/*
 *
 * scalprod.c : Example Athapascan-0b program
 *		Parallel Scalar Product
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * Feb 6, 1997
 *
 */

#include <a0b.h>
#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>


/* This is an example of a parallel program using Athapascan-0. It
   calculates the scalar product of two vectors, sequentially and in
   parallel. However, this is NOT a good example on how a scalar
   product should be done. The goal of this program is to show how to
   start remote threads, pass parameters, and receive results.
*/

/*************************************************************************
	generate_numbers - generates a random vector of floats
*************************************************************************/

void generate_numbers(int size, float *vector) {

  int i, num, den;

  for (i=0; i<size; i++) {
    num = 64 - (rand() & 127);
    den = 1 + (rand()&127);
    vector[i] = ((float) num) / ((float) den);
  }
}


/*************************************************************************
	multiply_scalar - calculates the scalar product of 2 vectors
*************************************************************************/

float multiply_scalar(int size, float *vector1, float *vector2) {

  int i, j;
  float sum = 0;

  for (i=0; i<size; i++) {
    /* The null for(;;) was inserted here in order to slow down this
       function. It makes communication more balanced with computation.
       The communication time needed to transfer the vectors is much
       greater than the computation time nedded to calculate the real
       scalar product. */
    for (j=0; j<1000; j++);
    sum += vector1[i] * vector2[i];
  }

  return sum;
}


/*************************************************************************
	multiply_service_function - service to call multiply_scalar
*************************************************************************/

a0tError multiply_service_function(a0tBuffer *input) {

  float *vector1, *vector2, result;
  int size;
  a0tPort answer_port;
  int answer_node;

  /* get info on where to receive data and send the result */
  a0Unpack(input, a0_PORT, &answer_port, 1);
  a0Unpack(input, a0_INT, &answer_node, 1);
  /* receive in tag=0 and send in tag=1 */

  /* get data */
  a0Unpack(input, a0_INT, &size, 1);
  vector1 = (float *) malloc(size * sizeof(float));
  vector2 = (float *) malloc(size * sizeof(float));
  if (vector1 && vector2) {
    a0Receive(&answer_port, answer_node, 0, NULL,
              a0_FLOAT, vector1, size);
    a0Receive(&answer_port, answer_node, 0, NULL,
              a0_FLOAT, vector2, size);
    result = multiply_scalar(size, vector1, vector2);
  }
  a0Send(&answer_port, answer_node, 1,
         NULL, a0_FLOAT, &result, 1);

  return A0ErrOk;
}


/*************************************************************************
	multiply_service - service number of multiply_service_function
*************************************************************************/
int multiply_service;


/*************************************************************************
	main_function_slaves - main function of nodes != 0
*************************************************************************/

void main_function_slaves(void) {
    /* do nothing */
}


/*************************************************************************
	do_parallel - calculates the scalar product in parallel
*************************************************************************/

float do_parallel(int size, float *vector1, float *vector2) {

  float result, *result_vector;
  int node, slice, curr_first, curr_size, which;
  size_t pksz, bufsz;
  a0tBuffer service_data;
  a0tPort answer_port;
  a0tRequest *request;
  a0tRequest **request_table;

  /* doing in parallel */
  /* how many items per-node */
  slice = (size + a0NodeCount - 1) / a0NodeCount;

  /* vector to store results from all nodes */
  result_vector = (float *) malloc(a0NodeCount * sizeof(float));

  /* vector to store the communication requests with all nodes */
  request = (a0tRequest *) malloc(a0NodeCount * sizeof(a0tRequest));

  /* vector to store the pointer to the comm.requests with all nodes */
  request_table = (a0tRequest **) malloc(a0NodeCount * sizeof(a0tRequest*));

  /* port to receive answer from all nodes */
  a0NewPort(&answer_port, 2, A0LocalCreation);

  /* get the size of packed data, to allocate buffer */
  a0GetPackedSize(a0_INT, &pksz);     bufsz = 2 * pksz;
  a0GetPackedSize(a0_PORT, &pksz);    bufsz += pksz;

  a0NewBuffer(&service_data, A0RTCBufferType, bufsz);

  curr_first = 0;
  for (node = 0; node<a0NodeCount; node++) {

    /* compute the size of a slice */
    curr_size = size - curr_first;
    if (curr_size > slice) curr_size = slice;
    /* pack port and node to the service answer */
    a0Pack(&service_data, a0_PORT, &answer_port, 1);
    a0Pack(&service_data, a0_INT, &a0SelfNode, 1);
    a0Pack(&service_data, a0_INT, &curr_size, 1);

    /* Create remote thread, passing data to it. Note the synchonous
       call to reuse the buffer. */
    a0StartRemoteThread(node, multiply_service,
        A0ServiceScheduling, A0ServicePriority, A0ServiceStack,
        NULL, /* null request */
        &service_data);
    a0ClearBuffer(&service_data); /* to restart packing */

    /* send slices of the vectors */
    /* receive in tag=1 and send in tag=0 */
    a0ISend(&answer_port, node, 0, NULL,
            a0_FLOAT, vector1 + curr_first, curr_size);
    a0ISend(&answer_port, node, 0, NULL,
            a0_FLOAT, vector2 + curr_first, curr_size);
    curr_first += curr_size;

    /* ask to receive result (asynchronous) */
    a0IReceive(&answer_port, node, 1, request + node,
        a0_FLOAT, result_vector + node, 1);
  }

  result = 0;
  /* prepare table to wait any */
  for (node = 0; node<a0NodeCount; node++)
    request_table[node] = request + node;

  /* wait all results and add it as soon as it arrives */
  for (node = 0; node<a0NodeCount; node++) {
    a0WaitAnyRequest(request_table, a0NodeCount, &which);
    result += result_vector[which];
    request_table[which] = NULL;
  }

  a0DisposeBuffer(&service_data);
  a0DisposePort(&answer_port);

  free(request);
  free(result_vector);
  free(request_table);

  return result;
}


/*************************************************************************
	main_function_master - main function of node 0
*************************************************************************/

void main_function_master(int size) {

  float *vector1, *vector2, result1, result2;
  struct timeval start_seq, end_seq, start_par, end_par;
  int i;

  vector1 = (float *) malloc(size * sizeof(float));
  vector2 = (float *) malloc(size * sizeof(float));
  if (!vector1 || !vector2) {
    fprintf(stderr, "scalprod: not enough memory\n");
    a0Abort(-2);
  }

  srand(time(NULL));
  generate_numbers(size, vector1);
  generate_numbers(size, vector2);

  printf("Vector 1:\n");
  for (i=0; (i<size) && (i<5); i++)
    printf("%f ", vector1[i]);
  printf("\nVector 2\n");
  for (i=0; (i<size) && (i<5); i++)
    printf("%f ", vector2[i]);
  printf("\n");

  /* doing sequentially */
  gettimeofday(&start_seq, 0);
  result1 = multiply_scalar(size, vector1, vector2);
  gettimeofday(&end_seq, 0);

  printf("Sequential: result=%f time=%f sec\n",
         result1,
         (float)(end_seq.tv_sec - start_seq.tv_sec) +
         ((float)(end_seq.tv_usec - start_seq.tv_usec)/1000000.0));

  /* doing in parallel */
  gettimeofday(&start_par, 0);
  result2 = do_parallel(size, vector1, vector2);
  gettimeofday(&end_par, 0);

  printf("Parallel:   result=%f time=%f sec\n",
         result2,
         (float)(end_par.tv_sec - start_par.tv_sec) +
         ((float)(end_par.tv_usec - start_par.tv_usec)/1000000.0));
}


/*************************************************************************
	main - main function
*************************************************************************/

void main(int argc, char *argv[]) {

  a0Init(&argc, &argv);
  a0NewService(&multiply_service, multiply_service_function,
               A0DefaultScheduling, A0DefaultPriority, A0DefaultStack);
  a0InitCommit();

  /* node 0 does not the same job as other nodes */
  if (a0SelfNode) main_function_slaves();
  else {
    if (argc != 2) {
      fprintf(stderr, "usage: scalprod <size>\n");
      a0Abort(-1);
    }

    main_function_master(atoi(argv[1]));
  }
  a0Terminate();
}
