/*
 * a0f-pipe.c : formatted put and get
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * Jun 30, 1997
 *
 * This program tests the formatted pipes of ath0b.
 *
 * This program must run in two nodes. If an argument is
 * passed, it is the number of messages sent by the node 1
 * to the node 0.
 *
 * The node 1 fills a table with a random number of integers with
 * random values. At the end, a checksum is added to verify if the
 * nubers received are the numbers sent. Data is sent in two puts, 
 * first contain the size, second contains data. Last put 
 * contains 0 and no data message is sent.
 *
 * The behavior of this program is the same as a0f-sr, but using a
 * pipe instead of messages.
 */

#include <a0format.h>
#include <stdio.h>


#define MAX	10000
#define TIMES	30


void fill_buffer(int *buf, int size)
{
	int i;

	if (size < 3)
		return;

	buf[0] = rand() % MAX;
	buf[1] = rand() % MAX;
	for (i = 2; i < size; i++)
		buf[i] = (i + buf[i - 1] * buf[i - 2]) % MAX;
}


int test_buffer(int *buf, int size)
{
	int i, p;

	if (size < 3)
		return 0;

	for (i = 2; i < size; i++) {
		p = (i + buf[i - 1] * buf[i - 2]) % MAX;
		if (buf[i] != p)
			return i + 1;
	}
	return 0;
}


int main(int argc, char *argv[])
{
	a0tRequest req;
	int size, ok, times;
	int buffer[MAX];
	a0tPipe mypipe;
	a0tPort myport;

	a0Init(&argc, &argv);
	a0UseFormats();
	a0NewPort(&myport, 1, 0);
	a0InitCommit();

	/* init a different random seed each time */
	srand(a0SelfNode * time(NULL));

	if (a0SelfNode == 0) {
		a0NewPipe(&mypipe, &myport, 1, 0, 100,
			  A0SyncPipe | A0InputPipe, A0InheritPriority);
		while (1) {
			a0FGet(&mypipe, &req, A0FInt, &size, 1);
			printf("received %d elements\n", size);
			if (!size)
				break;
			a0FGet(&mypipe, &req, A0FInt, buffer, size);
			ok = test_buffer(buffer, size);
			printf("status: %s (%d %d %d %d %d %d)\n",
			       ok ? "error!" : "ok.",
			       buffer[0], buffer[1], buffer[2],
			       buffer[3], buffer[4], buffer[5]);
		}
	}
	if (a0SelfNode == 1) {
		a0NewPipe(&mypipe, &myport, 0, 0, 100,
			  A0SyncPipe | A0OutputPipe, A0InheritPriority);
		if (argc > 1)
			times = atoi(argv[1]);
		else
			times = TIMES;

		while (1) {
			if (times--)
				size = (rand() % MAX) + 1;
			else
				size = 0;
			fill_buffer(buffer, size);
			a0FPut(&mypipe, &req, A0FInt, &size, 1);
			if (!size)
				break;
			a0FPut(&mypipe, &req, A0FInt, buffer, size);
		}
	}
	a0DisposePipe(&mypipe);
	a0Terminate();
}
