/*
 * ak-thr-sr.c : many sends and receives to different threads
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * Mar 10, 1996
 *
 * This program tests the multithreaded communication of aKernel.
 *
 * This program can run in any number of nodes. If an argument is
 * passed, it is the number of messages sent by the main thread of
 * the node 0.
 *
 * It starts three threads on each node, which receive a number,
 * multiply it by 10, 100 or 1000, depending on the thread, and
 * sent it back. The node 0 choses a number, a node and a thread
 * at random and sends the chosen number to the chosen thread in
 * the chosen node (it can be the node 0 itself). Every thread
 * receives from a different tag, so the main thread of the node 0
 * can use different tags to choose who to send the number.
 */

#include <akernel.h>
#include <stdio.h>

/* debugging function */
#define akDebug_on(a) printf a
#define akDebug(a)


int run = 1;

void *test_th1(void *x)
{
	int i;
	aktHandle handle;

	while (run) {
		akRecv(&i, 1, MPI_INT, 0, 1, MPI_COMM_WORLD, &handle);
		akDebug(("%d:th1 got %d\n", akNode, i));
		if (i < 0)
			return 0;
		i *= 10;
		akSend(&i, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &handle);
		akDebug(("%d:th1 put %d\n", akNode, i));
	}
}


void *test_th2(void *x)
{
	int i;
	aktHandle handle;

	while (run) {
		akRecv(&i, 1, MPI_INT, 0, 2, MPI_COMM_WORLD, &handle);
		akDebug(("%d:th2 got %d\n", akNode, i));
		if (i < 0)
			return 0;
		i *= 100;
		akSend(&i, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &handle);
		akDebug(("%d:th2 put %d\n", akNode, i));
	}
}


void *test_th3(void *x)
{
	int i;
	aktHandle handle;

	while (run) {
		akRecv(&i, 1, MPI_INT, 0, 3, MPI_COMM_WORLD, &handle);
		akDebug(("%d:th3 got %d\n", akNode, i));
		if (i < 0)
			return 0;
		i *= 1000;
		akSend(&i, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &handle);
		akDebug(("%d:th3 put %d\n", akNode, i));
	}
}


main(int argc, char **argv)
{
	akpthread_t t[3];
	void *(*f[3]) () =
	{test_th1, test_th2, test_th3};
	int i, tot;
	int rank, size;
	void *res;
	akpthread_attr_t attr;
	aktHandle handle;

	/* init a different random seed each time */
	srand(time(NULL));

	akInit(&argc, &argv);
	rank = akNode;
	size = akSize;

	if (argc > 1)
		tot = atoi(argv[1]);
	else
		tot = 10;

	akpthread_attr_init(&attr);
#ifdef A0_TAIX4
	akpthread_attr_setdetachstate(&attr, akPTHREAD_CREATE_JOINABLE);
#endif

	for (i = 0; i < 3; i++)
		akpthread_create(t + i, &attr, f[i], NULL);

	if (!rank) {
		int r, b, n, p;

		for (i = 0; i < tot; i++) {
			/* chose randomly a rank, a port and a number */
			r = rand() % size;
			b = (rand() % 3) + 1;
			n = (rand() & 63) + 10;

			printf("put %d to %d:%d", n, r, b);

			/* sends the number to the port of r */
			akISend(&n, 1, MPI_INT, r, b, MPI_COMM_WORLD, 0, 0, 0);

			/*and receives the result */
			akRecv(&p, 1, MPI_INT, r, 0, MPI_COMM_WORLD, &handle);

			printf(" - got result #%d: %d\n", i, p);
		}

		/* terminate everybody */
		n = -1;
		for (i = 0; i < size; i++) {
			akSend(&n, 1, MPI_INT, i, 1, MPI_COMM_WORLD, 0);
			akSend(&n, 1, MPI_INT, i, 2, MPI_COMM_WORLD, 0);
			akSend(&n, 1, MPI_INT, i, 3, MPI_COMM_WORLD, 0);
		}
	}
	for (i = 0; i < 3; i++)
		akpthread_join(t[i], &res);

	akFinalize();
}
