/*
 *
 * a0services.c : Define local services and call remote services
 *
 * by Ilan Ginzburg, Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * Feb 29, 1996
 *
 */


/* Athapascan0 source code */
#define A0_SOURCE

#include <akernel.h>
#include "a0error.h"
#include "a0thread.h"
#include "a0prio.h"
#include "a0services.h"
#include "a0ports.h"
#include "a0servd.h"
#include "a0urgentd.h"
#include "a0pt2pt.h"


/* debugging macro */
#define a0Debug_on(a) printf a
#define a0Debug(a)


/*
 * Use of _a0InitServices, _a0NewRRService, _a0NewRTCService, _a0LookupRRService
 * and _a0LookupRTCService.
 *
 * On all processors, all the services of the different (user program)
 * modules must be registered (_a0NewRRService and _a0NewRTCService) in the same
 * order, so that their identifiers will be the same across all the processors.
 * Such an identifier is used in a Remote Thread Creation call to create a
 * service, and the destination SERV daemon uses the
 * function _a0LookupRTCService to find out the adress of the code associated
 * with the service.
 * Similarly, the URGENT daemon uses _a0LookupRRService.
 * Calls to _a0NewRTCService AND _a0NewRRService must be done in the same order,
 * because the same counter is used for both types of services.
 */

/* Max number of different services used in an Athapascan program.
 * (this should diseappear when the association between a service and its
 * number is done more elegantly than an array...)
 */
#define A0MAXSERVICES   50


/*****************************************************************************/
/* Counter for service numbers... */
static int next_unused_number;


/* Array with all the information about all the services. The index in the
 * array is the service number.
 * All services have the same index in this array on all the nodes, but may
 * have different adresses (service field), and we can even think of different
 * Sched, Prio and StackSize, in heterogenous architectures...
 * To be continued...
 */
static struct {
  a0tRemoteFunction	Code;		/* Function  */
  a0tScheduling		Sched;		/* Default scheduling policy (if RTC) */
  int			Prio;		/* Default priority (if RTC) */
  size_t		StackSize;	/* Default stack size (if RTC) */
  } services_array[A0MAXSERVICES];


/*****************************************************************************/
/* Init the services table.
 */

a0tError _a0InitServices() {

  /* Starting service numbering at 0. */
  next_unused_number = 0;
}


/*****************************************************************************/
/* Returns the data associated with service service_num.
 * service_num must have been registered as an RTC service
 * (via _a0NewService()).
 */

a0tError _a0LookupRTCService(int service_num,
			 a0tRemoteFunction *RTCCode,
			 a0tScheduling *Sched,
			 int *Prio,
			 size_t *StackSize) {

  if (service_num >= next_unused_number || service_num < 0)
    return A0ErrInvalid;

  /* The adress of the function to execute */
  *RTCCode = services_array[service_num].Code;

  /* Scheduling policy, priority and stack size.
   * If the value of one of the following parameters is A0ServiceXxx, it means
   * that when the user did the call to _a0StartRemoteThread, he didn't pass
   * values for these parameters and intended the created service to use the
   * values defined at the time of _a0NewRTCService. So that's what we do.
   */
  if (*Sched == A0ServiceScheduling)
  	*Sched = services_array[service_num].Sched;
  if (*Prio == A0ServicePriority)
  	*Prio = services_array[service_num].Prio;
  if (*StackSize == A0ServiceStack)
  	*StackSize = services_array[service_num].StackSize;

  return A0ErrOk;
}


/*****************************************************************************/
/* Returns the data associated with service service_num.
 * service_num must have been registered as an RR service
 * (via _a0NewService()).
 */

a0tError _a0LookupRRService(int service_num,
			 a0tRemoteFunction *RRCode) {

  if (service_num >= next_unused_number || service_num < 0)
    return -1;

  /* Get all the information about the service */

  /* The adress of the function to execute (only one of these two is valid) */
  *RRCode = services_array[service_num].Code;

  return 0;
}


/*****************************************************************************/
/* Inserts a service in the services list, and returns the service identifier
 * associated with it (*service_num).
 * This service can be called by a Remote Thread Creation (RTC) and
 * by a Remote Request (RR).
 * All nodes must call this function in the same order, and the call defining
 * a service on a node must complete before that service is called on that node.
 */

a0tError _a0NewService(int *Service, a0tRTCFunction Code,
			a0tScheduling Sched, int Prio,
			size_t StackSize) {

  /* Test if Sched Prio and StackSize have 'correct' values: Inheritage is
   * not allowed for services (but they may be LOTS of other erroneous
   * values we don't test for...)
   */
  if (	(Sched == A0InheritScheduling) ||
  	(Prio == A0InheritPriority)  ||
  	(StackSize == A0InheritStack)  ) return A0ErrInvalid;
  
  /* Get a service identifier and register the number for the SERV daemon: */
  /* Test if we have room in the services table */
  if (next_unused_number >= A0MAXSERVICES) return -1;

  /* The code function to be called upon invocation of this service */
  services_array[next_unused_number].Code = Code;

  /* Default values for a call to that service */
  services_array[next_unused_number].Sched = Sched;
  services_array[next_unused_number].Prio = Prio;
  services_array[next_unused_number].StackSize = StackSize;

  /* We've got the number of the service */
  *Service = next_unused_number++;

  a0Debug(("Registered fct=%X for service %d "
		"prio=%d, sched=%d, stack=%d\n",
		Code, *Service, Prio, Sched, StackSize));

  return A0ErrOk;
}


/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/



/*****************************************************************************/
/* Remote Thread Creation call (RTC). (non-blocking)
 * Starts a new thread on a distant (or not) node. The thread started executes
 * a service previously registered.
 * The starting of the new thread is done by sending a message to the SERV
 * daemon on the distant node.
 * Parameters:
 *   IN node: node on which to start the thread
 *   IN service: service executed by the new thread
 *   IN sched: Scheduling policy of thread (or A0ServiceScheduling)
 *   IN prio: Priority of thread (or A0ServicePriority)
 *   IN stack: Stack size of thread (or A0ServiceStack)
 *   IN buffer: buffer to pass to the thread.
 */

extern a0tError _a0IStartRemoteThreadCallback(
			int node, int service,
			a0tScheduling sched, int prio, size_t stack,
			a0tRequest *requestp, a0tBuffer *buffer,
			void (*callback)(), void *callback_par) {

  long position_save;
  a0tError err;

  a0Debug(("Starting _a0IStartRemoteThreadCallback\n"));

  /* We've got a little hack here, but I didn't find any clean way to do that:
   * In the buffer we pass to the newly created thread, we start by packing
   * some (necessary) athapascan data. We have there:
   * service number to execute, scheduling, priority and stack size of the
   * thread.
   * When we allocated the buffer, we packed dummy variables (with no specific
   * value) into the buffer. Now we must go back to the beginning of the buffer
   * and to pack again the 'real' values.
   * If the size of a pack depends on the value of a variable and not only on
   * its type, this hack can't work.
   *
   *    Now let's pray.
   *
   */

  if (err = _a0CheckNode(node, 0)) return err;
  if (buffer->bufftype != A0ThreadBufferType) return A0ErrInvBuffType;

  /* Hey, who said a0tBuffer is an opaque type?... :-) */
  position_save = buffer->position;
  buffer->position = 0;

  _a0Pack(buffer, a0_INT, &service, 1);
  _a0Pack(buffer, a0_INT, &sched, 1);
  _a0Pack(buffer, a0_INT, &prio, 1);
  _a0Pack(buffer, a0_UNSIGNED_LONG, &stack, 1);

  buffer->position=position_save;
  /* Et voila. a0tBuffer is opaque again. :-) */

  a0Debug(("_a0IStartRemoteThreadCallback : tag=%d comm=%d stack=%ld node=%d "
		"service=%d sched=%d prio=%d\n",
  	_a0ServdPort.basetag, _a0ServdPort.comm, stack, node,
  	service, sched, prio));

  a0Debug(("_a0IStartRemoteThreadCallback before akSend. buffer->position=%d\n",
  			buffer->position));

  /* We send this buffer to the serv daemon on destination node and
     we're done. */

  return _a0KernelErrorConvert(
		akISend( buffer->storage, buffer->position, MPI_PACKED, node,
			_a0ServdPort.basetag,
			_a0Communicators[_a0ServdPort.comm],
			requestp, callback, callback_par));
}



/*****************************************************************************
	_a0IStartRemoteThread - see _a0IStartRemoteThreadCallback
*****************************************************************************/

extern a0tError _a0IStartRemoteThread(
			int node, int service,
			a0tScheduling sched, int prio, size_t stack,
			a0tRequest *requestp, a0tBuffer *buffer) {

  /* if no buffer is passed, build one */
  if (buffer)
    return _a0IStartRemoteThreadCallback(node, service,
					sched, prio, stack,
					requestp, buffer,
					0,0);
  else {
    a0tError err;
    a0tBuffer *internal;

    err = _a0AllocateBuffer(&internal, A0UrgentBufferType, 0);
    if (err) return err;

    return _a0IStartRemoteThreadCallback(node, service,
					sched, prio, stack,
					requestp, internal,
					(void (*)()) _a0DeallocateBuffer,
					(void *) internal);
  }
}


/*****************************************************************************
	_a0StartRemoteThread see _a0IStartRemoteThread
*****************************************************************************/

extern a0tError _a0StartRemoteThread(
			int node, int service,
			a0tScheduling sched, int prio, size_t stack,
			a0tRequest *requestp, a0tBuffer *buffer) {

  a0tError err;
  a0tRequest req_aux;

  if (!requestp) requestp = &req_aux;

  if (err = _a0IStartRemoteThread(node, service, sched, prio, stack,
			requestp, buffer))
	return err;

  return _a0WaitRequest(requestp);
}


/*****************************************************************************/
/* Remote Request call (RR). (non-blocking)
 * Executes a service function on a remote node, WITHOUT creating a new
 * thread. The function is executes by the URGENT daemon
 * Parameters:
 *   IN node: node on which to execute the request (service)
 *   IN service: service (request) to be executed
 *   IN buffer: buffer to pass to the service
 */

extern a0tError _a0IStartRemoteUrgentCallback(
			int node, int service,
			a0tRequest *requestp, a0tBuffer *buffer,
			void (*callback)(), void *callback_par) {

  long position_save;
  a0tError err;

  a0Debug(("Starting _a0IStartRemoteUrgentCallback\n"));

  /* We've got a little hack here, but I didn't find any clean way to do that:
   * In the buffer we pass to the newly created thread, we start by packing
   * some (necessary) athapascan data. We have there:
   * service number to execute.
   * When we allocated the buffer, we packed dummy variables (with no specific
   * value) into the buffer. Now we must go back to the beginning of the buffer
   * and to pack again the 'real' values.
   * If the size of a pack depends on the value of a variable and not only on
   * its type, this hack can't work.
   *
   *    Now let's pray.
   *
   */

  if (err = _a0CheckNode(node, 0)) return err;
  if (buffer->bufftype != A0UrgentBufferType) return A0ErrInvBuffType;

  /* Hey, who said a0tBuffer is an opaque type?... :-) */
  position_save = buffer->position;

  buffer->position = 0;
  _a0Pack(buffer, a0_INT, &service, 1);

  buffer->position = position_save;
  /* Et voila. a0tBuffer is opaque again. :-) */

  a0Debug(("_a0IStartRemoteUrgentCallback: tag %d, comm %d, node %d, service %d\n",
  	_a0UrgentdPort.basetag, _a0UrgentdPort.comm, node, service));

  a0Debug(("_a0IStartRemoteUrgentCallback before akSend. buffer->position=%d\n",
  			buffer->position));

  /* We send this buffer to the reqst daemon on destination node and
     we're done,. We do a non-blocking send and the callback changes if
     the buffer internal or not */

  return _a0KernelErrorConvert(
               akISend( buffer->storage, buffer->position, MPI_PACKED, node,
                       _a0UrgentdPort.basetag, _a0Communicators[_a0UrgentdPort.comm],
                       requestp, callback, callback_par));
}


/*****************************************************************************
	_a0IStartRemoteUrgent - see _a0IStartRemoteUrgentCallback
*****************************************************************************/

extern a0tError _a0IStartRemoteUrgent(
			int node, int service,
			a0tRequest *requestp, a0tBuffer *buffer) {

  /* if no buffer is passed, build one */
  if (buffer)
    return _a0IStartRemoteUrgentCallback(node, service,
					requestp, buffer,
					0,0);
  else {
    a0tError err;
    a0tBuffer *internal;

    err = _a0AllocateBuffer(&internal, A0UrgentBufferType, 0);
    if (err) return err;

    return _a0IStartRemoteUrgentCallback(node, service,
					requestp, internal,
					(void (*)()) _a0DeallocateBuffer,
					(void *) internal);
  }
}


/*****************************************************************************
	_a0StartRemoteUrgent blocking version of _a0IStartRemoteUrgent
*****************************************************************************/

extern a0tError _a0StartRemoteUrgent(
			int node, int service,
			a0tRequest *requestp, a0tBuffer *buffer) {

  a0tError err;
  a0tRequest req_aux;

  if (!requestp) requestp = &req_aux;

  if (err = _a0IStartRemoteUrgent(node, service, requestp, buffer))
	return err;

  return _a0WaitRequest(requestp);
}
