/*
 *
 * a0fnew.c : _a0 New format functions
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * May 23, 1996
 *
 */

#define A0_SOURCE
#include <ath0.h>
#include <stdlib.h>

#define A0F_SOURCE
#include "a0fuser.h"
#include "a0fmap.h"
#include "a0fpipe.h"


#define a0Debug(a)
#define a0Debug_on(a) printf a


/******************local variables & definitions*****************/

/* types defined only to use in a va_arg */
typedef a0tFormat *a0tFormatP;
typedef int	  *a0tIntP;
typedef void	  *a0tVoidP;


/* auxiliar structure to parse a format list */
typedef struct {
  va_list *parametersp;
  int count;
  MPI_Aint baseaddr;
  a0tMap **mapptab;
  int *blltab;
  MPI_Aint *dsptab;
} a0tMapList;


/*************************************************************************
		_a0MakeDatatype - produce a datatype to a map
*************************************************************************/

static a0tError _a0MakeDatatype(a0tMap *mapp) {

  a0tError r;
  int i;
  MPI_Datatype *oldtab;
  register a0tMap *oldmapp;

  a0Debug(("%d:Make datatype map=[%08X]\n", a0SelfNode, mapp));
  /* if the map already has a datatype, there is nothing to do */
  if (mapp->status & A0MapHasDatatype) return A0ErrOk;

#ifdef PUT_IT_BACK_MARCELO
  switch (mapp->maptype) {
  case A0ContiguousMap:
  case A0VectorMap:
  case A0HVectorMap:
  case A0IndexedMap:
  case A0HIndexedMap:
    oldmapp = mapp->umap.oldmapp;
    /* certify that there is a datatype for the oldmap */
    if (!(oldmapp->status & A0MapHasDatatype))
      if (r = _a0MakeDatatype(oldmapp)) return r;
    break;

  case A0StructMap:
    /* allocate a datatype table */
    if (!(oldtab = malloc(mapp->count * sizeof(MPI_Datatype))))
      return A0ErrNoMem;
    /* certify that there are datatypes for all maps in the table */
    for (i=0; i<mapp->count; i++) {
      oldmapp = mapp->umap.oldmapptab[i];
      /* certify that there is a datatype for the oldmap */
      if (!(oldmapp->status & A0MapHasDatatype))
	if (r = _a0MakeDatatype(oldmapp)) return r;
      oldtab[i] = oldmapp->datatype;
    }
    break;

  default:
    return A0ErrInvFormat;
  }

  /* call MPI to make the new datatypes */
  switch (mapp->maptype) {
  case A0ContiguousMap:
    if (akMPI_Type_contiguous(mapp->count,
				oldmapp->datatype,
				&(mapp->datatype)))
	_a0Error(A0ErrCommType, "make map");
    break;

  case A0VectorMap:
    if (akMPI_Type_vector(mapp->count,
				mapp->ubll.blocklength,
				mapp->udsp.stride,
				oldmapp->datatype,
				&(mapp->datatype)))
	_a0Error(A0ErrCommType, "make map");
    break;

  case A0HVectorMap:
    if (akMPI_Type_hvector(mapp->count,
				mapp->ubll.blocklength,
				(MPI_Aint) mapp->udsp.stride,
				oldmapp->datatype,
				&(mapp->datatype)))
	_a0Error(A0ErrCommType, "make map");
    break;

  case A0IndexedMap:
    if (akMPI_Type_indexed(mapp->count,
				mapp->ubll.blocklengthtab,
				mapp->udsp.displacementtab,
				oldmapp->datatype,
				&(mapp->datatype)))
	_a0Error(A0ErrCommType, "make map");
    break;

  case A0HIndexedMap:
    if (akMPI_Type_hindexed(mapp->count,
				mapp->ubll.blocklengthtab,
				mapp->udsp.Adisplacementtab,
				oldmapp->datatype,
				&(mapp->datatype)))
	_a0Error(A0ErrCommType, "make map");
    break;

  case A0StructMap:
    if (akMPI_Type_struct(mapp->count,
				mapp->ubll.blocklengthtab,
				mapp->udsp.Adisplacementtab,
				oldtab,
				&(mapp->datatype)))
	_a0Error(A0ErrCommType, "make map");
    break;
  }

  mapp->status |= A0MapHasDatatype;
  a0Debug(("Make datatype map=[%08X] ended OK (%X)\n", mapp,
		  mapp->datatype));

  return A0ErrOk;
#else
  return A0ErrInvFormat;
#endif
}


/*************************************************************************
		_a0GetMapDatatype - find out a commited MPI datatype
*************************************************************************/

extern a0tError _a0GetMapDatatype(a0tFormat *formatp,
					MPI_Datatype *datatypep) {

  a0tError r;
  register a0tMap *mapp;

  /* if the format is not a map: BOOM! */
  if (formatp->type != A0Map) return A0ErrInvFormat;

  mapp = formatp->u.mapp;
  /* if the map has not a datatype, produce it */
  if (!(mapp->status & A0MapHasDatatype))
    if (r = _a0MakeDatatype(mapp)) return r;

  /* if the datatype is not committed, do it */
  if (!(mapp->status & A0MapHasCommitted)) {
    if (akMPI_Type_commit(&(mapp->datatype)))
	_a0Error(A0ErrCommType, "commit format");
    mapp->status |= A0MapHasCommitted;
  }
  *datatypep = mapp->datatype;

  a0Debug(("%d:Using datatype %x\n", a0SelfNode, *datatypep));

  return A0ErrOk;
}


/*************************************************************************
	_a0GetMapOffset - find out the offset of a map element
*************************************************************************/

extern a0tError _a0GetMapOffset(a0tFormat *formatp, void *address, int count,
				void **offsetp) {

  a0tError r;
  register a0tMap *mapp;
  MPI_Aint extent;

  /* if the format is not a map: BOOM! */
  if (formatp->type != A0Map) return A0ErrInvFormat;

  a0Debug(("make extent %d\n", mapp->datatype));
  mapp = formatp->u.mapp;
  /* if the extent is already calculated: nothing to do :) */
  if (!(mapp->status & A0MapHasExtent)) {
    /* is there a datatype for the map? */
    if (!(mapp->status & A0MapHasDatatype))
      /* no datatype, produce it. */
      if (r = _a0MakeDatatype(mapp)) return r;

    /* Well, we don't know the extent: let's calculate it. */
    if (akMPI_Type_extent(mapp->datatype, &extent))
	_a0Error(A0ErrCommType, "extent format");
    mapp->extent = extent;
    mapp->status |= A0MapHasExtent;
  }
  *offsetp = (void *)( (char *)address + (mapp->extent * count) );

  a0Debug(("make extent %d is %d\n", mapp->datatype, mapp->extent));

  return A0ErrOk;
}


/*************************************************************************
		_a0GetMapSize - find out the map size
*************************************************************************/

extern a0tError _a0GetMapSize(a0tFormat *formatp,
				void *address, int count,
				size_t *sizep) {

  a0tError r;
  register a0tMap *mapp;
  int mpi_size;

  /* if the format is not a map: BOOM! */
  if (formatp->type != A0Map) return A0ErrInvFormat;

  a0Debug(("make size %d\n", mapp->datatype));
  mapp = formatp->u.mapp;
  /* if the size is already calculated: nothing to do :) */
  if (mapp->status & A0MapHasSize) {
    *sizep = mapp->size;
    return A0ErrOk;
  }

  /* is there a datatype for the map? */
  if (!(mapp->status & A0MapHasDatatype))
    /* no datatype, produce it. */
    if (r = _a0MakeDatatype(mapp)) return r;

  /* Well, we don't know the extent: let's calculate it. */
  if (akMPI_Type_size(mapp->datatype, &mpi_size))
	_a0Error(A0ErrCommType, "extent format");
  *sizep = mpi_size;
  mapp->status |= A0MapHasSize;
  mapp->size = mpi_size;
  a0Debug(("make size %d is %d\n", mapp->datatype, mapp->extent));
  return A0ErrOk;
}


#ifdef PUT_IT_BACK_MARCELO
/*************************************************************************
		_a0GetMapList - Get a format list from va_list
*************************************************************************/

static a0tError _a0GetMapList(a0tMapList *maplistp) {

  a0tFormat *oldformatp;
  int bll, count, dsp, size;
  a0tError r;
  void *addr;
  MPI_Aint mpi_addr;

  /* a format list structure is passed to _a0GetMapList, which
     contains the va_list being parsed and the number of formats
     already parsed. This function calls itself recursively and
     when there are no more formats to parse, it allocates a
     table of size <maplist.count>. After the return of the function,
     the values of the current parsed bll,format,addr are stored
     in the tables. (the tables are also passed inside maplist).
     The list to be parsed is ... bll, &format, bll, &format, 
     bll, &format, finished by a bll with a zero value. */

  /* save the current count */
  count = maplistp->count;

  /* get a new blocklength */
  bll  = va_arg(*(maplistp->parametersp), int);
  if (!bll) {
    /* No more formats, the list is ended. Allocate the table and return */
    if (!(maplistp->blltab = malloc(count * sizeof(int))))
      return A0ErrNoMem;
    if (!(maplistp->dsptab = malloc(count * sizeof(MPI_Aint)))) {
      free(maplistp->blltab);
      return A0ErrNoMem;
    }
    if (!(maplistp->mapptab = malloc(count * sizeof(a0tMap *)))) {
      free(maplistp->blltab);
      free(maplistp->dsptab);
      return A0ErrNoMem;
    }
    return A0ErrOk;
  }
  /* get a new format and its address */
  oldformatp = va_arg(*(maplistp->parametersp), a0tFormatP);
  addr       = va_arg(*(maplistp->parametersp), a0tVoidP);
  akMPI_Address(addr, &mpi_addr);

  /* only A0Map's can be used to compose new formats */
  if (oldformatp->type != A0Map) return A0ErrInvFormat;

  /* increments counter */
  maplistp->count++;
  /* take the rest of the list */
  if (r = _a0GetMapList(maplistp)) return r;

  /* insert this format in the table, if bll is positive */
  if (bll > 0) {
    maplistp->mapptab[count] = oldformatp->u.mapp;
    maplistp->blltab[count] = bll;
    maplistp->dsptab[count] = mpi_addr - maplistp->baseaddr;
  }
  return A0ErrOk;
}


/*************************************************************************
	_a0NewMapVaList - creates a map format from a va_list
*************************************************************************/

extern a0tError _a0NewMapVaList(a0tFormat *newformatp, 
			    a0tMapType maptype, va_list parameters) {

  int i;
  a0tMap map, *mapp;
  a0tFormat *oldformatp, **oldformatptab;
  a0tMapList maplist;
  a0tError r;
  void *addr;

  map.maptype = maptype;

  switch (maptype) {
  case A0MpiDatatype:
    /* _a0NewFormat(new, A0MpiDatatype, datatype) */
    map.count	= 0;
    map.datatype	= va_arg(parameters, MPI_Datatype);
    map.status	= A0MapHasDatatype;
    map.umap.oldmapp	= 0;
    break;

  case A0ListMap:
    /* _a0NewFormat(newfmt, type, baseaddr, cnt,fmt,addr, c,f,a, ...) */

    maplist.count = 0;
    addr = va_arg(parameters, a0tVoidP);
    akMPI_Address(addr, &maplist.baseaddr);
    maplist.parametersp = &parameters;

    /* take the rest of the list */
    if (r = _a0GetMapList(&maplist)) return r;
    a0Debug(("%d:list has %d elems\n", a0SelfNode, maplist.count));

    /* store the tables in the new format */
    map.umap.oldmapptab	= maplist.mapptab;
    map.ubll.blocklengthtab	= maplist.blltab;
    map.udsp.Adisplacementtab	= maplist.dsptab;
    map.count = maplist.count;
    /* a list is treated as a structure from now and then */
    map.maptype = A0StructMap;
    map.status	= A0MapHasNothing;
    break;
  case A0IndexedMap:
  case A0HIndexedMap:
  case A0HVectorMap:
  case A0StructMap:
    return A0ErrNotImplem;

  case A0ContiguousMap:
    /* _a0NewFormat(new, type, count, old) */
  case A0VectorMap:
    /* _a0NewFormat(new, type, count, old, blocklength, stride) */

    map.count  = va_arg(parameters, int);
    oldformatp = va_arg(parameters, a0tFormatP);

    if (oldformatp->type != A0Map) return A0ErrInvFormat;
    map.umap.oldmapp	= oldformatp->u.mapp;
    map.status	= A0MapHasNothing;

    if (maptype == A0ContiguousMap) break;
  
    map.ubll.blocklength = va_arg(parameters, int);
    map.udsp.stride      = va_arg(parameters, int);
    break;
  }

  /* we arrived here so far, allocate a new map */
  if (!(mapp = (a0tMap *)malloc(sizeof(a0tMap)))) return A0ErrNoMem;

  /* fill out the user format structure */
  newformatp->type	= A0Map;
  newformatp->u.mapp	= mapp;

  /* copy the filled map created to the new map */
  memcpy(mapp, &map, sizeof(a0tMap));

  return A0ErrOk;
}
#endif


/*************************************************************************
	_a0SetMapField - creates a map field
*************************************************************************/

extern a0tError _a0SetMapField(a0tMapField *fieldp, int index,
				void *base_addr, a0tFormat *formatp,
				void *address, int count) {

  MPI_Aint Abase, Afield;
  register a0tMap *mapp;
  a0tError r;

  /* if the format is not a map: BOOM! */
  if (formatp->type != A0Map) return A0ErrInvFormat;

  a0Debug(("map field %d\n", mapp->datatype));
  mapp = formatp->u.mapp;

  /* is there a datatype for the map? */
  if (!(mapp->status & A0MapHasDatatype))
    /* no datatype, produce it. */
    if (r = _a0MakeDatatype(mapp)) return r;

  akMPI_Address(base_addr, &Abase);
  akMPI_Address(address, &Afield);
  fieldp[index].Adisp = Afield - Abase;
  fieldp[index].blocklen = count;
  fieldp[index].datatype = mapp->datatype;
}


/*************************************************************************
	_a0NewMapFormat - creates a map format
*************************************************************************/

extern a0tError _a0NewMapFormat(a0tFormat *newformatp,
				int size,
				int count, a0tMapField *fields) {

  int i;
  a0tMap map, *mapp;
  a0tError r;
  void *addr;
#ifdef __GNUC__
  MPI_Datatype datatypetab[count+2];
  MPI_Aint Adisplacementtab[count+2];
  int blocklengthtab[count+2];
#else
  MPI_Datatype *datatypetab;
  MPI_Aint *Adisplacementtab;
  int *blocklengthtab;
#endif

  i = count + 2;

  mapp = malloc(sizeof(a0tMap));
  if (!mapp) return A0ErrNoMem;
#ifndef __GNUC__
  /* allocate all needed memory */
  datatypetab = malloc(i * sizeof(MPI_Datatype));
  blocklengthtab = malloc(i * sizeof(int));
  Adisplacementtab = malloc(i * sizeof(MPI_Aint));
  if (!datatypetab || !blocklengthtab || !Adisplacementtab) {
    if (datatypetab)	free(datatypetab);
    if (blocklengthtab) free(blocklengthtab);
    if (Adisplacementtab) free(Adisplacementtab);
    free(mapp);
    return A0ErrNoMem;
  }
#endif /* __GNUC__ */

  /* lower bound */
  datatypetab[0] = a0_LB;
  blocklengthtab[0] = 1;
  Adisplacementtab[0] = 0;

  /* user fields */
  for (i=0; i<count; i++) {
    datatypetab[i+1] = fields[i].datatype;
    blocklengthtab[i+1] = fields[i].blocklen;
    Adisplacementtab[i+1] = fields[i].Adisp;
  }

  /* upper bound */
  datatypetab[count + 1] = a0_UB;
  blocklengthtab[count + 1] = 1;
  Adisplacementtab[count + 1] = size;

  if (akMPI_Type_struct(count + 2, blocklengthtab, Adisplacementtab,
			datatypetab, &mapp->datatype))
	_a0Error(A0ErrCommType, "field map");

#ifndef __GNUC__
  free(datatypetab);
  free(blocklengthtab);
  free(Adisplacementtab);
#endif /* __GNUC__ */

#ifdef PUT_IT_BACK_MARCELO
  mapp->maptype = A0FieldMap;
#endif
  mapp->status	= A0MapHasDatatype;

  /* fill out the user format structure */
  newformatp->type	= A0Map;
  newformatp->u.mapp	= mapp;

  return A0ErrOk;
}


/*************************************************************************
		_a0MapPack - packs formatted to a buffer
*************************************************************************/

extern a0tError _a0MapPack(a0tBuffer *buffer,
		 a0tFormat *formatp, void *address, int count) {

  a0tDatatype datatype;
  a0tError ret;

  /* if the format is not a map: BOOM! */
  if (formatp->type != A0Map) return A0ErrInvFormat;

  if (buffer->bufftype == A0PipeBufferType)
      return _a0PipePack(buffer, formatp, address, count);

  /* Convert format to datatype */
  if (ret = _a0GetMapDatatype(formatp, &datatype))
	return ret;

  return _a0Pack(buffer, datatype, address, count);
}


/*************************************************************************
		_a0MapUnpack - unpacks formatted from a buffer
*************************************************************************/

extern a0tError _a0MapUnpack(a0tBuffer *buffer,
		 a0tFormat *formatp, void *address, int count) {

  a0tDatatype datatype;
  a0tError ret;

  /* if the format is not a map: BOOM! */
  if (formatp->type != A0Map) return A0ErrInvFormat;

  if (buffer->bufftype == A0PipeBufferType)
      return _a0PipeUnpack(buffer, formatp, address, count);

  /* Convert format to datatype */
  if (ret = _a0GetMapDatatype(formatp, &datatype))
	return ret;

  return _a0Unpack(buffer, datatype, address, count);
}
