/*
 *
 * a0fxdr.c : a0 xdr functions
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * May 28, 1996
 *
 */

#define A0_SOURCE
#include <ath0.h>
#include <stdarg.h>

#define A0F_SOURCE
#include "a0fuser.h"
#include "a0fxdr.h"


#define a0Debug(a)
#define a0Debug_on(a) printf a


/******************local variables & definitions*****************/

typedef struct {
  a0tFormat	*format;
  void		*address;
  int		count;
  a0tBuffer	buffer;
  void		(*callback)();
  void		*callback_par;
} a0tXdrReceiveRequest;

typedef struct {
  a0tBuffer	buffer;
  void		(*callback)();
  void		*callback_par;
} a0tXdrSendRequest;


/*************************************************************************
		_a0BuildXdrBuffer -
*************************************************************************/

static a0tError _a0BuildXdrBuffer(a0tBuffer *descr) {

  /* build a 'special' buffer (see _a0FPack) */
  descr->bufftype = A0XdrBufferType;
  descr->storage = 0;
  descr->size = 0;
}


/*************************************************************************
		_a0AddXdrSize - 
*************************************************************************/

extern a0tError _a0AddXdrSize(a0tBuffer *bufferp,
			 a0tFormat *formatp, void *address, int count){

  size_t size;
  a0tError ret;

  a0Debug(("_a0AddXdrSize: fmt=%x add=%x cnt=%x\n", formatp, address, count));

  /* accumulate element size */
  ret = _a0GetFormatSize(formatp, address, count, &size);
  bufferp->size += size;

  a0Debug(("_a0AddXdrSize: ret=%d\n", ret));
  return ret;
}


/*************************************************************************
		_a0GetXdrSize - find out the xdr format pack size
*************************************************************************/

extern a0tError _a0GetXdrSize(a0tFormat *formatp,
			void *address, int count,
			size_t *sizep) {

  a0tBuffer descr;
  a0tError ret;

  /* if the format is not a xdr: BOOM! */
  if (formatp->type != A0Xdr) return A0ErrInvFormat;

  /* if user supplied size function, yuppie! */
  if (formatp->u.xdrp->sizefun)
    return formatp->u.xdrp->sizefun(formatp, address, count, sizep);

  /* if user didn't supply the pack function either, return error */
  if (!formatp->u.xdrp->packfun) return A0ErrNotImplem;

  /* build a fake buffer descriptor */
  _a0BuildXdrBuffer(&descr);

  /* ask function to pack. It will only count space needed */
  ret = formatp->u.xdrp->packfun(&descr, formatp, address, count);
  *sizep = descr.size;

  a0Debug(("_a0GetXdrSize: returned %d size %d\n", ret, (int)*sizep));

  return ret;
}


/*************************************************************************
_a0XdrIReceiveCleanup - second cleanup of _a0XdrIReceiveCallback
*************************************************************************/

static void _a0XdrIReceiveCleanup(a0tXdrReceiveRequest *descrp) {

  /* see _a0XdrIReceiveCallback to know how this func works */

  a0Debug(("_a0XdrIReceiveCleanup: descr=%x\n", descrp));

  /* Call user function. Hope it's less bugged than athapascan! */
  descrp->format->u.xdrp->unpackfun(&descrp->buffer,
		descrp->format, descrp->address, descrp->count);

  /* cleanup */
  _a0DisposeBuffer(&descrp->buffer);

  /* call user callback */
  if (descrp->callback) descrp->callback(descrp->callback_par);

  free(descrp);
}


/*************************************************************************
_a0XdrIReceiveCallback - Receives a formatted message (non-blocking)
*************************************************************************/

extern a0tError _a0XdrIReceiveCallback(
		a0tPort *port, int src, int tag, a0tRequest *req,
		a0tFormat *formatp, void *address, int count,
		void (*callback)(), void *callback_par) {

  a0tError ret;
  a0tXdrReceiveRequest *descrp;
  unsigned long size;

  /* The receive of XDR data happens as follows:
	- Starts a blocking receive, to know how many bytes arrived.
	- Allocates a buffer which is large enough to contain all data
	  and post a receive for it.
	- The callback function calls the user unpack function, to unpack
	  all data from the buffer and frees the buffer.

     This is what we've called "almost non-blocking call." If you have
     a better idea, please, let us know!
  */

  /* I KNOW IT'S A FAKE RECEIVE!!! But it's not going to be used a lot,
     so, who cares? It's just done to keep the interface compatible. */

  /* if the format is not a map: BOOM! */
  if (formatp->type != A0Xdr) return A0ErrInvFormat;
  /* if user didn't supply the function, return */
  if (!formatp->u.xdrp->unpackfun) return A0ErrNotImplem;

  /* allocate and build a descriptor. This is done just to remember
     what the request was. */
  descrp = (a0tXdrReceiveRequest *) malloc(sizeof(a0tXdrReceiveRequest));
  if (!descrp) return A0ErrNoMem;

  a0Debug(("_a0XdrIReceiveCallback: descr=%x\n", descrp));

  descrp->format = formatp;
  descrp->address = address;
  descrp->count = count;
  descrp->callback = callback;
  descrp->callback_par = callback_par;

  /* receive size */
  ret =_a0Receive(port, src, tag, NULL,
		a0_UNSIGNED_LONG, &size, 1);

  a0Debug(("_a0XdrIReceiveCallback: size %ld\n", size));

  /* allocate buffer */
  if (_a0NewBuffer(&descrp->buffer, A0SendBufferType, size)) return;

  /* receive it */
  _a0IReceiveBufferCallback(port, src, tag, req, &descrp->buffer,
		_a0XdrIReceiveCleanup, (void *) descrp);

  a0Debug(("_a0XdrIReceiveCallback: ended %d\n", ret));

  return ret;
}


/*************************************************************************
	_a0XdrISendCleanup - cleanup of _a0XdrISendCallback
*************************************************************************/

static void _a0XdrISendCleanup(a0tXdrSendRequest *descrp) {

  /* see _a0XdrISendCallback to know how this func works */

  a0Debug(("_a0XdrISendCleanup: descr=%x\n", descrp));

  _a0DisposeBuffer(&descrp->buffer);

  /* call user callback */
  if (descrp->callback) descrp->callback(descrp->callback_par);

  free(descrp);
}


/*************************************************************************
_a0XdrISendCallback - Sends a formatted message (non-blocking)
*************************************************************************/

extern a0tError _a0XdrISendCallback(
		a0tPort *port, int dst, int tag, a0tRequest *req,
		a0tFormat *formatp, void *address, int count,
		void (*callback)(), void *callback_par) {

  a0tXdrSendRequest *descrp;
  a0tError ret = A0ErrOk;
  size_t size;

  /* The send of XDR data happens as follows:
	- Calculate the size of data.
	- Allocates a buffer and packs all data in it.
	- Requests to send the buffer.
	- A callback function is patched in the request.
	- When the request ends, the callback function is called. It
	  frees the buffer.
  */

  /* if the format is not a map: BOOM! */
  if (formatp->type != A0Xdr) return A0ErrInvFormat;
  /* if user didn't supply the function, return */
  if (!formatp->u.xdrp->packfun) return A0ErrNotImplem;

  /* allocate and build a descriptor and its buffer */
  descrp = (a0tXdrSendRequest *) malloc(sizeof(a0tXdrSendRequest));
  if (!descrp) return A0ErrNoMem;

  a0Debug(("_a0XdrISendCallback: descr=%x\n", descrp));

  /* Get the size to new buffer */
  if (ret = _a0GetXdrSize(formatp, address, count, &size))
    return ret;

  a0Debug(("_a0XdrISendCallback: xdr sz=%d\n", size));

  descrp->callback = callback;
  descrp->callback_par = callback_par;

  /* build a buffer to pack all data */
  if (!(ret = _a0NewBuffer(&descrp->buffer, A0SendBufferType, size))) {

    /* Call user function. Hope it's less bugged than athapascan! */
    ret = formatp->u.xdrp->packfun(&descrp->buffer, formatp, address, count);

    a0Debug(("_a0XdrISendCallback: buf sz=%d pos=%d\n", 
    		descrp->buffer.size, 
    		descrp->buffer.position));
    if (!ret) {
      /* ask to send data */
      ret = _a0ISend(port, dst, tag, NULL,
			a0_UNSIGNED_LONG, &descrp->buffer.position, 1);
    }
    if (!ret) {
      /* ask to send data */
      ret = _a0ISendBufferCallback(port, dst, tag, req,
				&descrp->buffer,
				_a0XdrISendCleanup, (void *) descrp);
    }
  }

  a0Debug(("_a0XdrISend: ended %d\n", ret));

  return ret;
}


/*************************************************************************
		_a0XdrPack - packs formatted to a buffer
*************************************************************************/

extern a0tError _a0XdrPack(a0tBuffer *buffer,
		 a0tFormat *formatp, void *address, int count) {

  /* if the format is not a xdr: BOOM! */
  if (formatp->type != A0Xdr) return A0ErrInvFormat;
  /* if user didn't supply the function, return */
  if (!formatp->u.xdrp->packfun) return A0ErrNotImplem;

  return formatp->u.xdrp->packfun(buffer, formatp, address, count);
}


/*************************************************************************
		_a0XdrUnpack - unpacks formatted from a buffer
*************************************************************************/

extern a0tError _a0XdrUnpack(a0tBuffer *buffer,
		 a0tFormat *formatp, void *address, int count) {

  /* if the format is not a xdr: BOOM! */
  if (formatp->type != A0Xdr) return A0ErrInvFormat;
  /* if user didn't supply the function, return */
  if (!formatp->u.xdrp->unpackfun) return A0ErrNotImplem;

  return formatp->u.xdrp->unpackfun(buffer, formatp, address, count);
}


#ifdef PUT_IT_BACK_MARCELO
/*************************************************************************
	_a0NewXdrVaList - creates an xdr format from a va_list
*************************************************************************/

extern a0tError _a0NewXdrVaList(a0tFormat *newformatp, 
			    a0tFormatType maptype, va_list parameters) {

  a0tXdr *xdrp;

  /* we arrived here so far, allocate a new xdr */
  if (!(xdrp = (a0tXdr *)malloc(sizeof(a0tXdr)))) return A0ErrNoMem;

  xdrp->packfun = 	va_arg(parameters, a0tPackFunction);
  xdrp->unpackfun = 	va_arg(parameters, a0tPackFunction);
  xdrp->sizefun = 	va_arg(parameters, a0tSizeFunction);
  xdrp->offsetfun = 	va_arg(parameters, a0tOffsetFunction);

  /* fill out the user format structure */
  newformatp->type	= A0Xdr;
  newformatp->u.xdrp	= xdrp;

  return A0ErrOk;
}
#endif


/*************************************************************************
	_a0NewFunctionFormat - creates an xdr format
*************************************************************************/

extern a0tError _a0NewFunctionFormat(a0tFormat *newformatp,
				a0tPackFunction packfun,
				a0tPackFunction unpackfun,
				a0tSizeFunction sizefun,
				a0tOffsetFunction offsetfun) {

  a0tXdr *xdrp;

  /* we arrived here so far, allocate a new xdr */
  if (!(xdrp = (a0tXdr *)malloc(sizeof(a0tXdr)))) return A0ErrNoMem;

  xdrp->packfun = 	packfun;
  xdrp->unpackfun = 	unpackfun;
  xdrp->sizefun = 	sizefun;
  xdrp->offsetfun = 	offsetfun;

  /* fill out the user format structure */
  newformatp->type	= A0Xdr;
  newformatp->u.xdrp	= xdrp;

  return A0ErrOk;
}
