
/*
 *
 * akcomm.c : Communication functions
 *
 *
 * by Marcelo Pasin, Ilan Ginzburg
 * Apache Project, IMAG, Grenoble, France
 * Feb 5, 1996
 *
 */

#define AKERNEL_SOURCE
#include "akthread.h"
#include "akconfig.h"
#include "akmpi.h"
#include "akcomm.h"
#include "akdaemon.h"
#include "akerror.h"

/* debugging function */
#define akDebug_on(a) printf a
#define akDebug(a) 

int akNode; /* Rank of the node */
int akSize; /* Number of nodes in the computation */

/* conters of blocked threads. one for each WC type + semaphore and mutex */
int akBlocked[akWCNTYPES + 2];

#ifdef A0_TRACE
/* Variable counting the total number of requests.
 * First call is returned 0, second 1 etc...
 * The numerotation reflects the ACTUAL order in which MPI has been called
 * (we guarantee this because as of the time of writing, all calls to MPI are
 * done in mutual exclusion, and we increment the counter and call MPI
 * atomically)
 */
static int akCounter[akWCNTYPES];
#endif /* A0_TRACE */

/*#define AK_COUNT*/
#ifdef AK_COUNT
static unsigned long akCountCommWait = 0;
static unsigned long akCountCommCompletedSend = 0;
static unsigned long akCountCommCompletedRecv = 0;
static unsigned long akCountCommCompletedProbe = 0;
static unsigned long akCountCommCompleted = 0;
static unsigned long akCountCommPostedSend = 0;
static unsigned long akCountCommPostedRecv = 0;
static unsigned long akCountCommPostedProbe = 0;
#endif


/*************************************************************************
akInit - Starts the kernel up
*************************************************************************/

extern int akInit(int *argcp, char ***argvp) {

  /* Temporary storage for the creation of akWakerUp thread */
  int n;

  akDebug(("Starting akInit...\n"));

  akpthread_init_np();

  /* startup semaphores: it must be changed to a dynamic structure,.
     in the present, we have a fixed quantity of semaphores. */
  if (akMPI_Init(argcp, argvp)) akError(akERRMPI, "init" );
  if (akMPI_Comm_rank(MPI_COMM_WORLD, &akNode)) akError(akERRMPI, "rank" );
  if (akMPI_Comm_size(MPI_COMM_WORLD, &akSize)) akError(akERRMPI, "size" );

#ifdef A0_TRACE
  /* get global clock initial time */
  a0_perf_define_initial_time ();
  /* Counting the number of send & receives */
  for (n=0; n<akWCNTYPES; n++) akCounter[n] = 0;
#endif

  /* Counting the number of blocked threads */
  for (n=0; n<akWCNTYPES; n++) akBlocked[n] = 0;

  akInitWC();

  akDebug(("Starting akDaemonInit...\n"));
 
 if (akDaemonInit()) akError(akERRDAEMON, "init" );

  akDebug(("Done akInit...\n"));
  return 0;
}


/*************************************************************************
akFinalize - Terminates the kernel
*************************************************************************/

extern int akFinalize(void) {

  akDebug(("%d:akFinalize BEGIN\n", akNode));

  if (akDaemonTerminate()) akError(akERRDAEMON, "terminate" );
  if (akMPI_Finalize()) akError(akERRMPI, "finalize" );

#ifdef AK_COUNT
  printf("node %d comm waits=%lu compl=%lu\n",
		akNode, akCountCommWait, akCountCommCompleted);
  printf("node %d comm posted send=%lu recv=%lu probe=%lu\n",
		akNode, akCountCommPostedSend,
		akCountCommPostedRecv,
		akCountCommPostedProbe);
  printf("node %d comm cmpl send=%lu recv=%lu probe=%lu\n",
		akNode, akCountCommCompletedSend,
		akCountCommCompletedRecv,
		akCountCommCompletedProbe);
#endif
  return 0;
}


/*************************************************************************
  * The Send and Receive functions do this sequence:
	- find out a wait context
	- ask MPI to communicate
	- if it's going to be waited (normally)
		initialise its semaphore
		store the handle address in the wait context,
			to the daemon put MPI status there
		store the wait context address in the handle,
			to the user wait or test on it later
	- put the w.c. in a state the daemon looks for it
*************************************************************************/


/*************************************************************************
akISend - Starts a nonblocking send
*************************************************************************/

extern int akISend(void *buf, int count, MPI_Datatype datatype, int dest,
		int tag, MPI_Comm comm, aktHandle *handlep,
		void (*callback)(), void *callback_par) {
  aktWaitContext *wcp;

  akDebug(("%d:akISend %d:%d %X\n", akNode, dest, tag, handlep));

  akpthread_mutex_lock(&akDaemonMutex);

  akDebug(("%d:akISend %d:%d %X got lock\n", akNode, dest, tag, handlep));

  /* Find (or produce) an unused wait context */
  akWCMoveToInUse1(&wcp);

  /* Wait context is for a send, mark it */
  wcp->flags |= akWCSEND;

  /* set the callback function */
  wcp->callback = callback;
  wcp->callback_par = callback_par;

  akDebug(("akISend: will MPI_Isend(%x,%d,%x,%d,%d,%x,%x)\n",
  		buf, count, datatype, dest, tag,
			comm, &(wcp->u.request)));

  /* Ask MPI to send the data. It returns a request identifier
     that's stored in the wait context. */
  if (MPI_Isend(buf, count, datatype, dest, tag,
			comm, &(wcp->u.request))) {
	akpthread_mutex_unlock(&akDaemonMutex);
	return akERRMPISEND;
  }

#ifdef AK_COUNT
  akCountCommPostedSend ++;
#endif

#ifdef A0_TRACE
  /* Added here for the trace module */
  wcp->serial_no = akCounter[akWCSEND] ++;
  /* handlep can be NULL if request isn't going to be waited on */
  if (handlep) handlep->serial_no = wcp->serial_no;
#endif

  akDebug(("akISend: after MPI_Isend()\n"));

#ifdef AK_DAENUL
    /* give only NULL handles to the daemon */
    if (handlep)
	  handlep->wc = wcp;
    else
	  akWCMoveToInUse2(wcp, handlep);
#else
    /* prepare the wc to be waited/tested on, give it to the daemon */
    akWCMoveToInUse2(wcp, handlep);
#endif

  akpthread_mutex_unlock(&akDaemonMutex);

  akDebug(("%d:akISend %d:%d %X (%X) done\n", akNode, dest, tag, handlep, wcp));

  return 0;
}


/*************************************************************************
akSend - Do a blocking send
*************************************************************************/

extern int akSend(void *buf, int count, MPI_Datatype datatype, int dest,
		int tag, MPI_Comm comm, aktHandle *handlep) {
  int r;

  r = akISend(buf, count, datatype, dest, tag, comm, handlep, 0, 0);
  if (r) return r;
  return akWait(handlep);
}


/*************************************************************************
akIRecv - Starts a nonblocking receive
*************************************************************************/

extern int akIRecv(void *buf, int count, MPI_Datatype datatype, int source,
		int tag, MPI_Comm comm, aktHandle *handlep,
		void (*callback)(), void *callback_par) {
  aktWaitContext *wcp;

  akDebug(("%d:akIRecv %d:%d %X\n", akNode, source, tag, handlep));

  akpthread_mutex_lock(&akDaemonMutex);

  akDebug(("%d:akIRecv %d:%d %X got lock\n", akNode, source, tag, handlep));

  /* Find (or produce) an unused wait context */
  akWCMoveToInUse1(&wcp);

  /* Wait context is for a receive, mark it */
  wcp->flags |= akWCRECV;

  /* set the callback function */
  wcp->callback = callback;
  wcp->callback_par = callback_par;

  akDebug(("akIRecv: will MPI_Irecv(%x,%d,%x,%d,%d,%x,%x)\n",
  		buf, count, datatype, source, tag,
			comm, &(wcp->u.request)));

  /* Ask MPI to receive the data. It returns a request identifier
     that's stored in the wait context. */
  if (MPI_Irecv(buf, count, datatype, source, tag,
			comm, &(wcp->u.request))) {
	akpthread_mutex_unlock(&akDaemonMutex);
	return akERRMPIRECV;
  }

#ifdef AK_COUNT
  akCountCommPostedRecv ++;
#endif

#ifdef A0_TRACE
  /* Added here for the trace module */;
  wcp->serial_no = akCounter[akWCRECV] ++;
  /* handlep can be NULL if request isn't going to be waited on */
  if (handlep) handlep->serial_no = wcp->serial_no;
#endif

#ifdef AK_DAENUL
    /* give only NULL handles to the daemon */
    if (handlep)
	  handlep->wc = wcp;
    else
	  akWCMoveToInUse2(wcp, handlep);
#else
    /* prepare the wc to be waited/tested on, give it to the daemon */
    akWCMoveToInUse2(wcp, handlep);
#endif

  akpthread_mutex_unlock(&akDaemonMutex);

  akDebug(("%d:akIRecv %d:%d %X (%X) done\n", akNode, source, tag, handlep, wcp));

  return 0;
}


/*************************************************************************
akRecv - Do a blocking receive
*************************************************************************/

extern int akRecv(void *buf, int count, MPI_Datatype datatype, int source,
		int tag, MPI_Comm comm, aktHandle *handlep) {
  int r;

  r = akIRecv(buf, count, datatype, source, tag, comm, handlep, 0, 0);
  if (r) return r;
  return akWait(handlep);
}


/*************************************************************************
akIProbe - Starts a non-blocking probe
*************************************************************************/

extern int akIProbe(int source, int tag, MPI_Comm comm,
			aktHandle *handlep,
		void (*callback)(), void *callback_par) {

  aktWaitContext *wcp;
  int flag; /* MPI_Iprobe's answer */

  akDebug(("akIProbe %d:%d comm=%d %X\n", source, tag, comm, handlep));

  /* Probe cannot have NULL handle */
  if (!handlep) return akERRNULLHANDLE;

  akpthread_mutex_lock(&akDaemonMutex);

#ifdef A0_TRACE
  /* Added here for the trace module */;
  handlep->serial_no = akCounter[akWCPROBE] ++;
#endif

  /* Well... Maybe just a nonblocking probe will do? */
  if (MPI_Iprobe(source, tag, comm, &flag, &handlep->status)) {
    akpthread_mutex_unlock(&akDaemonMutex);
    return akERRMPIRECV;
  }

#ifdef AK_COUNT
  akCountCommPostedProbe ++;
#endif

  if (flag) {
    /* God is great, the probe terminated */
    handlep->wc = 0;

#ifdef A0_TRACE
    akWCCallbackTrace(callback, callback_par, akWCPROBE,
			&(handlep->status), handlep->serial_no);
#else
    akWCCallback(callback, callback_par);
#endif

#ifdef AK_COUNT
    akCountCommCompletedProbe ++;
#endif

    akpthread_mutex_unlock(&akDaemonMutex);
    return 0;
  }
  /* Well... The probe didn't terminate, so now the daemon will redo it from
   * time to time, until it terminates.
   */

  /* Find (or produce) an unused wait context */
  akWCMoveToInUse1(&wcp);

  /* set the callback function */
  wcp->callback = callback;
  wcp->callback_par = callback_par;

  /* Wait context is for a probe, mark it */
  wcp->flags |= akWCPROBE;

  /* Copy the information to the wait context so the daemon knows what
     to probe */
  wcp->u.probe.source = source;
  wcp->u.probe.tag = tag;
  wcp->u.probe.comm = comm;

#ifdef A0_TRACE
  /* Added here for the trace module */;
  wcp->serial_no = handlep->serial_no;
#endif

#ifdef AK_DAENUL
  /* give only NULL handles to the daemon */
  if (handlep)
	handlep->wc = wcp;
  else
	akWCMoveToInUse2(wcp, handlep);
#else
  /* prepare the wc to be waited/tested on, give it to the daemon */
  akWCMoveToInUse2(wcp, handlep);
#endif

  akpthread_mutex_unlock(&akDaemonMutex);

  akDebug(("akIProbe %d:%d %X (%X) done\n", source, tag, handlep, wcp));

  return 0;
}


/*************************************************************************
akProbe - Do a blocking probe
*************************************************************************/

extern int akProbe(int source, int tag, MPI_Comm comm,
			aktHandle *handlep) {

  int r;

  r = akIProbe(source, tag, comm, handlep, 0, 0);
  if (r) return r;
  return akWait(handlep);
}


/*************************************************************************
akWait - Waits the completion of a nonblocking send or receive
*************************************************************************/

extern int akWait(aktHandle *handlep) {

  int i, r = 0, f;

  if (!handlep) return akERRNULLHANDLE;

  /* When the handle has already been waited on, we return (no error) */
  if (!handlep->wc) return 0;

  akDebug(("%d:akWait %X (%X)\n", akNode, handlep, handlep->wc));

  akpthread_mutex_lock(&akDaemonMutex);

  /* count blocked threads */
  akBlocked[handlep->wc->flags & akWCTYPE] ++;

#ifdef AK_BLKMPI
  /* if MPI is thread-safe, block inside of it */
  switch (handlep->wc->flags&akWCTYPE) {
  case akWCSEND:
  case akWCRECV:
#ifndef AK_DAENUL
    akWCMoveToInUse3(handlep->wc);
#endif
    akpthread_mutex_unlock(&akDaemonMutex);
    MPI_Wait(&(handlep->wc->u.request), &(handlep->status));
    akpthread_mutex_lock(&akDaemonMutex);
    break;
  case akWCPROBE:
#ifndef AK_DAENUL
    akWCMoveToInUse3(handlep->wc);
#endif
    akpthread_mutex_unlock(&akDaemonMutex);
    MPI_Probe(handlep->wc->u.probe.source, handlep->wc->u.probe.tag,
		handlep->wc->u.probe.comm, &(handlep->status));
    akpthread_mutex_lock(&akDaemonMutex);
    break;
  default:
#endif /* AK_BLKMPI */

    /* check if the request is ended. it always happens for short msgs */
    
 
    
    if (!(handlep->wc->flags & akWCDONE))
       akWCCheck(handlep->wc);

#ifdef AK_COUNT
    if (handlep->wc->flags & akWCDONE) {
      switch (handlep->wc->flags & akWCTYPE) {
      case akWCPROBE:	akCountCommCompletedProbe ++;	break;
      case akWCRECV:	akCountCommCompletedRecv ++;	break;
      case akWCSEND:	akCountCommCompletedSend ++;	break;
      }
      akCountCommCompleted ++;
    }
    akCountCommWait ++;
#endif

    /* Blocks to wait completion. The daemon unlocks the semaphore
       when it recognizes the completion. */
    while (!(handlep->wc->flags & akWCDONE))
      if(akpthread_cond_wait(&(handlep->wc->cond), &akDaemonMutex)<0)
	akError(akERRSEMAP, "p wait");
#ifdef AK_BLKMPI
  }
#endif /* AK_BLKMPI */

  /* count blocked threads */
  akBlocked[handlep->wc->flags & akWCTYPE] --;

  if (handlep->wc->flags & akCANCEL) r = akERRCANCEL;

  /* clean up and release the wait context */
  akWCMoveToFree(handlep->wc);
  akpthread_mutex_unlock(&akDaemonMutex);

  akDebug(("%d:akWait %X (%X) done\n", akNode, handlep, handlep->wc));

  /* We zero this for subsequent waits or tests */
  handlep->wc = 0;

  return r;
}


/*************************************************************************
akTest - Verifies the completion of a nonblocking send or receive
*************************************************************************/

extern int akTest(aktHandle *handlep, int *result) { 

  int r = 0;

  if (!handlep) return akERRNULLHANDLE;

  /* When the handle has already been waited on, return no error */
  if (!handlep->wc) {
    *result = TRUE;
    return 0;
  }

  akpthread_mutex_lock(&akDaemonMutex);
  
  if (!(handlep->wc->flags & akWCDONE))
       akWCCheck(handlep->wc);


  if (handlep->wc->flags & akCANCEL) r = akERRCANCEL;
  if (*result = handlep->wc->flags & akWCDONE) {
#ifdef AK_COUNT
    switch (handlep->wc->flags & akWCTYPE) {
    case akWCPROBE:	akCountCommCompletedProbe ++;	break;
    case akWCRECV:	akCountCommCompletedRecv ++;	break;
    case akWCSEND:	akCountCommCompletedSend ++;	break;
    }
    akCountCommCompleted ++;
#endif
    akWCMoveToFree(handlep->wc);
    handlep->wc = 0; /* zero for subsequent waits or tests */
  }
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


/*************************************************************************
akWaitAny - Waits the completion of several nonblocking sends or receives
*************************************************************************/

extern int akWaitAny(aktHandle *array[], int size, int *index) {
  akpthread_cond_t cond;
  int i, r, ct;

  r = akTestAny(array, size, index);

  if (*index == akNOCOMPLETED) {
    /* create cond variable to sleep waiting on it */
    if (akpthread_cond_init(&cond,akPTHREAD_CONDATTR_DEFAULT_NP))
	  akError(akERRCONDINIT, "wait-any");

    if (akpthread_mutex_lock(&akDaemonMutex))
	  akError(akERRMUTEXLOCK, "wait-any");
    /* point all wc->any to cond */
    for (i = ct = 0; i<size; i++)
      if (array[i]) {
        akDebug(("akWaitAny on %x (%d)\n", array[i], i));
#ifdef AK_DAENUL
	akWCMoveToInUse2(array[i]->wc, array[i]);
#endif
	array[i]->wc->any = &cond;
	array[i]->wc->flags |= akWCANY;
	ct ++;
      }
    if (!ct) {
      /* user passed a null table */
      if (akpthread_mutex_unlock(&akDaemonMutex))
	  akError(akERRMUTEXLOCK, "wait-any-ct");
      akDebug(("akWaitAny all done\n"));
      *index = akNOCOMPLETED;
      return 0;
    }
    while (*index == akNOCOMPLETED) {
      akpthread_cond_wait(&cond, &akDaemonMutex);
      for (i=0; i<size; i++) {
        akDebug(("akWaitAny test %x (%d)\n", array[i], i));
	if (array[i] && (array[i]->wc->flags & akWCDONE)) {
          akDebug(("akWaitAny completed %x (%d)\n", array[i], i));
	  *index = i;
	  break;
	}
      }
    }
    /* clear all wc->any */
    for (i=0; i<size; i++)
      if (array[i]) {
	array[i]->wc->flags &= ~akWCANY;
#ifdef AK_DAENUL
	akWCMoveToInUse3(array[i]->wc);
#endif
      }
    if (akpthread_mutex_unlock(&akDaemonMutex))
	  akError(akERRMUTEXLOCK, "wait-any");

    if (akpthread_cond_destroy(&cond))
	akError(akERRCONDDESTR, "wait-any");
  }

  return r;
}


/*************************************************************************
akTestAny - Verifies the completion of several nonblocking sends or receives
*************************************************************************/

extern int akTestAny(aktHandle *array[], int size, int *index) {
  int i, succ = 0, r;

  for (i=0; i<size; i++) {
    if (array[i]) {
      r = akTest(array[i], &succ);
      if (r) {
	*index = i;
	return r;
      }
      /* if one succeeded, returns the counter as index */
      if (succ) {
	*index = i;
	return 0;
      }
    }
  }

  /* no one completed */
  *index = akNOCOMPLETED;
  return 0;
}


/*************************************************************************
akBlock - Produces a fake request
*************************************************************************/

extern int akBlock(aktHandle *handlep,
		void (*callback)(), void *callback_par) {

  aktWaitContext *wcp;

  akDebug(("akBlock\n"));

  if (!handlep) return akERRNULLHANDLE;

  akpthread_mutex_lock(&akDaemonMutex);
  /* Find (or produce) an unused wait context */
  akWCMoveToInUse1(&wcp);

  /* set the callback function */
  wcp->callback = callback;
  wcp->callback_par = callback_par;

  /* Mark thye type of the request */
  wcp->flags |= akWCOTHER;

  /* remember handle address to the daemon store MPI status */
  wcp->statusp = &(handlep->status);
  /* remember w.c. address to the user wait/test on it */
  handlep->wc = wcp;

#ifdef A0_TRACE
  /* Added here for the trace module */;
  handlep->serial_no = wcp->serial_no = akCounter[akWCOTHER] ++;
#endif

  akpthread_mutex_unlock(&akDaemonMutex);

  akDebug(("akBlock: (%X,%X) done\n", wcp, handlep));

  return 0;
}


/*************************************************************************
akUnblock - Unblocks a request
*************************************************************************/

extern int akUnblock(aktHandle *handlep) {

  if (!handlep) return akERRNULLHANDLE;
  if (!handlep->wc) return akERRNULLHANDLE;

#ifdef AK_NSTOP
  akEverybodyBlocked = 0;
#endif

  akDebug(("akUnblock %x (%x)\n", handlep, handlep->wc));
  akpthread_mutex_lock(&akDaemonMutex);
  akWCUnblock(handlep->wc);
  akpthread_mutex_unlock(&akDaemonMutex);
  akDebug(("akUnblock %x (%x) DONE\n", handlep, handlep->wc));
}


/*************************************************************************
akCancel - Cancels a request
*************************************************************************/

extern int akCancel(aktHandle *handlep) {

  int type;

  if (!handlep) return akERRNULLHANDLE;
  if (!handlep->wc) return akERRNULLHANDLE;

  akDebug(("akCancel\n"));
  akpthread_mutex_lock(&akDaemonMutex);
  akWCMoveToInUse3(handlep->wc);
  akWCUnblock(handlep->wc);
  handlep->wc->flags |= akCANCEL;
  type = handlep->wc->flags & akWCTYPE;
  if ((type == akWCSEND) || (type == akWCRECV))
    if (MPI_Cancel(&(handlep->wc->u.request)))
	akError(akERRMPI, "cancel");
  akpthread_mutex_unlock(&akDaemonMutex);
}


/*************************************************************************
akNoWait - Marks the request to not be waited anymore
*************************************************************************/

extern int akNoWait(aktHandle *handlep) {

  int r = 0;

  /* This call is implemented in order to allow the higher level layer
     to submit a non-NULL request and ask it to be treated as if a NULL
     request had been posted. */

  if (!handlep) return akERRNULLHANDLE;

  /* When the handle has already been waited on, we return (no error) */
  if (!handlep->wc) return 0;

  akpthread_mutex_lock(&akDaemonMutex);
  akDebug(("akNoWait (%X:%X)\n", handlep->wc, handlep->wc->u.request));

  /* If the daemon already completed the request,
	clean up and release the wait context;
     else
	mark the w.c. as 'no wait'. The daemon will free it.
   */
  if (handlep->wc->flags & akWCDONE)
    akWCMoveToFree(handlep->wc);
  else
    akWCSetNoWait(handlep->wc);

  akpthread_mutex_unlock(&akDaemonMutex);

  akDebug(("akNoWait %X done\n", handlep->wc));

  /* We zero this for subsequent waits or tests */
  handlep->wc = 0;

  return r;
}

