/*
 *
 * akcomm.h : Communication functions
 *
 *
 * February 3rd, 1996. Ilan Ginzburg, Marcelo Pasin.
 *
 * Athapascan Kernel.
 *
 */

/* Eliminates multiple includes of this file */
#ifndef _AKCOMM_H
#define _AKCOMM_H


#include "akmpi.h"
#include "akthread.h"
#include "akwc.h"


/* Rank of the node in the computation. */
extern int akNode;
/* Number of nodes in the computation */
extern int akSize;
/* conters of blocked threads */
extern int akBlocked[];

/* index of the counters for blocked semaphores and mutexes */
#define akBLOCKSEM (akWCNTYPES)
#define akBLOCKMUT (akWCNTYPES+1)


/* akHandle must contain the status, because the status may be used after
 * completion of a receive for getting the number of elements received
 * (and the wait context for pending calls is destroyed once the call is
 * terminated).
 * When a call has completed, the status in the handle is valid and can
 * be used in subsequent calls (to our MPI interface). 
 */

typedef struct _aktHandle {
  MPI_Status	status;		/* Return status of MPI function */
  struct _aktWaitContext *wc;	/* Index into the wait context table */
#ifdef A0_TRACE
  unsigned long	serial_no;	/* serial number of the request */
#endif
} aktHandle;


/*************************************************************************
 *
 * Thread to thread communication on top of MPI + pthreads
 *
 * Comments:
 *   - port is passed to MPI as 'tag'. It is used by the receiving thread
 *     to specify what it wants to receive (and by the sending thread to
 *     specify to whom it wants to send...). port can have ANY value, but
 *     some functions are provided to allow for example coding of the id
 *     of the sending and receiving thread in port (but anything else is
 *     legal, possible, politically correct etc...).
 *     More than one thread, or more than one pending receive by one or more
 *     thread may have the same value for port. All that means is that it
 *     will be up to MPI to decide which receive to satisfy first. hehe, we
 *     got rid of our problems here. :-)
 *   - source can be a wildcard in the receive routines (wildcard
 *     passed to MPI).
 *   - We need to give precisions about the meaning of the 'completion' of a
 *     send/receive. Maybe we should give several nonblocking send and
 *     receives to match the different waiting semantics of MPI?
 *   - Only one thread is allowed to wait (using normal wait/test or the 'any'
 *     versions) on any pending request (send or receive). A simple way to
 *     warrant this is having only the thread that initiated the request do
 *     the wait, but this is not necessary. Remember: ONLY ONE THREAD WAITS
 *     FOR ANY PENDING REQUEST. got it?
 *   - Parameters buf, count, datatype, source, dest have the same meaning
 *     they have in MPI. MPI's functions can be used to construct datatypes.
 *     Parameter port has the meaning of the tag of MPI, but has been renamed
 *     because we decided it'll be that way.
 *     handle replaces MPI's request and MPI's status at the same time and
 *     more. handle is a struct composed of a pointer to a table used by the
 *     daemon and the status
 *
 *     handle looks like MPI's request but is more complicated (MPI's
 *     request is stored in a table to which handle points, or something
 *     alike, and we need synchronization (not present in the request)
 *     between the daemon and the (possibly) waiting thread).
 */


extern int akInit(int *argcp, char ***argvp);
extern int akFinalize(void);
extern int akSend(void *buf, int count, MPI_Datatype datatype, int dest,
		int tag, MPI_Comm comm, aktHandle *handlep);
extern int akRecv(void *buf, int count, MPI_Datatype datatype, int source,
		int tag, MPI_Comm comm, aktHandle *handlep);
extern int akProbe(int source,
		int tag, MPI_Comm comm, aktHandle *handlep);
extern int akISend(void *buf, int count, MPI_Datatype datatype, int dest,
		int tag, MPI_Comm comm, aktHandle *handlep,
		void (*callback)(), void *callback_par);
extern int akIRecv(void *buf, int count, MPI_Datatype datatype, int source,
		int tag, MPI_Comm comm, aktHandle *handlep,
		void (*callback)(), void *callback_par);
extern int akIProbe(int source,
		int tag, MPI_Comm comm, aktHandle *handlep,
		void (*callback)(), void *callback_par);

extern int akWait(aktHandle *handlep);
extern int akTest(aktHandle *handlep, int *result);
extern int akWaitAny(aktHandle *array[], int size, int *index);
extern int akTestAny(aktHandle *array[], int size, int *index);

extern int akBlock(aktHandle *handlep,
		void (*callback)(), void *callback_par);
extern int akUnblock(aktHandle *handlep);

extern int akCancel(aktHandle *handlep);

/* if no handles are completed, akTestAny returns akNOCOMPLETED */
#define akNOCOMPLETED (-1)


#ifdef AKERNEL_SOURCE
/****************************************************************************
		Included only for compiling the kernel
****************************************************************************/


#endif /* AKERNEL_SOURCE */

#endif /* _AKCOMM_H */
