/*
 *
 * akdaemon.c : Kernel daemon
 *
 *
 * by Ilan Ginzburg, Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * Feb 5, 1996
 *
 */


/* Kernel source file */
#define	AKERNEL_SOURCE
#include "akthread.h"
#include "akconfig.h"
#include "akmpi.h"
#include "akdaemon.h"
#include "akcomm.h"
#include "akwc.h"
#include "akerror.h"
#include "akinfo.h"
/* debugging function */
#define akDebug_on(a) printf a
#define akDebug(a)  


/*************************************************************************
	Global variables
*************************************************************************/

#ifdef AK_WAKER
static akpthread_t akWakerUpThread;	/* waker-up thread id */
#endif

#ifdef AK_DAEMON
/* reasons why daemon goest to sleep */
#define akDAEMONNOTHING 0
#define akDAEMONWRAPPED 1
#define akDAEMONCOMPLETED 2

static akpthread_t akDaemonThread;	/* Daemon thread id */
static akpthread_attr_t akDaemonAttr;
static volatile int akDaemonRun;	/* daemon termination flag */
akpthread_cond_t  akDaemonSleepCond;	/* daemon sleeps waiting on it */
#endif

#ifdef AK_TIMED
/* period that the daemon sleeps */
static volatile struct timespec akDaemonSleepDelta;
#endif

#ifdef AK_NSTOP
/* set by the low-priority thread to tell the daemon to run non-stop */
int volatile akEverybodyBlocked = 0;
#endif

/*#define AK_COUNT*/
#ifdef AK_COUNT
static unsigned long akCountDaemonSleep = 0;
static unsigned long akCountDaemonTimedSleep = 0;
static unsigned long akCountDaemonYield = 0;
static unsigned long akCountDaemonAdvance = 0;
static unsigned long akCountDaemonTestSend = 0;
static unsigned long akCountDaemonTestRecv = 0;
static unsigned long akCountDaemonTestProbe = 0;
static unsigned long akCountDaemonChecked = 0;
static unsigned long akCountDaemonCompletedSend = 0;
static unsigned long akCountDaemonCompletedRecv = 0;
static unsigned long akCountDaemonCompletedProbe = 0;
static unsigned long akCountDaemonCompleted = 0;
static unsigned long akCountDaemonWrapped = 0;
static unsigned long akCountDaemonNothing = 0;
#endif

/* the NEXT w.c. to be processed in the next loop of the daemon */
aktWaitContext *volatile akDaemonNext;


/*************************************************************************
akAdvance - Advances the communication daemon by waking daemon up
*************************************************************************/

extern int akAdvance(void) {

    /* Wake up the sleeping daemon so it can work a bit */
#ifndef AK_ASLEEP
  if (akWCInUse2)
#endif
  {
#ifdef AK_COUNT
	akCountDaemonAdvance ++;
#endif
    	if (akpthread_cond_signal(&akDaemonSleepCond))
		akError(akERRCONDSIG, "advance");
	aksched_yield();
  }
}


#ifdef AK_WAKER
/*************************************************************************
akWakerUp - Wakes up the daemon whenever possible
*************************************************************************/

static void *akWakerUp(void *arg) {

/* This function is started as a separate thread and its role is to wake
 * up the daemon when there's no work to do for advancing communications.
 */
  akDebug(("akWakerUp started\n"));

  while (akDaemonRun) {

#ifdef AK_NSTOP
    /* the variable akEverybodyBlocked is set to 1 here. It means that
       all other threads are blocked now. It is true, because this
       thread executes in low priority and no timer blocking exist. It 
       only works with priority scheduling and if no timers are used.
     */
    akEverybodyBlocked = 1; /* tell the daemon to run non-stop */
#endif

    if (akDaemonRun) akAdvance();
  }
}
#endif


/*************************************************************************
akDaemonSleep - The daemon goes to sleep
*************************************************************************/
#ifdef __GNUC__
inline
#endif
static int akDaemonSleep(int reason, aktWaitContext **currp) {

  /* This function is used by the daemon to unlock the mutex and give the

     other threads the access to MPI and the lists. If the reason is
     akDAEMONWRAPPED, the daemon has run enough and it wants to sleep for a
     while. If it is akDAEMONNOTHING, the daemon has no work to do and must be
     awaken if some w.c. is inserted in its list (InUse2). If it is
     akDAEMONCOMPLETED, a w.c. is just completed and the daemon wants to run
     the thread that it deblocks. */

#ifdef AK_TIMED
  struct timespec wakeup; /* time to wake-up */
#endif

#ifdef AK_NSTOP
  akEverybodyBlocked = 0;
#endif

  switch (reason) {
  case akDAEMONCOMPLETED:
  case akDAEMONWRAPPED:

    akDebug(("akDaemon goes to sleep(completed/wrapped).\n"));

#ifdef AK_ASLEEP
#ifdef AK_COUNT
    akCountDaemonSleep ++;
#endif
    if (akpthread_cond_wait(&akDaemonSleepCond, &akDaemonMutex))
	akError(akERRCONDWAIT, "sleep daemon");
#else /* AK_ASLEEP */
#ifdef AK_TIMED
    /* calculate wake-up time */
    /* this function adds akDaemonSleepDelta to the current time */
    if (akpthread_get_delta_np(&akDaemonSleepDelta, &wakeup))
		akError(akERRDAEMONTIMER, "get");

#ifdef AK_COUNT
    akCountDaemonTimedSleep ++;
#endif
    /* wait until time reaches wakeup */
    /* to test the error here is too awkward. */
    akpthread_cond_timedwait(&akDaemonSleepCond,
				 &akDaemonMutex, &wakeup);
#else /* !AK_TIMED */
    akpthread_mutex_unlock(&akDaemonMutex);
#ifndef AK_NOYLD
#ifdef AK_COUNT
    akCountDaemonYield ++;
#endif
    aksched_yield();
#endif /* AK_NOYLD */
    akpthread_mutex_lock(&akDaemonMutex);
#endif /* AK_TIMED */
#endif /* AK_ASLEEP */

    *currp = akDaemonNext ? akDaemonNext : akWCInUse2;
    break;

  case akDAEMONNOTHING:

    akDebug(("akDaemon goes to sleep(nothing).\n"));
    akDaemonNext = 0;

#ifdef AK_COUNT
    akCountDaemonSleep ++;
#endif
    if (akpthread_cond_wait(&akDaemonSleepCond, &akDaemonMutex))
	akError(akERRCONDWAIT, "sleep daemon");
    *currp = akWCInUse2;
  }
  akDebug(("akDaemon wakes up.\n"));
}


/*************************************************************************
akDaemon - The daemon thread itself.
*************************************************************************/

static void *akDaemon(void *arg) {

  aktWaitContext *curr; /* the w.c. processed by the daemon in this turn */
  int count;		/* count processed w.c., stop when all processed */

#ifdef A0_TRACE
  {
    unsigned long id;

    /* find a trace id and set it */
    akTraceNewThreadId(&id);
    akTraceSetThreadId(id);
  }
#endif

  akDebug(("akDaemon started\n"));

  /* Daemon has the mutex locked always, only unlock it when sleep. */
  if (akpthread_mutex_lock(&akDaemonMutex))
	akError(akERRMUTEXLOCK, "1 daemon");
  curr = akWCInUse2; /* Must start somewhere... */
  count = 0;

  while (akDaemonRun) {
    akDebug(("%d:Start of while (akDaemonRun)\n", akNode));

    /* If there's nothing to do, go to sleep */
    if (!curr) {
#ifdef AK_COUNT
      akCountDaemonNothing ++;
#endif
      akDaemonSleep(akDAEMONNOTHING, &curr);
      count = 0;
      continue;
    }

#ifdef AK_COUNT
    switch (curr->flags & akWCTYPE) {
    case akWCPROBE:	akCountDaemonTestProbe ++;	break;
    case akWCRECV:	akCountDaemonTestRecv ++;	break;
    case akWCSEND:	akCountDaemonTestSend ++;	break;
    }
#endif

    akDebug(("checking %X (flags=%X) ct=%d \n", curr, curr->flags, count));
    akWCCheck(curr);
    count ++;

    if (curr->flags & akWCDONE) {
#ifdef AK_COUNT
      switch (curr->flags & akWCTYPE) {
      case akWCPROBE:	akCountDaemonCompletedProbe ++;	break;
      case akWCRECV:	akCountDaemonCompletedRecv ++;	break;
      case akWCSEND:	akCountDaemonCompletedSend ++;	break;
      }
      akCountDaemonCompleted ++;
#endif
      akDaemonSleep(akDAEMONCOMPLETED, &curr);
      count = 0;
      continue;
    }

#ifdef AK_NSTOP
    /* Daemon does not stop after a turn, if everybody is blocked. */
    if (akEverybodyBlocked) {
      curr = curr->next ? curr->next : akWCInUse2;
      continue;
    }
#endif

    if (count >= akWCInUse2Count) {  /* completed a turn */
#ifdef AK_COUNT
      akCountDaemonChecked += count;
      akCountDaemonWrapped ++;
#endif
      akDaemonSleep(akDAEMONWRAPPED, &curr);
      count = 0;
      continue;
    }

    /* no reason to stop, take next w.c. */
    curr = curr->next ? curr->next : akWCInUse2;
  }

#ifdef AK_CANCEL
  /* Cancel all the MPI requests, so it doesn't blame the ``abrupt''
     termination. Ideally, only the final user requests should be
     cancelled. Here, all daemons AND user requests will be cancelled.
   */

  curr = akWCInUse2;
  while (curr) {
    if (((curr->flags & akWCTYPE) == akWCSEND) ||
	((curr->flags & akWCTYPE) == akWCRECV))
      if (MPI_Cancel(&(curr->u.request)))
		akError(akERRMPI, "cancel daemon");
    curr = curr->next;
  }
#endif

  akpthread_mutex_unlock(&akDaemonMutex);

  akInfo(DAEMON_TYPE);
  akDebug(("%d:akDaemon: I'm deeeaaaaaaaaaaad!\n", akNode));
}


/*************************************************************************
	akDaemonInit - akernel daemon startup
*************************************************************************/

extern int akDaemonInit(void) {

/* Initializes the table of wait contexts, the lists, the mutexes
 * and then launches the daemon thread.
 * Must be called while still single threaded.
 */

  int i;
  akpthread_attr_t attr;
  struct aksched_param params;
  int sched;

  akDebug(("akDaemonInit BEGIN\n"));

  /* set up our scheduling & prio */
  /* This function returns here a non-null value but still seems to
   * operate properly.
   */
  params.aksched_priority = AK_USER_PRIO;
  akpthread_setschedparam(akpthread_self(), AK_USER_SCHED, &params);
  akpthread_getschedparam(akpthread_self(), &sched, &params);

  akDebug(("Athapascan Kernel started: priority=%d, sched=%d.\n",
	     params.aksched_priority, sched));

  /* set up the daemon environment */
  /* akDaemonRun is set to 0 when the daemon must stop */
  akDaemonRun = 1;

  /* create cond variable to sleep waiting on it */
   if (akpthread_cond_init(&akDaemonSleepCond,
			   akPTHREAD_CONDATTR_DEFAULT_NP))
     akError(akERRCONDINIT, "daemon-sleep");

#ifdef AK_TIMED
  /* period that the daemon sleeps if no requests are completed */
  akDaemonSleepDelta.tv_sec = 0;				/* seconds */
  akDaemonSleepDelta.tv_nsec = AK_DAEMON_SLEEPDELTA;		/* nanoseconds */
#endif /* AK_TIMED */

#ifdef A0_TRACE
  /* init thread id tracing */
  akTraceInit();
#endif

#ifdef AK_DAEMON

  if(akpthread_attr_init(&attr))
	akError(akERRDAEMON, "daemon attr");

#if defined(A0_TSOL2) || defined(A0_TLIN2) || \
    defined(A0_TSGI6) ||defined(A0_TAIX4)
if(akpthread_attr_setscope(&attr, akPTHREAD_SCOPE_SYSTEM))
        akError(akERRDAEMON, "daemon scope");
#endif 

  /* set daemon scheduling & prio */
  if(akpthread_attr_setinheritsched(&attr, akPTHREAD_EXPLICIT_SCHED))
    akError(akERRDAEMON, "daemon inherit");
  if(akpthread_attr_setschedpolicy(&attr, AK_DAEMON_SCHED))
    akError(akERRDAEMON, "daemon policy");
  params.aksched_priority = AK_DAEMON_PRIO;


#ifndef A0_TLIN2
  if(akpthread_attr_setschedparam(&attr, &params))
    akError(akERRDAEMON, "deamon prio");
#endif


#ifdef A0_TAIX4
  if(akpthread_attr_setdetachstate(&attr, akPTHREAD_CREATE_JOINABLE))
	akError(akERRDAEMON, "deamon detach");
#endif

  akDebug(("akDaemonInit: will start daemon\n"));

  /* Start the daemon thread */
  

  if(akpthread_create(&akDaemonThread, &attr,
			akDaemon, NULL))
	akError(akERRDAEMON, "start daemon");

  akDebug(("akDaemonInit: daemon started!\n"));

  if(akpthread_attr_destroy(&attr))
	akError(akERRDAEMON, "daemon destr");

#endif /* AK_DAEMON */

#ifdef AK_WAKER
  /* Starting the low priority thread that wakes up the daemon when there's
     nothing else to do. */

  if(akpthread_attr_init(&attr))
	akError(akERRDAEMON, "waker attr");

  /* set daemon scheduling & prio */
  if(akpthread_attr_setinheritsched(&attr, akPTHREAD_EXPLICIT_SCHED))
	akError(akERRDAEMON, "waker inherit");
  if(akpthread_attr_setschedpolicy(&attr, AK_WAKER_SCHED))
	akError(akERRDAEMON, "waker policy");
  params.aksched_priority = AK_WAKER_PRIO;
  if(akpthread_attr_setschedparam(&attr, &params))
    akError(akERRDAEMON, "deamon prio");

#ifdef A0_TAIX4
  if(akpthread_attr_setdetachstate(&attr, akPTHREAD_CREATE_JOINABLE))
	akError(akERRDAEMON, "waker detach");
#endif

  /* Start the waker thread */
  if(akpthread_create(&akWakerUpThread, &attr,
			akWakerUp, NULL))
	akError(akERRDAEMON, "start waker");

  if(akpthread_attr_destroy(&attr))
	akError(akERRDAEMON, "waker destr");
#endif

  akDebug(("akDaemonInit END\n"));

  return 0;
}


/*************************************************************************
	akDaemonTerminate - daemon termination
*************************************************************************/

extern int akDaemonTerminate(void) {

  void *status;

  akDebug(("%d:akDaemonTerminate BEGIN\n", akNode));
  
  /* tell the demon to stop in the next loop */
  akDaemonRun = 0;

  /* just in case the daemon is sleeping */
  akpthread_cond_signal(&akDaemonSleepCond);

  akDebug(("%d:akDaemonTerminate after signal\n", akNode));

#ifdef AK_DAEMON
  /* wait the daemon thread end */
  if (akpthread_join(akDaemonThread, &status))
	akError(akERRDAEMON, "daemon join");

  akDebug(("%d:akDaemonTerminate after join daemon\n", akNode));
#endif

#ifdef AK_WAKER
  /* wait the waker-up thread end */
  if (akpthread_join(akWakerUpThread, &status))
	akError(akERRDAEMON, "waker join");

  akDebug(("%d:akDaemonTerminate after join waker\n", akNode));
#endif

#ifdef AK_DBGTIME
  { int i;
  /* Next lines for debug only! */
    for (i=0; i<akDebugIndex; i++)
      printf("[%d]: %ds %dns, since [0]:%ds %dns\n", i,
		akDebugTimed[i][1].tv_sec-akDebugTimed[i][0].tv_sec,
		akDebugTimed[i][1].tv_nsec-akDebugTimed[i][0].tv_nsec,
		akDebugTimed[i][1].tv_sec-akDebugTimed[0][0].tv_sec,
		akDebugTimed[i][1].tv_nsec-akDebugTimed[0][0].tv_nsec);
  }
#endif /* AK_DBGTIME */

  akDebug(("%d:akDaemonTerminate END\n", akNode));

#ifdef AK_COUNT
  printf("node %d daemon completed=%lu wrapped=%lu nothing=%lu chk/wrap=%f\n",
		akNode, akCountDaemonCompleted, akCountDaemonWrapped,
		akCountDaemonNothing,
		(float) akCountDaemonChecked /
			(float) akCountDaemonWrapped);
  printf("node %d daemon sleep=%lu yield=%lu timed=%lu advance=%lu\n",
		akNode, akCountDaemonSleep, akCountDaemonYield,
		akCountDaemonTimedSleep, akCountDaemonAdvance);
  printf("node %d daemon test send=%lu recv=%lu probe=%lu\n",
		akNode, akCountDaemonTestSend, akCountDaemonTestRecv,
		akCountDaemonTestProbe);
  printf("node %d daemon cmpl send=%lu recv=%lu probe=%lu\n",
		akNode, akCountDaemonCompletedSend, akCountDaemonCompletedRecv,
		akCountDaemonCompletedProbe);
#endif

  return 0;
}
