/*
 *
 *  aksemap.c
 *
 *  Semaphores on top of pthread mutexes and condition variables.
 *
 *
 *  Ilan Ginzburg.  January 29th, 1996.
 *  Marcelo Pasin.  Rewritten Mar 15, 1996.
 *
 */

#include <pthread.h>
#include <stdlib.h>

#define AKERNEL_SOURCE
#include "akthread.h"
#include "aksemap.h"
#include "akerror.h"
#include "akcomm.h"

/* debugging function */
#define akDebugOff(a) printf a
#define akDebug(a)

/*
 * in the struct semaphore:
 *   mutex: mutex associated with the semaphore
 *   cond: condition variable associated with the semaphore.
 *   counter: counter of the semaphore. a P on the semaphore is
 *      possible if counter >0. When this entry in the semaphore table
 *      is unused (free), counter contains the number of the next free
 *      entry. (or NO_NEXT_FREE if no next free entry).
 */


/************************************************************************
 * akSemaphoreCreate creates a new semaphore in sem.
 * init_value is the initial value of the counter of the semaphore
 * Return akERROK if OK, and a negative code if errors.
 * Non fatal return codes (program left in coherent state):
 *   akERRNOMEM: no more semaphores 
 ************************************************************************/

extern int akSemaphoreCreate(aktSemaphore *sem, int init_value) {

  /* Create the mutex for the semaphore */
  if (akpthread_mutex_init(&sem->mutex, akPTHREAD_MUTEXATTR_DEFAULT_NP)<0)
	return akERRMUTEXINIT;

  /* Create the condition variable for the semaphore */
  if (akpthread_cond_init(&sem->cond, akPTHREAD_CONDATTR_DEFAULT_NP)<0)
	return akERRCONDINIT;

  /* Update the counter and condition variable of the semaphore */
  sem->counter = init_value;

  return akERROK; /* No errors */
}


/************************************************************************
 * akSemaphoreDestroy deletes a semaphore
 * sem is the semaphore to delete (previously returned by a call to
 *     akSemaphoreCreate)
 * Return akERROK if OK, and a negative code if errors.
 * Non fatal return codes (program left in coherent state):
 *   akERRINVALID: invalid value for sem
 *   akERREMPTY: semaphore is free (not in use)
 *   akERROBSOL: deleted semaphore is obsolete (wrong version number)
 ************************************************************************/

extern int akSemaphoreDestroy(aktSemaphore *sem) {

  /* delete the semaphore's mutex */
  if (akpthread_mutex_destroy(&sem->mutex)<0)
	return akERRMUTEXDESTR;
  /* delete the semaphore's condition variable */
  if (akpthread_cond_destroy(&sem->cond)<0)
	return akERRCONDDESTR;

   return akERROK;
}


/************************************************************************
 * akSemaphoreP make a P (get?) on a semaphore if it's possible (the counter
 *    is positive non zero), or block and wait until it's possible. Decrement
 *    the counter.
 * sem: semaphore on which to do the P.
 * Return 0 if OK, and a negative code if errors.
 * Non fatal return codes (program left in coherent state):
 *   akERRINVALID: invalid value for sem
 *   akERREMPTY: semaphore is free (not in use)
 *   akERROBSOL: semaphore is obsolete (wrong version number)
 ************************************************************************/

extern int akSemaphoreP(aktSemaphore *sem) {

  /* Get the mutex */
  if (akpthread_mutex_lock(&sem->mutex)<0)
	return akERRMUTEXLOCK;

  akBlocked[akBLOCKSEM] ++;

  /* Wait on the condition which allows to decrement the counter */
  while (sem->counter <= 0)
    if(akpthread_cond_wait(&sem->cond, &sem->mutex)<0)
	return akERRCONDWAIT; /* erroneous return from akpthread_cond_wait */

  akBlocked[akBLOCKSEM] --;

  /* We got through. Update the value of the counter */
  sem->counter--;

  /* release the mutex of the semaphore */
  if (akpthread_mutex_unlock(&sem->mutex)<0)
	return akERRMUTEXUNLOCK;

   /* Everything went fine */
   return akERROK;
}


/************************************************************************
 * akSemaphoreTryP try to make a P (get?) on a semaphore. Do not bloc. If the
 *    P has been done, succeeded returns TRUE. If not, succeeded returns FALSE.
 *    In case the P has been done, the counter of the semaphore is decremented.
 * sem: semaphore on which to do the P.
 * succeeded: value returned (TRUE if P done, FALSE if not done)
 * Return 0 if OK, and a negative code if errors.
 * Non fatal return codes (program left in coherent state):
 *   akERRINVALID: invalid value for sem
 *   akERREMPTY: semaphore is free (not in use)
 *   akERROBSOL: semaphore is obsolete (wrong version number)
 ************************************************************************/

extern int akSemaphoreTryP(aktSemaphore *sem, int *succeeded) {

  /* Get the mutex */
  if (akpthread_mutex_lock(&sem->mutex)<0)
	return akERRMUTEXLOCK;

  /* Check the counter to see if a P is possible */
  if (sem->counter > 0) {
    /* The P is possible. We do it. */
    *succeeded = 1; /* for returning */
    sem->counter --;
  }
  else *succeeded = 0; /* the P couldn't be done */

  /* release the mutex of the semaphore */
  if (akpthread_mutex_unlock(&sem->mutex)<0)
	return akERRMUTEXUNLOCK;

  /* Everything went fine */
  return akERROK;
}


/************************************************************************
 * akSemaphoreV make a V (release?) on a semaphore. Increment the counter.
 *    potentially releases one of the threads blocked on that semaphore.
 *    (potentially, because if the semaphore was created with a negative
 *    initial value, a certain number of 'V' are needed before the
 *    first 'P' can be satisfied...).
 * sem: semaphore on which to do the V.
 * Return akERROK if OK, and a negative code if errors.
 * Non fatal return codes (program left in coherent state):
 *   akERRINVALID: invalid value for sem
 *   akERREMPTY: semaphore is free (not in use)
 *   akERROBSOL: semaphore is obsolete (wrong version number)
 ************************************************************************/

extern int akSemaphoreV(aktSemaphore *sem) {

  /* Get the mutex */
  if (akpthread_mutex_lock(&sem->mutex)<0)
	return akERRMUTEXLOCK;

  /* Increment the counter */
  sem->counter ++;
  /* Signal the condition variable (no broadcast needed here because at most
   * one blocked thread can be released). */
  if(akpthread_cond_signal(&sem->cond)<0)
	return akERRCONDSIG;

  /* release the mutex of the semaphore */
  if (akpthread_mutex_unlock(&sem->mutex)<0)
	return akERRMUTEXUNLOCK;

   /* Everything went fine */
   return akERROK;
}
