/*
 *
 * function.c : a0format Function Format Demo
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * Mar 28, 1997
 *
 * This program is a demostration on how to use function formats.
 * It builds a function format that allows users to send and receive
 * data structured in lists.
 */

#include <a0format.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>


/**********************typedefs***************************************/

/* data of an element in the list */
typedef struct {
  int foo;
  int bar;
  int node;			/* node that generated this element */
} elem_t;

/* element in the list */
typedef struct _list_t {
  struct _list_t *next;		/* next element of the list */
  elem_t          data;		/* data of this element */
} list_t;


/**********************global variables*******************************/

/* the format of an element of a list */
a0tFormat elem_format;

/* the format of a list */
a0tFormat list_format;


/**********************************************************************
	pack_list - a function that packs a list
**********************************************************************/

a0tError pack_list(a0tBuffer *bufferp,
		a0tFormat *formatp, void *address, int count) {

  /* The prototype of this function must be the same as a0FPack().
     It packs a list of elements, starting at a given address.
   */

  list_t *elemp;
  int size;
  a0tError ret;

  /* User pass count == 0 when all the list must be packed. 
   * Note that the behavior is not identical to a0Pack(), since
   * "count" and address mean whatever I want.
   */
  if (count == 0) {
    size = 0;
    elemp = (list_t *) address;
    while (elemp) {
      size ++;
      elemp = elemp->next;
    }
  }
  else size = count;

  /* pack the size of the list */
  ret = a0FPack(bufferp, A0FInt, &size, 1);
  if (ret) return ret;

  elemp = (list_t *) address;
  while (size) {
    ret = a0FPack(bufferp, &elem_format, &(elemp->data), 1);
    if (ret) return ret;
    elemp = elemp->next;
    size --;
  }

  return A0ErrOk;
}


/**********************************************************************
	unpack_list - a function that unpacks a list
**********************************************************************/

a0tError unpack_list(a0tBuffer *bufferp,
		a0tFormat *formatp, void *address, int count) {

  /* The prototype of this function must be the same as a0FUnpack().
     It unpacks a list of elements, starting at a given address.
   */

  list_t *elemp;
  int size, i;
  a0tError ret;

  /* unpack the size of the list */
  ret = a0FUnpack(bufferp, A0FInt, &size, 1);
  if (ret) return ret;

  /* User pass count == 0 when all the list must be unpacked.
   * Here, "address" is considered as a (list_t **). A list is
   * allocated and its address is returned in "address". If
   * "count" is other than zero, the function considers that
   * the list is already allocated and the elements are received
   * in it.
   */
  if (count == 0) {
    /* Allocate a list */
    elemp = (list_t *) malloc(size * sizeof(list_t));
    if (!elemp) return A0ErrNoMem;

    /* store the allocated list in "address" */
    *((list_t **)address) = elemp;

    /* Make the list chain, with NULL ate end. */
    for (i=0; i<(size-1); i++) elemp[i].next = elemp + i + 1;
    elemp[i].next = NULL;
  }
  else {
    /* "count" is size and store list starting at "address" */
    size = count;
    elemp = (list_t *) address;
  }

  /* unpack all elements */
  while (size) {
    ret = a0FUnpack(bufferp, &elem_format, &(elemp->data), 1);
    if (ret) return ret;
    elemp = elemp->next;
    size --;
  }

  return A0ErrOk;
}


/**********************************************************************
get_list_size - a function that gives the size of a packed list
**********************************************************************/

a0tError get_list_size(a0tFormat *formatp, void *address, int count,
			unsigned long *ret_size) {

  /* The prototype of this function must be the same as a0GetFormatSize().
   * It computes the size it takes to pack a list of elements, starting at
   * a given address.
   */

  list_t *elemp;
  int size;
  a0tError ret;
  unsigned long size_bytes, list_bytes;

  /* User pass count == 0 when all the list must be packed. 
   * Note that the behavior is not identical to a0GetFormatSize(),
   * since "count" and "address" mean whatever I want.
   */
  if (count == 0) {
    size = 0;
    elemp = (list_t *) address;
    while (elemp) {
      size ++;
      elemp = elemp->next;
    }
  }
  else size = count;

  /* get the size it takes for an int */
  ret = a0GetFormatSize(A0FInt, &size, 1, &size_bytes);
  if (ret) return ret;

  elemp = (list_t *) address;
  /* get the size it takes for "size" elements */
  ret = a0GetFormatSize(&elem_format, &(elemp->data), 1, &list_bytes);
  if (ret) return ret;

  *ret_size = size_bytes + size * list_bytes;
  return A0ErrOk;
}


/**********************************************************************
get_list_offset - a function that gives the address of the Nth element
**********************************************************************/

a0tError get_list_offset(a0tFormat *formatp, void *address, int count,
			void **offsetp) {

  /* The prototype of this function must be the same as a0GetFormatOffset().
   * It computes the the address of an element inside formatted data. The
   * address of the wanted element is found by skipping count elements of
   * format "formatp", from the starting address "address".
   */

  list_t *elemp;

  elemp = (list_t *) address;
  while (elemp && count--) elemp = elemp->next;

  (*offsetp) = (void *) elemp;
  return A0ErrOk;
}


/**********************************************************************
	print_list - print contents of a list
**********************************************************************/

void print_list(list_t *list) {

  while (list) {
    printf("[%d,%d] (%d) %x\n",
		list->data.foo, list->data.bar, list->data.node,
		list->next);
    list = list->next;
  }
}


/**********************************************************************
	concatenate_lists - glues two lists together
**********************************************************************/

list_t *concatenate_lists(list_t *list1, list_t *list2) {

  list_t *curr;

  if (!list1) return list2;
  if (!list2) return list1;

  curr = list1;
  while (curr->next) curr = curr->next;

  curr->next = list2;
  return list1;
}


/**********************************************************************
	free_list - frees memory used by a list
**********************************************************************/

void free_list(list_t *list) {

  list_t *next;

  while (list) {
    next = list->next;
    free(list);
    list = next;
  }
}


/**********************************************************************
	generate_list - generate a list
**********************************************************************/

list_t *generate_list(int count) {

  /* This function generates a list of "count" elements with dummy
   * values.
   */

  list_t *new, *before = NULL, *first = NULL;

  while (count) {
    new = (list_t *) malloc(sizeof(list_t));

    if (!first) first = new;
    if (before) before->next = new;
    if (!new) {
      free_list(first);
      return NULL;
    }

    new->data.foo = -count;
    new->data.bar = 1000 + (rand() % 1000);
    new->data.node = a0SelfNode;
    before = new;
    count --;
  }
  new->next = NULL;
  return first;
}


/**********************************************************************
	main
**********************************************************************/

void main(int argc, char *argv[]) {

  a0tPort my_port;
  a0tBuffer my_buffer;


  a0Init(&argc, &argv);
  a0UseFormats();
  a0NewPort(&my_port, 10, A0GlobalCreation);
  a0NewBuffer(&my_buffer, A0SendBufferType, 100000);

  {
    elem_t elem;		/* allocate to describe fields */
    a0tMapField field[3];	/* three fields */

    /* first field: integer "foo" */
    a0SetMapField(field+0, &elem, A0FInt, &elem.foo, 1);
    /* second field: integer "bar" */
    a0SetMapField(field+1, &elem, A0FInt, &elem.bar, 1);
    /* third field: integer "node" */
    a0SetMapField(field+2, &elem, A0FInt, &elem.node, 1);

    /* map has sizeof(elem_t) bytes, 3 fields */
    a0NewMapFormat(&elem_format, sizeof(elem_t), 3, field);

  } /* free elem & field */

  a0NewFunctionFormat(&list_format, pack_list, unpack_list,
			get_list_size, get_list_offset);

  a0InitCommit();

  sleep(60);

  /* at least 2 nodes */
  if (a0NodeCount < 2) a0Abort(-1);


  /******************************************************************
   * First part: node 0 creates a list with 5 elements and sends it *
   * to node 1.                                                     *
   ******************************************************************/

  /*************** node 0 code ***************/
  if (a0SelfNode==0) {

    list_t *list;

    printf("node %d will send list:\n", a0SelfNode);
    list = generate_list(5);	/* make a list with 5 elems. */
/*    print_list(list);*/

    a0FSend(/*port*/ &my_port, /*node*/ 1, /*tag*/ 0, /*request*/ NULL,
	/*format*/ &list_format, /*address*/ list, /*count*/ 0);
    free_list(list);
  }

  /*************** node 1 code ***************/
  if (a0SelfNode==1) {

    list_t *list;

    a0FReceive(/*port*/ &my_port, /*node*/ 0, /*tag*/ 0, /*request*/ NULL,
	/*format*/ &list_format, /*address*/ &list, /*count*/ 0);

    printf("node %d received list:\n", a0SelfNode);
    print_list(list);
    free_list(list);
  }


  /******************************************************************
   * Second part: node 1..n create a lists with 5 elements and      *
   * send it to node 0. Node 0 concatenated all lists received.     *
   ******************************************************************/


  /*************** node 0 code ***************/
  if (a0SelfNode==0) {

    list_t *list = NULL, *curr = NULL;
    int n;

    for (n=1; n<a0NodeCount; n++) {
      a0FReceive(/*port*/ &my_port, /*node*/ A0AnySource, /*tag*/ 1,
		/*request*/ NULL, /*format*/ &list_format,
		/*address*/ &curr, /*count*/ 0);

      list = concatenate_lists(list, curr);
    }

    printf("node %d received all lists:\n", a0SelfNode);
    print_list(list);
    free_list(list);
  }

  /*************** node 1..n code ***************/
  if (a0SelfNode>0) {

    list_t *list;

    printf("node %d will send list:\n", a0SelfNode);
    list = generate_list(5);	/* make a list with 5 elems. */
/*    print_list(list);*/

    a0FSend(/*port*/ &my_port, /*node*/ 0, /*tag*/ 1, /*request*/ NULL,
	/*format*/ &list_format, /*address*/ list, /*count*/ 5);

    free_list(list);
  }

  a0Terminate();
}
