
/*
 *
 * akwc.h : Daemon Kernel wait context operations
 *
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * Mar 13, 1997
 *
 */

/* Eliminates multiple includes of this file */
#ifndef _AKWC_H
#define _AKWC_H


#ifdef AKERNEL_SOURCE
/****************************************************************************
		Included only for compiling the kernel
****************************************************************************/

#include "akthread.h"
#include "akmpi.h"


/****************************************************************************
		Wait context definitions
****************************************************************************/

/* Nombre maximal de wait contexts */
#define akMAXWC	1024

/* A wait context can be for waiting for:
 * - a send or a receive to complete
 * - a probe to complete
 * The type of event waited on is known by looking at one bit
 * (akWCSENDRECV) of flags. if TRUE: send or receive, if FALSE: probe.
 */
typedef struct _aktWaitContext {
  int flags;		/* See above & below */
  akpthread_cond_t cond;	/* condition to sleep in akWait */
  akpthread_cond_t *any;	/* condition to sleep in akWaitAny */
  struct _aktWaitContext *next; /* Link to next wait context or NULL. */
  struct _aktWaitContext *prev; /* Link to prev wait context. See below */
  MPI_Status	*statusp; /* Place to store status */
  void	(*callback)();	/* function called at completion */
  void	*callback_par;	/* parameter of callback function */
  union {		/* Following fields only used in akSend and akRecv */
    MPI_Request request; /* request associated with non blocking MPI call */
    struct {		/* Following fields only used in akProbe */
      int	source;	/* source to probe on */
      int	tag;	/* tag to probe on */
      MPI_Comm	comm;	/* Communicator to probe on */
    } probe;
  } u;
#ifdef A0_TRACE
  unsigned long serial_no; /* serial number of the request */
#endif
} aktWaitContext;

/* Defines for aktWaitContext flags (bitwise AND):

akWCTYPE: This bitmask gives the type of the wait context. It can be:
	akWCOTHER, akWCSEND, akWCRECV, akWCPROBE.
akWCSTATE: This bitmask gives the state of the wait context. It can be:
	akWCFREE: The wait context is free. It is linked on the
		akWCFree list.
	akWCINUSE1: The wait context is in state InUse1. It is not
		linked in either of the two lists (akWCFree and akWCInUse2).
	akWCINUSE2: The wait context is in state InUse2. It is linked
		on the akWCInUse2 list.
	akWCINUSE3: The wait context is in state InUse3. It is not
		linked in either of the two lists (akWCFree and akWCInUse2).
akWCNOWAIT: the request will not be waited for (issued with a
	NULL handle).
akWCANY: then the request is waited for by a call to akWaitAny
	and sem_any is defined (as well as sem).
akWCDONE: the request already completely processed by the kernel. A
	wait ot test on this request will return 'completed' as result.
akCANCEL: the request was cancelled and akWait must return an error.
*/

/* the type must be the lower bits of the flags! */
#define akWCOTHER	0x00
#define akWCSEND	0x01
#define akWCRECV	0x02
#define akWCPROBE	0x03
#define akWCNTYPES	0x04 /* number of WC types */
#define akWCTYPE	0x03 /* 2-bit bitmask */

#define akWCFREE	0x00
#define akWCINUSE1	0x10
#define akWCINUSE2	0x20
#define akWCINUSE3	0x30
#define akWCSTATE	0x30 /* 2-bit bitmask */

#define akWCNOWAIT	0x40
#define akWCANY		0x80

#define akWCDONE	0x100
#define akCANCEL	0x200

/* Different possibilities (bitwise AND between flags and the akWC...):
 *  akWCNOWAIT            : sem and sem_any are not defined.
 * !akWCNOWAIT && !akWCANY: sem is defined, sem_any is not.
 * !akWCNOWAIT &&  akWCANY: sem and sem_any are defined.
 */

/* akWCTable
 * Every entry (wait context) of this table has 3 states:
 * Free:    It can be allocated to a thread for a new send/receive
 * In Use1: A thread has 'reserved' the wait context and is initializing
 *          a communication on it, but the wait context isn't ready yet for
 *          the daemon to test. (transient state, only the time of the
 *          akISend or akIRecv call. more or less...).
 * In Use2: There is a pending MPI communication on this wait context, whose
 *          completion hasn't been noticed yet by the daemon. The daemon must
 *          check this communication for completion.
 * In Use3: The daemon has found that the communication has completed, but
 *          the thread hasn't waited for it yet.
 * 
 */

/* When in state Free, the 'next' field of a wait context points to the
 * next wait context on the free list.
 * When in state InUse2, the 'next' field of a wait context points to the
 * next wait context on the InUse2 list. The 'prev' field point to the
 * previous wait context in the list, to easy extraction.
 * When in other states, the 'next' field has no meaning.
 */


/* Array of all wait contexts */
extern aktWaitContext akWCTable[];

/* First free wait context index or NULL */
extern aktWaitContext *akWCFree;

/* First in-use wait context index or NULL */
extern aktWaitContext *akWCInUse2;

/* wait context InUse2 list counter */
extern int akWCInUse2Count;

/*
 * akWCFreeMutex must be held locked by any thread modifying the akWCFree
 * list (getting a new free wait context or recycling a no-more-used one).
 *
 * akWCInUse2Mutex must be held locked by any thread modifying the akWCInUse2
 * list (or intending to modify it, under certain circumstances... :-). This
 * lock must be held by the daemon when it does a V on one of the semaphores
 * of the wait contexts (must be held prior to doing the V and must be released
 * only after the wait context has been removed from the akWCInUse2 list (state
 * of the wait context changed to InUse3 in this case)).
 * When a thread does an akWaitAny, it must do a loop of akSemaphoreTryP with
 * the akWCInUse2Mutex *locked*, and if none of the tryP succedes, then that
 * thread must add sem_any to all of the wait context while *still* holding
 * the lock, and only after that can he (the thread) unlock the mutex.
 *  To be Continued...
 *
 * they are no longer used. we use only one kernel mutex.
extern akpthread_mutex_t akWCFreeMutex;
extern akpthread_mutex_t akWCInUse2Mutex;
*/

/*****************************functions****************************/


extern int akWCInit(void);
extern int akWCMoveToInUse1(aktWaitContext **wcpp);
extern int akWCMoveToInUse2(aktWaitContext *wcp, void *handlep);
extern int akWCMoveToInUse3(aktWaitContext *curr);
extern int akWCMoveToFree(aktWaitContext *wc);
extern int akWCSetNoWait(aktWaitContext *wcp);
extern int akWCUnblock(aktWaitContext *wc);
extern int akWCCheck(aktWaitContext *curr);

#ifdef A0_TRACE
int akWCCallbackTrace(void (*callback)(), void *callback_par,
		int type, MPI_Status *statusp, unsigned long serial_no);
#else
int akWCCallback(void (*callback)(), void *callback_par);
#endif


#endif /* AKERNEL_SOURCE */

#endif /* _AKWC_H */



