/***************************************************************************/
/* File a0_perf_trace_buffer.c                                             */
/* Philippe Waille 05/07/96 - modified by Gustavo 16/12/97                 */
/***************************************************************************/

#include <malloc.h>
#include <stdlib.h>

#include "writev.h"

#include <akernel.h>
#include "a0semap.h"
#include "a0thread.h"
#include "a0init.h"
#include "a0prio.h"
#include "a0_perf_trace_buffer.h"
#include "a0_perf_event_macros.h"
#include "a0_perf_event_codes.h"

static char release [A0_PERF_RELEASE_SIZE] = A0_PERF_RELEASE;

static a0_perf_trace_buffer *free_list_first=NULL;
static a0_perf_trace_buffer *free_list_last=NULL;

static a0_perf_trace_buffer *full_list_first=NULL;
static a0_perf_trace_buffer *full_list_last=NULL;

static a0_perf_trace_buffer *flushing_list_first=NULL;
static a0_perf_trace_buffer *flushing_list_last=NULL;

static a0_perf_trace_buffer *global_list_first=NULL;
static a0_perf_trace_buffer *global_list_last=NULL;

static akpthread_mutex_t mutex;

static akpthread_cond_t	full_list_not_empty;
static akpthread_cond_t	free_list_not_empty;

a0tThread a0_flushing_daemon_id;
static int flushing_time;
a0tSemaphore flushing_dameon_died;

a0_perf_trace_buffer *last_buffer_to_dump = NULL;
a0_perf_trace_buffer *service_daemon_trace_buffer;

a0tContext a0_perf_Trace_buffer_context;

int trace_buffer_size = TRACE_BUFFER_SIZE;
int nb_trace_buffers = NB_TRACE_BUFFERS;
char trace_file_name [300] = DEFAULT_TRACE_FILE_NAME;
a0_perf_declare_time_variable(a0_perf_global_time);

FILE *ftrace;
FILE *fout = stdout;

struct iovec iov [IOV_MAX]; 

int a0_perf_nb_wait_buffer = 0;

static char header_buffer [3*A0_NODE_SIZE];


/* Get buffer : remove a buffer from the free list */
/* If none, waits for the daemons to return some flushed buffer */


a0_perf_trace_buffer *a0_perf_get_free_trace_buffer ()
{
a0_perf_trace_buffer *result;
akpthread_mutex_lock (&mutex);
if (free_list_first == NULL) a0_perf_nb_wait_buffer++;
while ((result = free_list_first) == NULL)
   {
   akpthread_cond_wait (&free_list_not_empty,&mutex); /* No buffer : wait */
   }
free_list_first = result-> next;		    /* removes an element */
if (free_list_first == NULL) free_list_last = NULL;
akpthread_mutex_unlock (&mutex);
return result;
}

/* Put buffer : add the buffer to the full queue */
/* Daemon can not sleep any more */

void a0_perf_put_full_trace_buffer (a0_perf_trace_buffer *buffer)
{
int size;
akpthread_mutex_lock (&mutex);
size = buffer -> index - buffer -> content;
/* Update size field : marks the buffer as already processed */
buffer -> size = size;
/* enqueue and wake up daemon */
buffer -> next = NULL;
if (full_list_last == NULL)
  {
  full_list_first = buffer;     /* empty queue : buffer becomes first element */
  }
else
  {
  full_list_last -> next = buffer;      /* put buffer after the existing last */
  }
full_list_last = buffer;        /* buffer becomes the last one */
akpthread_cond_signal (&full_list_not_empty);
akpthread_mutex_unlock (&mutex);
}

/* Add buffers to the free list and wakes up threads waiting for a buffer */

void a0_perf_put_free_trace_buffers (a0_perf_trace_buffer *first,
			             a0_perf_trace_buffer *last)
{
a0_perf_trace_buffer *buf;
int do_not_stop =1;
akpthread_mutex_lock (&mutex);
/* Reset all buffers in the list to empty state */
last -> next = NULL;
buf = first;
while ( buf != NULL)
  {
  buf -> size = A0_PERF_INITIAL_NULL_SIZE;
  buf -> index = buf -> content;
  buf = buf -> next;
  }
/* enqueue buffers list */
if (free_list_last != NULL) 
    free_list_last -> next = first;
if (free_list_first == NULL)
    free_list_first = first;
free_list_last = last;
akpthread_cond_signal (&free_list_not_empty);
akpthread_mutex_unlock (&mutex);
}

int a0_perf_get_full_trace_buffers ()
{
int no_stop = 1;
akpthread_mutex_lock (&mutex);
while ( (flushing_list_first = full_list_first) == NULL)
  {
  akpthread_cond_wait (&full_list_not_empty,&mutex); /* No buffer : wait */
  }
flushing_list_last = full_list_last;
full_list_first=NULL;
full_list_last=NULL;
if (flushing_time) no_stop=0;
akpthread_mutex_unlock (&mutex);
return(no_stop);
}


void a0_perf_init_free_list ()
{
int i;
a0_perf_trace_buffer *buf;
for (i=0;i<nb_trace_buffers;i++)
  {
  buf = (a0_perf_trace_buffer *) calloc (trace_buffer_size,1);
  buf -> high_water_mark = (char *) buf + trace_buffer_size -
			   A0_PERF_MAX_EVENT_RECORD_LENGTH;
  buf -> next_global = global_list_first;
  global_list_first = buf;
  if (global_list_last == NULL)
    global_list_last = buf;
  buf -> next = free_list_first;
  free_list_first = buf;
  if (free_list_last == NULL)
    free_list_last = buf;
  }
a0_perf_put_free_trace_buffers (global_list_first,global_list_last);
}

void a0_perf_flush_trace_buffers ()
{
a0_perf_trace_buffer *buf = global_list_first;
int size;
akpthread_mutex_lock (&mutex);
while (buf != NULL)
  {
  size = buf -> index - buf -> content;
  if ((buf -> size == A0_PERF_INITIAL_NULL_SIZE) &&  (size != 0))
    {   
    /* Update size field : marks the buffer as already processed */
    buf -> size = size;
    /* enqueue and wake up daemon */
    buf -> next = NULL;
    if (full_list_last == NULL)
      {
      full_list_first = buf;   /* empty queue : becomes first element */
      }
    else
      {
      full_list_last -> next = buf;    /* put buf after the last */
      }
    full_list_last = buf;      /* buf becomes the last one */
    }
  buf = buf -> next_global;
  }
/* There should exist at least a buffer submitted : the main thread
   trace buffer is never submitted by a end_of_thread event */
flushing_time =1; /* Last daemon iteration ... */
/* Wake up the flushing daemon */
akpthread_cond_signal (&full_list_not_empty);
akpthread_mutex_unlock (&mutex);
a0_perf_wait_daemon_end ();
}

void  a0_perf_trace_flushing_daemon ()
{
int j;
int size;
a0_perf_trace_buffer *current_buffer;
a0_perf_reader_buffer *rbuffer;
int iov_length;
char message [80];
char *a0_perf_index;
int do_not_stop =1;
while (do_not_stop)
  {
  do_not_stop = a0_perf_get_full_trace_buffers ();
  current_buffer = flushing_list_first;
  iov_length = 0;
  while (current_buffer != NULL)
    {
    size = current_buffer -> size;
    /* Skip header : pass only size and content part of the buffer */
    rbuffer = (a0_perf_reader_buffer *) &(current_buffer->size);

    /* Rewrite size field in variable length format */
    a0_perf_index = (char *) & (rbuffer -> size);
    a0_perf_put_integer (size);

    /* fills the buffer descriptor array for writev */

    iov [iov_length].iov_base = (char *) rbuffer;

    /* length : content size + size of size field (followed by content field) */
    iov [iov_length].iov_len = size + (char *) &rbuffer->content -
				      (char *) &rbuffer-> size;
    /* DEBUG 
    _a0_perf_decode_buffer (iov [iov_length].iov_base);
    to allow ascii buffer dump */
    iov_length ++;
    if (iov_length == IOV_MAX)
      {
      /* flush the first buffers to disk */
      writev (fileno(ftrace),iov,iov_length);
      /* now get ready for the next set of IOV_MAX buffers to dump */
      iov_length = 0;
      }
    current_buffer = current_buffer -> next;
    }
  /* flush the remaining buffers */
  if (iov_length >0)  
    {
     writev (fileno(ftrace),iov,iov_length);
    }

  /* free all the dumped buffers */
  a0_perf_put_free_trace_buffers (flushing_list_first,flushing_list_last);
  }
fclose (ftrace);
/* Wake Up main thread. Next action : main thread run akFinalize or exit */
_a0VSemaphore (&flushing_dameon_died);
}


void a0_perf_init_trace_buffers (int argc,char **argv)
{
/* akpthread_mutexattr_t mutex_attribute;*/
/* akpthread_condattr_t cond_attribute;  */
int result;
int i;
int value;
for (i=0;i<argc;i++) 
  {
  if (!strcmp (argv[i],"-h")) 
      printf ("Perf. options : -trace_file=myfile -nbuf=nb_buf -sbuf=buffer_size_in_bytes\n Min (default)  :  file : %s nbuf : %d (%d) sbuf : %d (%d)\n\n",DEFAULT_TRACE_FILE_NAME,MIN_NB_TRACE_BUFFERS,NB_TRACE_BUFFERS,MIN_TRACE_BUFFER_SIZE,TRACE_BUFFER_SIZE);
  if (!strncmp (argv[i],"-trace_file=",12))
     {
      strcpy (trace_file_name,argv[i]+12);
     }
  if (!strncmp (argv[i],"-nbuf=",6))
     {
     value = atoi (argv[i]+6);
     if (value>MIN_NB_TRACE_BUFFERS) nb_trace_buffers = value;
     }
  if (!strncmp (argv[i],"-sbuf=",6))
     {
     value = atoi (argv[i]+6);
     if (value>MIN_TRACE_BUFFER_SIZE) trace_buffer_size = value;
     }
  }

/* Set the key for the trace buffers as specific data */
_a0NewContext (&a0_perf_Trace_buffer_context);

/* Mutex, conditions and attributes for the buffer queues */
akpthread_mutex_init (&mutex,akPTHREAD_MUTEXATTR_DEFAULT_NP);
akpthread_cond_init (&full_list_not_empty,akPTHREAD_CONDATTR_DEFAULT_NP);
akpthread_cond_init (&free_list_not_empty,akPTHREAD_CONDATTR_DEFAULT_NP);
/* Create trace buffers */
a0_perf_init_free_list ();
/* Flushing daemon should run until main thread start the terminate action */
flushing_time=0;
_a0NewSemaphore (&flushing_dameon_died,0);
}

void a0_perf_start_flushing_daemon ()
{
void *arg;
char *a0_perf_index;
/* Open the trace file */
sprintf (trace_file_name + strlen (trace_file_name),".%d",a0SelfNode);
ftrace = fopen (trace_file_name,"w");
/* Write Standard infos to trace file : Release, TotalNode, Node */
write (fileno(ftrace),release,A0_PERF_RELEASE_SIZE);
a0_perf_index = header_buffer;
a0_perf_put_integer(a0NodeCount);
write (fileno(ftrace),header_buffer,A0_NODE_SIZE);
a0_perf_index = header_buffer;
a0_perf_put_integer(a0SelfNode);
write (fileno(ftrace),header_buffer,A0_NODE_SIZE);

/* Write the time reference */
a0_perf_index = header_buffer;
a0_perf_put_initial_time;
write (fileno(ftrace),header_buffer,A0_PERF_TIME_SIZE);

_a0NewThread (&a0_flushing_daemon_id,
                _a0DefaultScheduling,_a0DefaultPriority,
                FLUSHING_DAEMON_STACK,
                (a0tFunction) a0_perf_trace_flushing_daemon, arg);
}

void a0_perf_define_initial_time ()
{
a0_perf_set_initial_time;
}

void a0_perf_wait_daemon_end ()
{
void *status;
_a0PSemaphore (&flushing_dameon_died);
}

