/*
 *
 * _a0ports.c : Functions for the Ath0 'ports'
 *
 * by Ilan Ginzburg
 * Apache Project, IMAG, Grenoble, France
 * May 28, 1996
 *
 */


/* Athapascan0 source code */
#define A0_SOURCE

#include <akernel.h>
#include <errno.h>

#include "a0error.h"
#include "a0init.h"
#include "a0mpi.h"
#include "a0ports.h"


/* debugging macro */
#define a0Debug_on(a) printf a
#define a0Debug(a)

/* Error macro */
#define _a0PError(a) ( printf("_a0PError. %s %d:", __FILE__, __LINE__), printf a )

/* Communicators array for the ports */
MPI_Comm *_a0Communicators;

/* Current base tags for unique and non-unique ports creation */
static int _a0BaseTagUnique, _a0BaseTagNonUnique;
/* Mutex protecting access to above variables */
static akpthread_mutex_t _a0BaseTagMutex;

/***********************************************************************/

/* Does the necessary initializations concerning the ports and the associated
 * communicators:
 *    Creates all the communicators,
 *    Initialize base tags for unique and non-unique ports,
 *    Create mutex for protecting base tags variables.
 */
extern a0tError _a0InitPorts()
{
  int i;
  a0tPort port[2]; /* For creating the datatype */

  _a0Communicators = (MPI_Comm *) malloc((akSize+1) * sizeof(MPI_Comm));
  if (!_a0Communicators) return A0ErrNoMem;

  /* We create the array of communicators */
  for (i=0; i<=akSize; i++)
    akMPI_Comm_dup(MPI_COMM_WORLD, _a0Communicators + i);

  /* We initialize the counters of the base tags */
  _a0BaseTagUnique=0; /* Associated with _a0Communicators[akNode+1] */
  _a0BaseTagNonUnique=0; /* Associated with _a0Communicators[0] */

  /* Mutex protecting access to _a0BaseTagUnique and _a0BaseTagNonUnique */
  if (akpthread_mutex_init(&_a0BaseTagMutex, akPTHREAD_MUTEXATTR_DEFAULT_NP))
    return A0ErrNoMem;

  /* Creating the a0_PORT datatype */
  _a0NewStructDatatype(&a0_PORT, &port[0], 5,
  				&port[0],	1, MPI_LB,
  				&port[0].comm,	1, MPI_INT,
  				&port[0].basetag,1, MPI_INT,
  				&port[0].maxtag,1, MPI_INT,
  				&port[1],	1, MPI_UB);

  return A0ErrOk;
}


/***********************************************************************/
/* _a0NewPort. */
/***********************************************************************/
 
extern a0tError _a0NewPort(a0tPort *port, int tags, int unique) {

/* Creates a new port.
 * On the new port, tags tag are reserved (valid tags are in 0..tags-1),
 * if unique, the returned port is system-wide unique.
 * if !unique, ALL nodes must do the call and all are returned with the same
 * port. (all non-unique calls to _a0NewPort must be done in the same order in
 * all the nodes).
 */

  int basetag;

  if (!tags) tags = 1; /* 0 or 1 if only user tag 0 allowed */

  /* Increment the correct bas tag variable in a critical section */
  if (akpthread_mutex_lock(&_a0BaseTagMutex))
    return (errno==EINVAL) ? A0ErrInvalid : A0ErrLockedMutex;

  if (unique) {
    basetag = _a0BaseTagUnique;
    _a0BaseTagUnique += tags;
  }
  else {
    basetag = _a0BaseTagNonUnique;
    _a0BaseTagNonUnique += tags;
  }

  if (akpthread_mutex_unlock(&_a0BaseTagMutex)) return A0ErrInvalid;

  port->basetag = basetag; /* First MPI tag to use */
  port->maxtag = tags - 1; /* maximum user tag (min user tag is 0) */

  /* Communicator n. 0 is for non-unique allocation. For unique allocation,
   * every node uses it's 'own' communicator.
   */
  port->comm = unique ? akNode+1 : 0;

  a0Debug(("Port created: basetag %d, comm %d, maxtag %d\n",
  		port->basetag, port->comm, port->maxtag));
  return A0ErrOk;
}


/***********************************************************************/
/* _a0DisposePort. */
/***********************************************************************/

extern a0tError _a0DisposePort(a0tPort *port) {

/* Destroys a (unique) port, thus allowing the system to return it again
 * on a future call to _a0NewPort.
   * Questions to be considered by whom may implement this:
   * - ports freed on another node than the one allocated on.
   * - ports with more than one possible tag.
   */

  return A0ErrOk;
}
