/*
 *
 * a0pt2pt.c : Point to point communications
 *
 * by Ilan Ginzburg
 * Apache Project, IMAG, Grenoble, France
 * May 31, 1996
 *
 */


/* Athapascan0 source code */
#define A0_SOURCE

#include <akernel.h>
#include "a0error.h"
#include "a0pt2pt.h"
#include "a0buffer.h"
#include "a0thread.h"


/* debugging macro */
#define a0Debug_on(a) printf a
#define a0Debug(a)

#define A0CannotAny 0
#define A0CanAny 1


/*****************************************************************************
	_a0CheckNode - verifies if it's good node number
*****************************************************************************/

extern a0tError _a0CheckNode(int node, int any) {

  if (node>=akSize) return A0ErrInvNode;
  if (node==A0AnySource) {
    if (any!=A0CannotAny) return A0ErrOk;
    else		  return A0ErrInvNode;
  }
  if (node<0) return A0ErrInvNode;
  return A0ErrOk;
}


/*****************************************************************************
	_a0WaitRequest - Waits for a request to terminate.
*****************************************************************************/

extern a0tError _a0WaitRequest(a0tRequest *request) {

  a0tError err1;

  a0ThreadsBlockedComm++;
  err1 = _a0KernelErrorConvert(akWait(request));
  a0ThreadsBlockedComm--;

  return err1;
}



/*****************************************************************************
	_a0TestRequest - Tests for the termination of a request.
*****************************************************************************/

extern a0tError _a0TestRequest(a0tRequest *request, int *completed) {

  a0tError err1;

  err1 = _a0KernelErrorConvert(akTest(request, completed));

  return err1;
}


/*****************************************************************************
	_a0WaitAnyRequest - Waits for any request of table
*****************************************************************************/

extern a0tError _a0WaitAnyRequest(a0tRequest *reqtab[], int size,
					int *which) {

  a0tError err1;

  a0ThreadsBlockedComm++;
  err1 = _a0KernelErrorConvert(akWaitAny(reqtab, size, which));
  a0ThreadsBlockedComm--;

  return err1;
}


/*****************************************************************************
_a0TestAnyRequest - Tests for the termination of any request of a table
*****************************************************************************/

extern a0tError _a0TestAnyRequest(a0tRequest *reqtab[], int size,
					int *which) {

  a0tError err1;

  err1 = _a0KernelErrorConvert(akTestAny(reqtab, size, which));

  return err1;
}


/*****************************************************************************
	_a0GetRequestData - Get info from a request, after its completion
*****************************************************************************/

extern int _a0GetRequestData(a0tRequest *request,
				a0tRequestData *data) {

  data->source = request->status.MPI_SOURCE;
  if (akMPI_Get_count(&request->status, MPI_PACKED, &data->size))
    return A0ErrInvReq;
#ifdef A0_TRACE
  data->serial_no = request->serial_no;
#endif

  return A0ErrOk;
}


/*****************************************************************************
	_a0IProbeCallback - Non-blocking probe of a port, with callback
*****************************************************************************/

extern a0tError _a0IProbeCallback(
			a0tPort *port, int source, int tag,
			a0tRequest *request,
			void (*callback)(), void *callback_par) {

  a0tError r;

  if (r = _a0CheckNode(source, A0CanAny)) return r;

  /* callback is a function passed to the kernel which is executed
     when the request is completed. Buffers often need some cleanup
     after the transfer. */

  return _a0KernelErrorConvert(akIProbe(source, port->basetag + tag,
			_a0Communicators[port->comm],
			request, callback, callback_par));
}


/*****************************************************************************
	_a0IProbe - Non-blocking probe of a port
*****************************************************************************/

extern a0tError _a0IProbe(a0tPort *port, int source, int tag,
				a0tRequest *req) {

  return _a0IProbeCallback(port, source, tag, req, 0, 0);
}


/*****************************************************************************
	_a0Probe - blocking probe of a port
*****************************************************************************/

extern a0tError _a0Probe(a0tPort *port, int source, int tag,
				a0tRequest *req) {

  a0tError err;
  a0tRequest req_aux;

  if (!req) req = &req_aux;

  if (err = _a0IProbeCallback(port, source, tag, req, 0, 0))
    return err;

  return _a0WaitRequest(req);
}


/*****************************************************************************
_a0ISendBufferCallback - nonblocking send of a buffer, with callback
*****************************************************************************/

extern a0tError _a0ISendBufferCallback(
			a0tPort *port, int dest, int tag,
			a0tRequest *request, a0tBuffer *buffer,
			void (*callback)(), void *callback_par) {

  /* Performs a nonblocking send of a buffer. Must be received as
     a buffer on the other side.
     callback is a function passed to the kernel which is executed
     when the request is completed. Buffers often need some cleanup
     after the transfer. */

  a0tError err;
  int mpi_size;
  unsigned char *mpi_buff; /* this should be (byte *) */

  a0Debug(("_a0ISendBufferCallback: buf=%X siz=%d req=%x\n",
		buffer->storage, buffer->position, request));

  if (err = _a0CheckNode(dest, A0CannotAny)) return err;

  /* if buffer type is not A0SendBufferType, the user data in the storage
     are not right in the beginning of the storage. */

  mpi_buff = buffer->storage;
  if (buffer->bufftype != A0SendBufferType)
    mpi_buff += buffer->userdata;

  /* Check if tag is in allowed range */
  if (tag<0 || tag>port->maxtag)
    return A0ErrInvTag;

  /* We don't have much things to pack into the buffer, on the space that is
   * reserved before where the user packed his data. So we don't pack anything
   * there. This may change... :-) (and please change the comment too).
   */

  mpi_size = buffer->position;
  err = _a0KernelErrorConvert(
  		akISend(
  			(void *)mpi_buff, mpi_size, MPI_PACKED, dest,
  			port->basetag + tag, _a0Communicators[port->comm],
			request, callback, callback_par
		));

  a0Debug(("_a0ISendBufferCallback END ok\n"));
  return err;
}


/*****************************************************************************
	_a0ISendBuffer - non-blocking send of a buffer
*****************************************************************************/

extern a0tError _a0ISendBuffer(a0tPort *port, int dest, int tag,
				a0tRequest *request, a0tBuffer *buffer) {

  return _a0ISendBufferCallback(port, dest, tag, request, buffer,0,0);
}


/*****************************************************************************
	_a0SendBuffer - blocking send of a buffer
*****************************************************************************/

extern a0tError _a0SendBuffer(a0tPort *port, int dest, int tag,
				a0tRequest *request, a0tBuffer *buffer) {

  a0tError err;
  a0tRequest req_aux;

  if (!request) request = &req_aux;

  if (err = _a0ISendBufferCallback(port, dest, tag, request, buffer,0,0))
    return err;
  
  return _a0WaitRequest(request);
}


/*****************************************************************************
_a0ISendCallback -  non-blocking send of a datatype with callback
*****************************************************************************/

extern a0tError _a0ISendCallback(a0tPort *port, int dest, int tag,
			a0tRequest *request, a0tDatatype datatype,
			void *address, int count,
			void (*callback)(), void *callback_par) {

  a0tError err;

  a0Debug(("_a0ISendCallback: req=%x adr=%X siz=%d\n",
		request, address, count));

  if (err = _a0CheckNode(dest, A0CannotAny)) return err;

  /* Check if tag is in allowed range */
  if (tag<0 || tag>port->maxtag)
    return A0ErrInvTag;

  if (err = _a0KernelErrorConvert(
  		akISend(
  			address, count, datatype, dest,
  			port->basetag + tag, _a0Communicators[port->comm],
			request, callback, callback_par
		)))
    return err;

  a0Debug(("_a0ISendCallback req=%x END ok\n", request));

  return A0ErrOk;
}


/*****************************************************************************
	_a0ISend - non-blocking send of a datatype
*****************************************************************************/

extern a0tError _a0ISend(a0tPort *port, int dest, int tag,
			a0tRequest *request, a0tDatatype datatype,
			void *address, int count) {

  return _a0ISendCallback(port, dest, tag, request,
				datatype, address, count, 0, 0);
}


/*****************************************************************************
	_a0Send - blocking send of a datatype
*****************************************************************************/

extern a0tError _a0Send(a0tPort *port, int dest, int tag,
				a0tRequest *request, a0tDatatype datatype,
				void *address, int count) {

  a0tError err;
  a0tRequest req_aux;

  if (!request) request = &req_aux;

  if (err = _a0ISendCallback(port, dest, tag, request,
				datatype, address, count, 0, 0))
    return err;
  
  return _a0WaitRequest(request);
}


/*****************************************************************************
_a0IReceiveBufferCallback - nonblocking receive in a buffer, with callback
*****************************************************************************/

extern a0tError _a0IReceiveBufferCallback(
			a0tPort *port, int source, int tag,
			a0tRequest *request, a0tBuffer *buffer,
			void (*callback)(), void *callback_par) {

  /* Performs a nonblocking receive in a buffer. The message must
     have been sent as a buffer on the other side. The buffer must
     be of type A0SendBufferType.
     callback is a function passed to the kernel which is executed
     when the request is completed. Buffers often need some cleanup
     after the transfer. */

  a0tError err;
  long dummy;
  int mpi_size;
  unsigned char *mpi_buff; /* this should be (byte *) */

  if (err = _a0CheckNode(source, A0CanAny)) return err;

  /* if buffer type is not A0SendBufferType, the user data in the storage
     are not right in the beginning of the storage. */

  mpi_buff = buffer->storage;
  if (buffer->bufftype != A0SendBufferType)
    mpi_buff += buffer->userdata;

  /* Check if tag is in allowed range */
  if (tag<0 || tag>port->maxtag)
    return A0ErrInvTag;

  a0Debug(("_a0IReceiveBuffer. tag %d, comm %d\n", port->basetag + tag, port->comm));

  mpi_size = buffer->size;
  if (err = _a0KernelErrorConvert(
  		akIRecv(
  			(void *)mpi_buff, mpi_size, MPI_PACKED, source,
  			port->basetag + tag, _a0Communicators[port->comm],
			request, callback, callback_par
		)))
    return err;

  /* Ok... Nothing has been unpacked yet, we position the buffer accordingly */
  buffer->position = buffer->userdata;

  return A0ErrOk;
}


/*****************************************************************************
	_a0IReceiveBuffer - non-blocking receive in a buffer
*****************************************************************************/

extern a0tError _a0IReceiveBuffer(a0tPort *port, int source, int tag,
				a0tRequest *request, a0tBuffer *buffer) {

  return _a0IReceiveBufferCallback(port, source, tag, request, buffer,0,0);
}


/*****************************************************************************
	_a0ReceiveBuffer - blocking receive in a buffer
*****************************************************************************/

extern a0tError _a0ReceiveBuffer(a0tPort *port, int source, int tag,
				a0tRequest *request, a0tBuffer *buffer) {

  a0tError err;
  a0tRequest req_aux;

  if (!request) request = &req_aux;

  if (err = _a0IReceiveBufferCallback(port, source, tag, request, buffer,0,0))
    return err;
  
  return _a0WaitRequest(request);
}


/*****************************************************************************
_a0IReceiveCallback - nonblocking receive of a datatype, with callback
*****************************************************************************/

extern a0tError _a0IReceiveCallback(a0tPort *port, int source, int tag,
				a0tRequest *request, a0tDatatype datatype,
				void *address, int count,
			void (*callback)(), void *callback_par) {

  a0tError err;

  a0Debug(("_a0IReceiveCallback: req=%x adr=%X siz=%d\n",
		request, address, count));

  if (err = _a0CheckNode(source, A0CanAny)) return err;

  /* Check if tag is in allowed range */
  if (tag<0 || tag>port->maxtag)
    return A0ErrInvTag;

  if (err = _a0KernelErrorConvert(
  		akIRecv(
  			address, count, datatype, source,
  			port->basetag + tag, _a0Communicators[port->comm],
			request, callback, callback_par
		)))
    return err;

  a0Debug(("_a0IReceiveCallback req=%x END ok\n", request));

  return A0ErrOk;
}


/*****************************************************************************
	_a0IReceive - nonblocking receive of a datatype
*****************************************************************************/

extern a0tError _a0IReceive(a0tPort *port, int source, int tag,
				a0tRequest *request, a0tDatatype datatype,
				void *address, int count) {

  return _a0IReceiveCallback(port, source, tag, request,
				datatype, address, count, 0, 0);
}


/*****************************************************************************
	_a0Receive - blocking receive of a datatype
*****************************************************************************/

extern a0tError _a0Receive(a0tPort *port, int source, int tag,
				a0tRequest *request, a0tDatatype datatype,
				void *address, int count) {

  a0tError err;
  a0tRequest req_aux;

  if (!request) request = &req_aux;

  if (err = _a0IReceiveCallback(port, source, tag, request,
				datatype, address, count, 0, 0))
    return err;
  
  return _a0WaitRequest(request);
}
