/*
 *
 * a0servd.c : Functions for the SERV daemon.
 *
 * The SERV daemon is the one responsible for doing remote service invocation.
 *    (= remote thread creation).
 *
 * by Ilan Ginzburg
 * Apache Project, IMAG, Grenoble, France
 * Feb 16, 1996
 *
 */


/* Athapascan0 source code */
#define A0_SOURCE

#include <akernel.h>
#include "a0error.h"
#include "a0servd.h"
#include "a0specific.h"
#include "a0services.h"
#include "a0thread.h"
#include "a0mpi.h"
#include "a0pt2pt.h"
#include "akinfo.h"

#ifdef A0_TRACE
#include "a0_perf_start_remote_service.h"
#endif


/* debugging macro */
#define a0Debug_on(a) printf a
#define a0Debug(a)


/*****************************************************************************/

/* Port used for communications with the SERV daemon */
a0tPort _a0ServdPort;

/* Daemon thread id */
static a0tThread _a0ServDaemonThread;

static _a0ServDaemonRun = 1;


/*****************************************************************************/
/* This function is called when the buffer for the RTC has been allocated
 * and received. The function unpacks what necessary from the buffer, and calls
 * the requested service. It's here only to keep _a0ServDaemonFunction() short.
 */

static a0tError _a0CallThatServicePlease(a0tThreadSpecific *threadData,
					a0tRequest *requestp) {

  /* values packed in the header of the buffer */
  int service_number, priority;
  a0tScheduling scheduling;
  size_t stack_size;
  a0tThread thread;
  a0tRemoteFunction Code;

  a0tRTCFunction service_function;

  /* We unpack athapascan data from the beginning of the buffer. */
  _a0Unpack(&threadData->buffer, a0_INT, &service_number, 1);
  _a0Unpack(&threadData->buffer, a0_INT, &scheduling, 1);
  _a0Unpack(&threadData->buffer, a0_INT, &priority, 1);
  _a0Unpack(&threadData->buffer, a0_UNSIGNED_LONG, &stack_size, 1);

  a0Debug(("SERV received: serv# %d, sched %d, prio %d, stack %ld "
		"buf %ld pos %ld\n",
  		service_number, scheduling, priority, stack_size,
  		threadData->buffer.size, threadData->buffer.position));

  if (service_number == A0TerminationService) {
      akInfo(SERVICE_TYPE);
      akpthread_exit(0);
  }

  /* Now the buffer is ready to be passed to the service thread, because
   * everything packed by athapascan has been unpacked.
   */

  /* Get the function to call, and possibly modify scheduling, priority and/or
   * stack size.
   */
  if (_a0LookupRTCService(service_number,
  		      &Code,
		      &scheduling,
		      &priority,
		      &stack_size))
    _a0Error(A0ErrInvService, "RTC");

  a0Debug(("After _a0LookupRTCService: service_num=%d\n",
    			service_number));

  /* a service gets its buffer as parameter */
  threadData->function = (a0tFunction) Code;
  threadData->param = (void *) &(threadData->buffer);

  a0Debug(("Call Service: spc=%X fcn=%X, buf=%x, prio=%d, "
			"sched=%d, stack=%d\n",
    			threadData, threadData->function,
    			threadData->param,
    			priority, scheduling, stack_size));

  /* ...and start this service */

#ifdef A0_TRACE
   {
   /* Insert here begin of the thread creation trace : records date + service */
   a0_perf_start_remote_service_begin (service_number)
#endif

  if (_a0StartThreadPlease(&thread,
			scheduling, priority, stack_size,
			threadData)) {
	_a0Error(A0ErrDaemonThreadCreate, "RTC");
  }

#ifdef A0_TRACE
  /* TRACE: Insert here end of macro for RTC thread creation tracing.
   * The logical thread id of the created thread is in threadData->logical_id
   */
  a0_perf_start_remote_service_end (thread.logical_id, requestp);
  }
#endif

  a0Debug(("Daemon before _a0DetachThread\n"));

  /* We won't wait for this thread to terminate, so we let it die by itself */
  if (_a0DetachThread(&thread)) {
	_a0Error(A0ErrDetach, "RTC");
  }
}


/*****************************************************************************/
/* This is the SERV daemon thread */

static int _a0ServDaemonFunction(void *arg) {

  a0tThreadSpecific *threadData;
  a0tRequest request;
  a0tRequestData req_data;

  a0Debug(("Start of _a0ServDaemonFunction (%d)\n", akpthread_self()));

#ifdef A0_TRACE
  {
    unsigned long id;
    /* find a trace id and set it */
    akTraceNewThreadId(&id);
    akTraceSetThreadId(id);
  }
#endif /* A0_TRACE */

  while (_a0ServDaemonRun) {
    /* Get a new thread specific area for the to-be-created thread */
    a0Debug(("_a0ServDaemonFunction: before _a0NewSpecificData\n"));
    _a0NewSpecificData(&threadData);
    a0Debug(("_a0ServDaemonFunction: after _a0NewSpecificData\n"));

    /* Do a blocking probe to get an incoming request.
     * We do a probe so we can allocate the buffer to do the receive on,
     * as remote thread creation buffers can have an arbitrary size.
     */
    /* We do a horrible gororoba to compensate for thread accounting...
     * (we don't want the accounting to reflect athapascan's daemons
     * activity)
     */
    a0ThreadsBlockedComm--; /* gororoba part 1/2 */
    _a0Probe(&_a0ServdPort, A0AnySource, 0, &request);
    a0ThreadsBlockedComm++; /* gororoba part 2/2 */

    /* Get the size of the arriving message */
    _a0GetRequestData(&request, &req_data);

    /* Allocate a RTC buffer of the correct size. (Actually the allocated
     * buffer is greater than necessary, because size includes the size of
     * the header of the buffer, and in _a0NewBuffer we allocate size bytes +
     * the size of the header. But who cares?... )
     */
    _a0NewBuffer(&threadData->buffer, A0ThreadBufferType, req_data.size);
    threadData->flags |= A0SpecificBuffer; /* We've got a buffer there... */

    /* Get the request in the buffer (hopefully the probe worked...) */
    akRecv(threadData->buffer.storage, threadData->buffer.size,
                MPI_PACKED, A0AnySource, _a0ServdPort.basetag,
		_a0Communicators[_a0ServdPort.comm], &request);

    /* The buffer has just been received, nothing unpacked, so the position
     * is at the beginning of the buffer.
     */
    threadData->buffer.position = 0;

    /* Do all the necessary unpacking and call the service... */
    _a0CallThatServicePlease(threadData, &request);

  }
}

/*****************************************************************************/

/* This function does the initializations needed for the SERV daemon, creates
 * the SERV daemon thread and returns.
 * Upon return from this function, all that's needed by the threads
 * on this node to adress SERV daemons on other nodes is initialized (as well
 * as all that's needed by the SERV daemon to answer requests from
 * other nodes).
 */
a0tError _a0InitServ()
{
  int rc;

  /* Global creation of a port for communicating with the SERV daemon */
  _a0NewPort(&_a0ServdPort, 0, FALSE);

  /* Gororoba hack... We decrease the total number of threads because
   * _a0NewThread will increase it, and we don't want to include Athapscan
   * threads in this counter.
   */
  a0ThreadsTotal--; /* Sorry 'bout that! */

  /* Creation of the SERV daemon thread.  */
  rc = _a0NewThread(&_a0ServDaemonThread,
    		 _a0DaemonScheduling,
		 _a0PriDaemonHigh,
		 _a0DefaultStack,
		 _a0ServDaemonFunction,
		 (void *) NULL);

  a0Debug(("_a0NewThread SERV returned %d\n", rc));
}

/*****************************************************************************/

a0tError _a0TermServ()
{
	void *res;
	a0tBuffer buf;
	a0tRequest re;

	_a0NewBuffer(&buf, A0ThreadBufferType, 0);
	_a0ServDaemonRun = 0;
	_a0StartRemoteThread(akNode, A0TerminationService, 
				A0ServiceScheduling, A0ServicePriority,
				A0ServiceStack, &re, &buf);
	akpthread_join(_a0ServDaemonThread.pth_id, &res);
}
