#include "a0_perf_internal_fmem.h"
#include "../basic/a0_perf_event_codes.h"
#include "../basic/a0_perf_event_macros.h"

/* Called within a0fread on the initiator node side */
/* RTC will be followed by a recv_reply from the remote node */
/* Records the serial_no for the RTC call */
/* Implicit informations : remote service = Read, recv port = MemReply */
/* Recv_reply serial_no found in the fread request parameter */

int a0_perf_init_fread (int tag, int remote_node, long RTC_serial_no)
{
unsigned long serial_no;
a0_perf_begin_event;
a0_perf_get_index;
a0_perf_put_event_id(A0_PERF_RECORD_ID_INIT_FREAD);
a0_perf_put_integer(remote_node);
a0_perf_put_integer(tag);
a0_perf_put_integer(RTC_serial_no);
a0_perf_put_index;
}

/* Called within a0fwrite on the initiator node side */
/* RTC will be followed by a send_data and a recv_reply */
/* Records the tag used for the receive reply from the remote node */
/* Records the serial_no for the RTC call */
/* Records the serial_no for the send_data */
/* Implicit informations : remote service = Write */
/* send_data port = MemAsk, recv_reply port = MemReply */
/* tag common to both send and recv */
/* Recv_reply serial_no found in the fread request parameter */

int a0_perf_init_fwrite (int tag, int remote_node, 
			  long RTC_serial_no, long send_serial_no)
{
a0_perf_begin_event;
a0_perf_get_index;
a0_perf_put_event_id(A0_PERF_RECORD_ID_INIT_FWRITE);
a0_perf_put_integer(remote_node);
a0_perf_put_integer(tag);
a0_perf_put_integer(RTC_serial_no);
a0_perf_put_integer(send_serial_no);
a0_perf_put_index;
}

/* Called within a0fcopy on the initiator node side */
/* RTC to source node will be followed by a recv_reply from dest node */
/* Records the serial_no for the RTC call */
/* Implicit informations : remote service = Copy, recv port = MemReply */
/* Recv_reply serial_no found in the fread request parameter */

int a0_perf_init_fcopy (int tag, int source_node, int dest_node,
			  long RTC_serial_no)
{
a0_perf_begin_event;
a0_perf_get_index;
a0_perf_put_event_id(A0_PERF_RECORD_ID_INIT_FCOPY);
a0_perf_put_integer(source_node);
a0_perf_put_integer(dest_node);
a0_perf_put_integer(tag);
a0_perf_put_integer(RTC_serial_no);
a0_perf_put_index;
}


/* Called within a0flock on the initiator node side */
/* RTC to source node will be followed by a recv_reply from remote node */
/* Records the serial_no for the RTC call */
/* Implicit informations : remote service = Lock, recv port = MemReply */
/* Recv_reply serial_no found in the fread request parameter */

int a0_perf_init_flock (int tag, int remote_node, long RTC_serial_no)
{
unsigned long serial_no;
a0_perf_begin_event;
a0_perf_get_index;
a0_perf_put_event_id(A0_PERF_RECORD_ID_INIT_FLOCK);
a0_perf_put_integer(remote_node);
a0_perf_put_integer(tag);
a0_perf_put_integer(RTC_serial_no);
a0_perf_put_index;
}

/* Called within a0flock when a0SetLockRead was performed before */
/* Same as before with a recv_data between RTC and recv_reply */
/* Records recv_data serial_no */
/* Tag comon to both and recv */
/* Implicit recv_data port : MemReply */

int a0_perf_init_flockread (int tag, int remote_node, 
			long RTC_serial_no, long recv_serial_no)
{
a0_perf_begin_event;
a0_perf_get_index;
a0_perf_put_event_id(A0_PERF_RECORD_ID_INIT_FLOCKREAD);
a0_perf_put_integer(remote_node);
a0_perf_put_integer(tag);
a0_perf_put_integer(RTC_serial_no);
a0_perf_put_integer(recv_serial_no);
a0_perf_put_index;
}


/* Called within a0funlock on the initiator node side */
/* RTC to source node will be followed by a recv_reply from remote node */
/* Records the serial_no for the RTC call */
/* Implicit informations : remote service = Unlock, recv port = MemReply */
/* Recv_reply serial_no found in the fread request parameter */

int a0_perf_init_funlock (int tag, int remote_node, long RTC_serial_no)
{
a0_perf_begin_event;
a0_perf_get_index;
a0_perf_put_event_id(A0_PERF_RECORD_ID_INIT_FUNLOCK);
a0_perf_put_integer(remote_node);
a0_perf_put_integer(tag);
a0_perf_put_integer(RTC_serial_no);
a0_perf_put_index;
}


/* Called within a0flock when a0SetLockRead was performed before */
/* Same as before with a send_data between RTC and recv_reply */
/* Records send_data serial_no */
/* Tag comon to both send and recv */
/* Implicit send_data port : MemAsk */

int a0_perf_init_funlockwrite (int tag, int remote_node, 
			  long RTC_serial_no,long send_serial_no)
{
a0_perf_begin_event;
a0_perf_get_index;
a0_perf_put_event_id(A0_PERF_RECORD_ID_INIT_FUNLOCKWRITE);
a0_perf_put_integer(remote_node);
a0_perf_put_integer(tag);
a0_perf_put_integer(RTC_serial_no);
a0_perf_put_integer(send_serial_no);
a0_perf_put_index;
}

/* Called within the copy service thread on the source (region1) node */
/* RTC to dest node followed by a send data */
/* Implicit service : Write */
/* Implicit send_data port : MemAsk */
/* Records both RTC and send_data seriel_no */
/* Implicit source node : RTC Copy initiator */

int a0_perf_write_copy (int tag, int dest_node, 
			 long RTC_serial_no, long send_serial_no)
{
a0_perf_begin_event;
a0_perf_get_index;
a0_perf_put_event_id(A0_PERF_RECORD_ID_WRITECOPY);
a0_perf_put_integer(dest_node);
a0_perf_put_integer(tag);
a0_perf_put_integer(RTC_serial_no);
a0_perf_put_integer(send_serial_no);
a0_perf_put_index;
}


/* Called within the read service thread */
/* Sends data to initiator node */
/* Implicit send_data port : MemReply */
/* Records send_data seriel_no */

int a0_perf_reply_read (int tag, int reply_node,long serial_no)
{
a0_perf_begin_event;
a0_perf_get_index;
a0_perf_put_event_id(A0_PERF_RECORD_ID_REPLYREAD);
a0_perf_put_integer(reply_node);
a0_perf_put_integer(tag);
a0_perf_put_integer(serial_no);
a0_perf_put_index;
}

/* Called within the write service thread */
/* Recv data from source node */
/* Sends a reply to initiator node */
/* Same source and initiator for a fwrite */
/* Not the same for a write within a fcopy */
/* Implicit recv_data port : MemAsk */
/* Implicit send_reply  port : MemReply */
/* Records both recv_data and send_reply seriel_no */


int a0_perf_reply_write (int tag, int source_node, int reply_node,
			  long recv_serial_no, long send_serial_no)
{
a0_perf_begin_event;
a0_perf_get_index;
a0_perf_put_event_id(A0_PERF_RECORD_ID_REPLYWRITE);
a0_perf_put_integer(source_node);
a0_perf_put_integer(reply_node);
a0_perf_put_integer(tag);
a0_perf_put_integer(recv_serial_no);
a0_perf_put_integer(send_serial_no);
a0_perf_put_index;
}
