/*
 *
 * a0flock.c : a0 Functions of DMA lock
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * Oct 18, 1996
 *
 */

#define A0_SOURCE
#include <ath0.h>
#include <stdlib.h>

#define A0F_SOURCE
#include "a0fmem.h"
#include "a0flock.h"
#include "a0flockd.h"


#define a0Debug_on(a) printf a
#define a0Debug(a)


/******************local variables & definitions*****************/

typedef struct {
  a0tLock	*locktable;
  int		count;
  a0tRequest	*userreqp;
} a0tOngoingLock;


/* mutex to control access to the lock table */
static a0tMutex _a0OngoingMutex;


/*************************************************************************
	_a0InitLocks
*************************************************************************/

a0tError _a0InitLocks(void) {

  a0tError ret;

  ret = _a0InitLockServer();
  if (ret) return ret;

  ret = _a0NewMutex(&_a0OngoingMutex);
  return ret;
}


/*************************************************************************
	_a0RegionCompare - 
*************************************************************************/

static int _a0RegionCompare(const void *p1, const void *p2) {

  int dif, n;
  unsigned char *s1, *s2;
  a0tLock *l1, *l2;

  l1 = (a0tLock *) p1;
  l2 = (a0tLock *) p2;

  /* nodes are different, return difference */
  if (dif = (l1->regionp->node - l2->regionp->node)) return dif;

  /* addresses are different, return difference */
  s1 = (unsigned char *) &l1->regionp->data;
  s2 = (unsigned char *) &l2->regionp->data;
  for (n=0; n<sizeof(a0tMemoryAddress); n++)
    if (dif = *s1++ - *s2++) return dif;

  /* formats are different, return difference */
  s1 = (unsigned char *) &l1->regionp->format;
  s2 = (unsigned char *) &l2->regionp->format;
  for (n=0; n<sizeof(a0tMemoryAddress); n++)
    if (dif = *s1++ - *s2++) return dif;

  /* starts are different, return difference */
  if (dif = (l1->start - l2->start)) return dif;

  /* sizes are different, return difference */
  if (dif = (l1->size - l2->size)) return dif;

  /* nothing is different */
  return 0;
}


/*************************************************************************
	_a0AskLock
*************************************************************************/

static a0tError _a0AskLock(a0tLock *lock, a0tRequest *requestp,
			void (*callback)(), void *callback_par) {


  a0tMemRequest *message;
  a0tError r;
  int tag;

  message = (a0tMemRequest *) malloc(sizeof(a0tMemRequest));
  if (!message) return A0ErrNoMem;
  _a0FindMemTag(&tag);

  message->type =  (lock->read.formatp)? A0LockReadRequest: A0LockRequest;
  message->region = *(lock->regionp);
  message->region_start = lock->start;
  message->region_size = lock->size;
  message->reply_node = akNode;
  message->reply_tag = tag;

  r = _a0ISendCallback(&_a0MemServerPort, lock->regionp->node, 0, NULL,
		a0_MEM_REQUEST, message, 1, free, message);
  if (r) return r;

  a0Debug(("post lock(%s) req to %d tag=%d fmt=%x d=%x st=%d sz=%d\n",
	(lock->read.formatp?"read":""), lock->regionp->node, tag,
	lock->regionp->format.address, lock->regionp->data.address,
	lock->start, lock->size));

  /* if has read-at-lock, receive data */
  if (lock->read.formatp) {
    r = _a0IFReceive(&_a0MemReplyPort, lock->regionp->node, tag,
			NULL,
			lock->read.formatp,
			lock->read.address,
			lock->read.count);
    if (r) return r;
  }
  r = _a0IFReceiveCallback(&_a0MemReplyPort, lock->regionp->node, tag,
			requestp, A0FAddress, &lock->lockaddr, 1,
			callback, callback_par);

  a0Debug(("_a0AskLock END\n"));

  return r;
}


/*************************************************************************
	_a0LockCleanup - 
*************************************************************************/

static a0tError _a0LockCleanup(a0tOngoingLock *ong) {

  int done;

  /* callback function for an elementary lock of a multiple lock */

  a0Debug(("_a0LockCleanup %d\n", ong->count));

  if (ong->count) {
    ong->count --;
    /* lock the next */
    _a0AskLock(ong->locktable + ong->count, NULL,
    			(void(*)())_a0LockCleanup, ong);
  }
  else {
    a0Debug(("_a0LockCleanup unblock %x\n", ong->userreqp));
    if (ong->userreqp) akUnblock(ong->userreqp);
    free(ong);
  }
}


/*************************************************************************
	_a0ILock
*************************************************************************/

a0tError _a0ILock(a0tLock *lock, int count, a0tRequest *requestp) {

  a0tOngoingLock *ong;
  int i;
  a0tError r;

  for (i=0; i<count; i++)
    if ((lock[i].start+lock[i].size) > lock[i].regionp->size)
      return A0ErrInvalid;

  if (count > 1) {
    ong = (a0tOngoingLock *) malloc (sizeof(a0tOngoingLock));
    if (!ong) return A0ErrNoMem;

    ong->locktable = lock;
    ong->userreqp = requestp;
    ong->count = count - 1;
    if (requestp) akBlock(requestp, 0, 0);

    /* sort the locks */
    qsort(lock, count, sizeof(a0tLock), _a0RegionCompare);

    /* lock the first */
    r = _a0AskLock(lock + ong->count, NULL, (void(*)())_a0LockCleanup, ong);
    return r;
  }
  else
    return _a0AskLock(lock, requestp, 0, 0);
}


/*************************************************************************
	_a0Lock
*************************************************************************/

a0tError _a0Lock(a0tLock *lock, int count, a0tRequest *requestp) {

  a0tError ret;
  a0tRequest req_aux;

  if (!requestp) requestp = &req_aux;
  ret = _a0ILock(lock, count, requestp);
  if (ret) return ret;
  ret = _a0WaitRequest(requestp);

#ifdef DEBUG_LOCK
  { int i;
    for (i=0; i<count; i++)
      printf("_a0Lock[%d]=%x\n", i, lock[i].lockaddr.address);
  }
#endif
  return ret;
}


/*************************************************************************
	_a0IFReadPartLock
*************************************************************************/

a0tError _a0IFReadPartLock(a0tLock *lock,
			a0tDMARegion *region, int start, int size,
			a0tRequest *requestp,
			a0tFormat *formatp, void *address, int count) {

  a0tError ret;

  ret = _a0SetLockDMARegionPart(lock, 0, region, start, size);
  if (ret) return ret;
  _a0SetLockRead(lock, 0, formatp, address, count);

  ret = _a0ILock(lock, 1, requestp);
  return ret;
}


/*************************************************************************
	_a0FReadPartLock
*************************************************************************/

a0tError _a0FReadPartLock(a0tLock *lock,
			a0tDMARegion *region, int start, int size,
			a0tRequest *requestp,
			a0tFormat *formatp, void *address, int count) {

  a0tError ret;
  a0tRequest req_aux;

  ret = _a0SetLockDMARegionPart(lock, 0, region, start, size);
  if (ret) return ret;
  _a0SetLockRead(lock, 0, formatp, address, count);

  if (!requestp) requestp = &req_aux;
  ret = _a0ILock(lock, 1, requestp);
  if (ret) return ret;
  ret = _a0WaitRequest(requestp);
}


/*************************************************************************
	_a0AskUnlock
*************************************************************************/

static a0tError _a0AskUnlock(a0tLock *lock, a0tRequest *requestp,
			void (*callback)(), void *callback_par) {

  a0tMemRequest *message;
  a0tError r;
  int tag;

  message = (a0tMemRequest *) malloc(sizeof(a0tMemRequest));
  if (!message) return A0ErrNoMem;
  _a0FindMemTag(&tag);

  /* dirty stuff: to keep the mem request compatible, pass lock address
     and answer tag inside the region. region.data = lock address and
     region.node = reply tag. */
  message->type =  (lock->write.formatp)? A0UnlockWriteRequest: A0UnlockRequest;
  message->region.data = lock->lockaddr;
  message->region.node = tag;

  r = _a0ISendCallback(&_a0MemServerPort, lock->regionp->node, 0, NULL,
		a0_SHORT_MEM_REQUEST, message, 1, free, message);
  if (r) return r;

  a0Debug(("post unlock(%s) %x to %d fmt=%x d=%x st=%d sz=%d tg=%d\n",
	(lock->write.formatp?"write":""), lock->lockaddr.address,
	lock->regionp->node,
	lock->regionp->format.address, lock->regionp->data.address,
	lock->start, lock->size, tag));

  /* if a write-at-unlock is done, send data */
  if (lock->write.formatp) {
    r = _a0IFSend(&_a0MemAskPort, lock->regionp->node, tag,
			NULL,
			lock->read.formatp,
			lock->read.address,
			lock->read.count);
    if (r) return r;
  }
  return _a0IReceiveCallback(&_a0MemReplyPort, lock->regionp->node, tag,
			requestp, a0_INT, &_a0MemDummy, 0,
			callback, callback_par);
}


/*************************************************************************
	_a0UnlockCleanup - 
*************************************************************************/

static a0tError _a0UnlockCleanup(a0tOngoingLock *ong) {

  int count;

  /* callback function for an elementary unlock of a multiple lock */

  if (_a0LockMutex(&_a0OngoingMutex)) _a0Error(A0ErrMutexLock, "ong");
  count = (-- ong->count);
  if (_a0UnlockMutex(&_a0OngoingMutex)) _a0Error(A0ErrMutexUnlock, "ong");

  a0Debug(("_a0UnlockCleanup\n"));

  if (!count) {
    a0Debug(("_a0UnlockCleanup unblock\n"));
    if (ong->userreqp) akUnblock(ong->userreqp);
    free(ong);
  }
}


/*************************************************************************
	_a0IUnlock
*************************************************************************/

a0tError _a0IUnlock(a0tLock *lock, int count, a0tRequest *requestp) {

  a0tOngoingLock *ong;
  int i;
  a0tError r;

  if (count > 1) {
    ong = (a0tOngoingLock *) malloc (sizeof(a0tOngoingLock));
    if (!ong) return A0ErrNoMem;

    ong->userreqp = requestp;
    ong->count = count;
    if (requestp) akBlock(requestp, 0, 0);

    for (i=0; i<count; i++) {
      r = _a0AskUnlock(lock+i, NULL, (void(*)()) _a0UnlockCleanup, ong);
      if (r) return r;
      /* should free(ong), but the callback can stil run, so... */
    }
    return A0ErrOk;
  }
  else
    return _a0AskUnlock(lock, requestp, 0, 0);
}


/*************************************************************************
	_a0Unlock
*************************************************************************/

a0tError _a0Unlock(a0tLock *lock, int count, a0tRequest *requestp) {

  a0tError ret;
  a0tRequest req_aux;

  if (!requestp) requestp = &req_aux;
  ret = _a0IUnlock(lock, count, requestp);
  if (ret) return ret;
  return _a0WaitRequest(requestp);
}


/*************************************************************************
	_a0IFWriteUnlock
*************************************************************************/

a0tError _a0IFWriteUnlock(a0tLock *lock, a0tRequest *requestp,
			a0tFormat *formatp, void *address, int count) {

  a0tError ret;

  _a0SetUnlockWrite(lock, 0, formatp, address, count);

  ret = _a0IUnlock(lock, 1, requestp);
  return ret;
}


/*************************************************************************
	_a0FWriteUnlock
*************************************************************************/

a0tError _a0FWriteUnlock(a0tLock *lock, a0tRequest *requestp,
			a0tFormat *formatp, void *address, int count) {

  a0tError ret;
  a0tRequest req_aux;

  _a0SetUnlockWrite(lock, 0, formatp, address, count);

  if (!requestp) requestp = &req_aux;
  ret = _a0IUnlock(lock, 1, requestp);
  if (ret) return ret;
  return _a0WaitRequest(requestp);
}
