
/*
 *
 * a0fmem.c : wa0 Functions of the local memory
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * Feb 16, 1996
 *
 */

#define A0_SOURCE
#include <ath0.h>

#define A0F_SOURCE
#include "a0fmem.h"
#include "a0fread.h"
#include "a0fwrite.h"
#include "a0fcopy.h"
#include "a0flock.h"


#define a0Debug_on(a) printf a
#define a0Debug(a)


/******************local variables & definitions*****************/

/* increment tag until A0MemLastTag then come back to A0MemFirstTag */
#define A0MemFirstTag	0
#define A0MemLastTag	255


/* memory server request port */
a0tPort _a0MemServerPort;

/* memory server data port */
a0tPort _a0MemAskPort;

/* memory server replies port */
a0tPort _a0MemReplyPort;

/* stub tag counter & its mutex */
static int _a0MemStubTag = A0MemFirstTag;
static a0tMutex _a0MemStubTagMutex;

/* datatype to send/receive regions */
a0tDatatype a0_DMA_REGION;

/* datatypes to mem requests */
a0tDatatype a0_MEM_REQUEST, a0_SHORT_MEM_REQUEST, a0_LONG_MEM_REQUEST;

/* datatype to send/receive addresses */
a0tDatatype a0_ADDRESS;

/* dummy synch variable */
int _a0MemDummy;

/* mem daemon thread */
a0tThread _a0MemDaemonThread;

#ifdef A0_TRACE
/* mutex and counter to generate trace id's to regions */
static a0tMutex _a0MemTraceIdMutex;
int _a0NewTraceId = 0;
#endif

/* termination chain, stores the a0Init's termination chain call */
void (*_a0MemTerminateChain)(void);


/*************************************************************************
	_a0NewDMARegion - makes a region reference to a local address
*************************************************************************/

extern a0tError _a0NewDMARegion(a0tDMARegion *region,
			a0tFormat *formatp, void *pointer, int size) {

  region->node = a0SelfNode;
  region->format.address = formatp;
  region->data.address = pointer;
  region->size = size;

#ifdef A0_TRACE
  _a0LockMutex(&_a0MemTraceIdMutex);
  region->trace_id = _a0NewTraceId++;
  _a0UnlockMutex(&_a0MemTraceIdMutex);
#endif

  return A0ErrOk;
}


/*************************************************************************
_a0FindMemTag - calculates a new tag for memory communication
*************************************************************************/

extern a0tError _a0FindMemTag(int *tagp) {

  if (_a0LockMutex(&_a0MemStubTagMutex)) _a0Error(A0ErrMutexLock, "mem");

  *tagp = _a0MemStubTag ++;
  /* if all tags were used, restart it */
  if (_a0MemStubTag > A0MemLastTag) _a0MemStubTag = A0MemFirstTag;

  if (_a0UnlockMutex(&_a0MemStubTagMutex)) _a0Error(A0ErrMutexUnlock, "mem");
}


/*************************************************************************
	_a0MemServer
*************************************************************************/

static a0tError _a0MemServer(void *nada)
{
	a0tMemRequest message;
	int run = 1;
	int count_term = 0;

#ifdef A0_TRACE
	{
		/* find a trace id and set it */
		unsigned long id;
		akTraceNewThreadId(&id);
		akTraceSetThreadId(id);
	}
#endif /* A0_TRACE */

	while (run) {
		_a0Receive(&_a0MemServerPort, A0AnySource, 0, NULL,
				a0_LONG_MEM_REQUEST, &message, 1);

		a0Debug(("_a0MemServer got %s request\n",
				message.type==A0ReadRequest?"read":
				message.type==A0WriteRequest?"write":
				message.type==A0LockReadRequest?"lock-read":
				message.type==A0UnlockWriteRequest?"unlock-write":
				message.type==A0LockRequest?"lock":
				message.type==A0UnlockRequest?"unlock":
				message.type==A0MemTerminate?"terminate":
				"unknown"));

		switch (message.type) {
		case A0ReadRequest:
			_a0ReplyRead(&message);
			break;
		case A0WriteRequest:
			_a0ReplyWrite(&message);
			break;
		case A0LockRequest:
		case A0LockReadRequest:
			_a0ReplyLock(&message);
			break;
		case A0UnlockRequest:
		case A0UnlockWriteRequest:
			_a0ReplyUnlock(&message);
			break;
		case A0MemTerminate:
			count_term ++;
			if (count_term == akSize)
				run = 0;
		}
		a0Debug(("_a0MemServer served.\n"));
	}
	a0Debug(("_a0MemServer terminating.\n"));
}


/*************************************************************************
		_a0TerminateDMA
*************************************************************************/

static void _a0TerminateDMA(void)
{
	a0tMemRequest message;
	int i;
	a0tError r;
	void *res;

	if (_a0MemTerminateChain)
		_a0MemTerminateChain();

	message.type = A0MemTerminate;
	for (i=0; i<akSize; i++)
		_a0ISend(&_a0MemServerPort, i, 0, NULL,
			a0_SHORT_MEM_REQUEST, &message, 1);

	/* wait local damon to die */
	akpthread_join(_a0MemDaemonThread.pth_id, &res);
}


/*************************************************************************
		_a0UseDMA - mem module init
*************************************************************************/

extern a0tError _a0UseDMA(void) {

  a0tDMARegion	region[2];
  a0tMemoryAddress address[2];
  a0tMemRequest message[2];
  a0tError	ret;
  int i;
  a0tThread daemon;

  /* init servers */
  _a0InitCopy();
  _a0InitLocks();

  /* Create a Port to the memory service */
  if (ret = _a0NewPort(&_a0MemAskPort, A0MemLastTag+1, 0))
    return ret;

  /* Create a Port to replies to the memory service */
  if (ret = _a0NewPort(&_a0MemReplyPort, A0MemLastTag+1, 0))
    return ret;

  /* create mutex to protect the StubTag incrementation */
  if (ret = _a0NewMutex(&_a0MemStubTagMutex))
    return ret;

  /* create MPI struct a0tMemoryAddress */
  _a0NewStructDatatype(&a0_ADDRESS, &address[0], 3,
			&address[0],		1,		a0_LB,
			address[0].bytes,	A0MaxAddressSize, a0_BYTE,
			&address[1],		1,		a0_UB);

  /* store in format _a0FAddress the datatype just created */
  _a0SetBasicFormat(A0FAddress, a0_ADDRESS);

  /* create MPI struct a0tDMARegion.{site, address} */
#ifdef A0_TRACE
  _a0NewStructDatatype(&a0_DMA_REGION, &region[0], 7,
  			&region[0],		1,		a0_LB,
			&region[0].node,	1, 		a0_INT,
			region[0].format.bytes, A0MaxAddressSize,a0_BYTE,
			region[0].data.bytes,  A0MaxAddressSize,a0_BYTE,
			&region[0].size,	1, 		a0_INT,
			&region[0].trace_id,	1, 		a0_INT,
			&region[1],		1,		a0_UB);
#else /* not A0_TRACE */
  _a0NewStructDatatype(&a0_DMA_REGION, &region[0], 6,
  			&region[0],		1,		a0_LB,
			&region[0].node,	1, 		a0_INT,
			region[0].format.bytes, A0MaxAddressSize,a0_BYTE,
			region[0].data.bytes,  A0MaxAddressSize,a0_BYTE,
			&region[0].size,	1, 		a0_INT,
			&region[1],		1,		a0_UB);
#endif /* A0_TRACE */

  /* store in format _a0FDMARegion the datatype just created */
  _a0SetBasicFormat(A0FDMARegion, a0_DMA_REGION);

#ifdef A0_TRACE
  /* create mutex to protect the trace_id incrementation */
  if (ret = _a0NewMutex(&_a0MemTraceIdMutex))
    return ret;
#endif

  /* messages to mem server */
  _a0NewStructDatatype(&a0_LONG_MEM_REQUEST, &message[0], 10,
  			&message[0],		1,		a0_LB,
  			&message[0].type,	1, 		a0_INT,
  			&message[0].region,	1,		a0_DMA_REGION,
  			&message[0].region_start, 1, 		a0_INT,
  			&message[0].region_size, 1, 		a0_INT,
  			&message[0].reply_node,	1, 		a0_INT,
  			&message[0].reply_tag,	1, 		a0_INT,
  			&message[0].writer_node, 1, 		a0_INT,
  			&message[0].writer_tag,	1, 		a0_INT,
			&message[1],		1,		a0_UB);
  _a0NewStructDatatype(&a0_MEM_REQUEST, &message[0], 8,
  			&message[0],		1,		a0_LB,
  			&message[0].type,	1, 		a0_INT,
  			&message[0].region,	1,		a0_DMA_REGION,
  			&message[0].region_start, 1, 		a0_INT,
  			&message[0].region_size, 1, 		a0_INT,
  			&message[0].reply_node,	1, 		a0_INT,
  			&message[0].reply_tag,	1, 		a0_INT,
			&message[1],		1,		a0_UB);
  _a0NewStructDatatype(&a0_SHORT_MEM_REQUEST, &message[0], 4,
  			&message[0],		1,		a0_LB,
  			&message[0].type,	1, 		a0_INT,
  			&message[0].region,	1,		a0_DMA_REGION,
			&message[1],		1,		a0_UB);

  /* Create a Port to mem server */
  if (ret = _a0NewPort(&_a0MemServerPort, 1, 0))
    return ret;

  /* start mem server */
  _a0NewThread(&_a0MemDaemonThread,
    		 _a0DaemonScheduling,
		 _a0PriDaemonLow,
		 _a0DefaultStack,
		 _a0MemServer,
		 (void *) NULL);

    _a0MemTerminateChain = _a0TerminateChain;
  _a0TerminateChain = _a0TerminateDMA;
  return A0ErrOk;
}

