/*
 *
 * a0fpipe.c : _a0 Pipe format functions
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * June 10, 1996
 *
 */


#define A0_SOURCE
#include <ath0.h>

#define A0F_SOURCE
#include "a0fuser.h"
#include "a0fpipe.h"
#include "a0fxdr.h"


#define a0Debug(a)
#define a0Debug_on(a) printf a


/* flow control. MPI-CH blows up if there is no flow control.
   maybe there's a bug in Athapascan... who knows? */
#define A0PipeMaxSend 5

#define buffer_available(pipe)	(pipe->buffer->size - \
				pipe->buffer->position)
#define buffer_empty_fit(pipe, size)	(pipe->bufsize >= (size))


typedef struct _a0tPipeQueue {
  a0tFormat	*format;
  void		*address;
  int		count;
  a0tRequest	*request_ready;
  struct _a0tPipeQueue *next;
} a0tPipeQueue;


/*************************************************************************
	_a0PipeEnqueue
*************************************************************************/

static a0tError _a0PipeEnqueue(a0tPipe *pipe,
	int protocol, unsigned long *serial_nop,
	a0tFormat *format, void *address, int count,
	a0tRequest *request) {

  int bad;
  a0tPipeQueue *prq;
#ifdef A0_TRACE
  a0tRequest	null_ifput;
#endif

  /* check if pipe in not synchronous */
  if (pipe->flags & A0SyncPipe) return A0ErrInvPipe;

  /* check/set input flag */
  switch (protocol) {
  case A0InputPipe:	bad = A0OutputPipe; break;
  case A0OutputPipe:	bad = A0InputPipe; break;
  default:		bad = 0; break;
  }

  if (pipe->flags & bad) return A0ErrInvPipe;
  pipe->flags |= protocol;

  prq = (a0tPipeQueue *)malloc(sizeof(a0tPipeQueue));
  if (!prq) return A0ErrNoMem;

  /* remember all put/get data */
  prq->format = format;
  prq->address = address;
  prq->count = count;
  prq->request_ready = request;

#ifdef A0_TRACE

  /* a request is needed to the tracer: it must get its serial number. */
  if (!request) request = &null_ifput;
  akBlock(request, 0 ,0);
  if (serial_nop) *serial_nop = request->serial_no;
  if (request == &null_ifput) akNoWait(&null_ifput);

#else /* no A0_TRACE */

  /* ask a blocked request to the kernel */
  if (request) akBlock(request, 0, 0);

#endif

  /* Put the data element at the tail of the queue. */
  _a0LockMutex(pipe->mutex);

  /* if immediate puts happen, wake up the daemon if queue was empty */
  if (!pipe->queue_head)
    _a0VSemaphore(pipe->sleep);

  *pipe->queue_tail = prq;
  pipe->queue_tail = &(prq->next);
  prq->next = 0;

  _a0UnlockMutex(pipe->mutex);

  a0Debug(("%d:_a0PipeEnqueue (%X)\n", a0SelfNode, prq));
}


/*************************************************************************
	_a0PipeRequeue
*************************************************************************/

static void _a0PipeRequeue(a0tPipe *pipe, a0tPipeQueue *prq) {

  /* Put the data element at the head of the queue. */

  _a0LockMutex(pipe->mutex);

  prq->next = pipe->queue_head;
  pipe->queue_head = prq;

  _a0UnlockMutex(pipe->mutex);

  a0Debug(("%d:_a0PipeRequeue (%X)\n", a0SelfNode, prq));
}


/*************************************************************************
	_a0PipeDequeue
*************************************************************************/

static void _a0PipeDequeue(a0tPipe *pipe, a0tPipeQueue **new) {

  /* Take a data element out of the head of the queue. */

  a0tPipeQueue *prq;

  _a0LockMutex(pipe->mutex);

  prq = pipe->queue_head;
  *new = prq;
  if (!prq) {
    _a0UnlockMutex(pipe->mutex);
    return;
  }

  if (pipe->queue_tail == &(prq->next))
    pipe->queue_tail = &(pipe->queue_head);
  pipe->queue_head = prq->next;

  _a0UnlockMutex(pipe->mutex);

  a0Debug(("%d:_a0PipeDequeue (%X)\n", a0SelfNode, prq));
}


/*************************************************************************
	_a0PipeFreeQueue
*************************************************************************/

static a0tError _a0PipeFreeQueue(a0tPipeQueue *data) {

  /* if there is a request, unblock it */
  if (data->request_ready)
    akUnblock(data->request_ready);

  a0Debug(("%d:_a0PipeFreeQueue (%X)\n", a0SelfNode, data));

  free(data);
}


/************************************************************************
	_a0PipeSendBuffer - 
************************************************************************/

static a0tError _a0PipeSendBuffer(a0tPipe *pipe) {

  a0tError ret;

  a0Debug(("%d:_a0PipeSendBuffer: BEGIN p=%d:%d n=%d t=%d %X %X\n",
		a0SelfNode, pipe->port->comm, pipe->port->basetag,
		pipe->node, pipe->tag,
		pipe->request, pipe->buffer));

#ifdef DEBUG_DUMP_BUFFER
{
  int i;
  char tab[1024]="";
  for (i=0; i<pipe->buffer->size; i++)
    sprintf(tab+i*3, "%02x ", ((char*)pipe->buffer->storage)[i]);
  printf("out=[%s]\n", tab);
}
#endif

  if (pipe->flags & A0PipeHasData) {
    a0Debug(("%d:_a0PipeSendBuffer: send it\n", a0SelfNode));

#ifndef A0_TRACE
    ret = _a0SendBuffer(pipe->port, pipe->node, pipe->tag,
				pipe->request, pipe->buffer);
#else
    ret = _a0_perf_SendBuffer(pipe->port, pipe->node, pipe->tag,
				pipe->request,	pipe->buffer);
#endif /* A0_TRACE */

    if (ret) return ret;
  }

  if ((pipe->flags & A0FlowPipe) && ((++pipe->flowct)==A0PipeMaxSend)) {
    int i;
    _a0Receive(pipe->port, pipe->node, pipe->tag, pipe->request,
		a0_INT, &i, 1);
    pipe->flowct = 0;
  }

  ret = _a0ClearBuffer(pipe->buffer);
  pipe->flags &= ~A0PipeHasData;

  a0Debug(("%d:_a0PipeSendBuffer: END (%x)\n", a0SelfNode, pipe));
  return ret;
}


/************************************************************************
	_a0PipeReceiveBuffer - 
************************************************************************/

static a0tError _a0PipeReceiveBuffer(a0tPipe *pipe) {

  a0tError ret;
  a0tRequestData data;

  a0Debug(("%d:_a0PipeReceiveBuffer: BEGIN p=%d:%d n=%d t=%d %X %X\n",
		a0SelfNode, pipe->port->comm, pipe->port->basetag,
		pipe->node, pipe->tag,
		pipe->request, pipe->buffer));

  /* set buffer size to the original size */
  pipe->buffer->size = pipe->initsiz;

#ifndef A0_TRACE
  ret = _a0ReceiveBuffer(pipe->port, pipe->node, pipe->tag,
				pipe->request, pipe->buffer);
#else
  ret = _a0_perf_ReceiveBuffer(pipe->port, pipe->node, pipe->tag,
				pipe->request,	pipe->buffer);
#endif /* A0_TRACE */

  if (ret) return ret;

#ifdef DEBUG_DUMP_BUFFER
{
  int i;
  char tab[1024]="";
  for (i=0; i<pipe->buffer->size; i++)
    sprintf(tab+i*3, "%02x ", ((char*)pipe->buffer->storage)[i]);
  printf("in=[%s]\n", tab);
}
#endif

  if ((pipe->flags & A0FlowPipe) && ((++pipe->flowct)==A0PipeMaxSend)) {
    int i;
    _a0Send(pipe->port, pipe->node, pipe->tag, pipe->request,
		a0_INT, &i, 1);
    pipe->flowct = 0;
  }

  /* set buffer size to the received size */
  ret = _a0GetRequestData(pipe->request, &data);
  pipe->buffer->size = data.size;

  a0Debug(("%d:_a0PipeReceiveBuffer: DONE\n", a0SelfNode));
  return ret;
}


/************************************************************************
		_a0PipeDaemon - input/output pipe daemon
************************************************************************/

static a0tError _a0PipeDaemon(a0tPipe *pipe) {

  a0tPipeQueue *data;
  a0tError ret;
  a0tPackFunction function = 0;
  int must_run = 1;

  a0Debug(("%d:_a0PipeDaemon: (%X) BEGIN\n", a0SelfNode, pipe));

  while (must_run) {

    a0Debug(("%d:_a0PipeDaemon: (%X) go to sleep\n", a0SelfNode, pipe));

#ifndef A0_TRACE
    /* sleeps while there's nothing in queue */
    _a0PSemaphore(pipe->sleep);
#else
    /* trace the P */
    _a0_perf_PSemaphore(pipe->sleep);
#endif /* A0_TRACE */

    a0Debug(("%d:_a0PipeDaemon: (%X) just awaken\n", a0SelfNode, pipe));

    if (!function) {
      /* the first time the daemon wakes up, it discovers its direction */
      if (pipe->flags & A0OutputPipe) {
	a0Debug(("%d:_a0PipeDaemon: (%X) I'm OUTput\n", a0SelfNode, pipe));
	function = _a0FPack;
      }
      else
      if (pipe->flags & A0InputPipe) {
	a0Debug(("%d:_a0PipeDaemon: (%X) I'm INput\n", a0SelfNode, pipe));
	function = _a0FUnpack;
      }
      else
        _a0Error(A0ErrFatal, "pipe daemon I/O?");
     }

    _a0PipeDequeue(pipe, &data);

    while (data) {
      /* NULL format to flush, not NULL, pack or unpack */
      if (data->format) {
	a0Debug(("%d:_a0PipeDaemon (%X) will %spack\n", a0SelfNode, pipe,
			(pipe->flags & A0OutputPipe)?"":"un"));
	if (ret = function(pipe->fake_buffer,
				data->format, data->address, data->count))
		break;

	a0Debug(("%d:_a0PipeDaemon (%X) %spack done\n", a0SelfNode, pipe,
			(pipe->flags & A0OutputPipe)?"":"un"));
      }
      else {
 	a0Debug(("%d:_a0PipeDaemon (%X) will flush\n", a0SelfNode,pipe));
	if (!(pipe->flags & A0PipeOpen)) {
	  /* this is the final flush */
	  must_run = 0;
 	  a0Debug(("%d:_a0PipeDaemon (%X) FINAL flush\n", a0SelfNode,pipe));
	}

	/* fake request, just to flush */
	if (pipe->flags & A0OutputPipe) {
	  if (ret = _a0PipeSendBuffer(pipe)) break;
	}
 	a0Debug(("%d:_a0PipeDaemon (%X) flush done\n", a0SelfNode,pipe));
      }
      _a0PipeFreeQueue(data);
      _a0PipeDequeue(pipe, &data);
    }

    a0Debug(("%d:_a0PipeDaemon: (%X) !data (%d)\n", a0SelfNode, pipe, ret));

    if (ret) break;
  }

  a0Debug(("%d:_a0PipeDaemon: (%X) END (%d)\n", a0SelfNode, pipe, ret));

  _a0ExitThread((void *) ret);
}


/*************************************************************************
	_a0IFPut
*************************************************************************/

extern a0tError _a0IFPut(a0tPipe *pipe, a0tRequest *request,
			a0tFormat *format, void *address, int count) {

  /* User-level put function. It checks the end of a possible
     former communication, allocates space to store the data and
     put it at queue tail. If communication ended, try to pack
     data into buffer and send it if nedded. If not ended, check
     if next data can be sent anyway. */

  a0Debug(("%d:_a0IFPut (%X) %X %X %d %X\n", a0SelfNode,
		pipe, format, address, count, request));

  /* enqueue data */
  return _a0PipeEnqueue(pipe, A0OutputPipe, NULL,
			format, address, count, request);
}


/*************************************************************************
	_a0IFGet
*************************************************************************/

extern a0tError _a0IFGet(a0tPipe *pipe, a0tRequest *request,
			a0tFormat *format, void *address, int count) {

  /* User-level get function. It checks the end of a possible
     former communication, allocates space to store the data and
     put it at queue tail. If communication ended, try to unpack
     data from buffer and receive it if nedded. If not ended, check
     if next data can be received anyway. */

  /* See _a0FIPut for trace explanations. */

  a0Debug(("%d:_a0IFGet (%X) %X %X %d %X\n", a0SelfNode,
		pipe, format, address, count, request));

  /* enqueue data */
  return _a0PipeEnqueue(pipe, A0InputPipe, NULL,
			format, address, count, request);

}


#ifdef A0_TRACE
/*************************************************************************
	_a0IFPutTrace
*************************************************************************/

extern a0tError _a0IFPutTrace(a0tPipe *pipe, a0tRequest *request,
			unsigned long *serial_nop,
			a0tFormat *format, void *address, int count) {

  /* This function was modified for the tracer, adding serial_nop.
     When tracing an a0IFPut with a NULL request, the tracer should pass
     a substitute request to _a0IFPut, get the request serial number,
     and call akNoWait right after, to make the kernel behave as if
     the request had been NULL. But the pipe request queue keeps the
     pointer to the request sent with akBlock to the kernel, in order
     to akUnblock it when it completes. So, the solution is as follows:
     the tracer passes the user request to _a0IFPut, even if it's NULL.
     If a NULL request comes to _a0PipeEnqueue, it uses an internal
     request field instead. It calls akBlock to get a serial number to
     return to the tracer and call akNoWait, to tell the kernel to forget
     about it after completion. Same for _a0IFGet. */

  /* enqueue data */
  return _a0PipeEnqueue(pipe, A0OutputPipe, serial_nop,
			format, address, count, request);
}


/*************************************************************************
	_a0IFGetTrace
*************************************************************************/

extern a0tError _a0IFGetTrace(a0tPipe *pipe, a0tRequest *request,
			unsigned long *serial_nop,
			a0tFormat *format, void *address, int count) {

  /* See _a0FIPutTrace for trace explanations. */

  /* enqueue data */
  return _a0PipeEnqueue(pipe, A0InputPipe, serial_nop,
			format, address, count, request);
}
#endif


/*************************************************************************
	_a0Flush
*************************************************************************/

extern a0tError _a0Flush(a0tPipe *pipe) {

  /* User-level flush function. It puts a fake data descriptor in the
     queue and waits its completion. At that moment, the daemon 
     guarantees that all requests are performed and the buffer is empty
     deblocking the calling thread.*/

  a0tError ret;
  a0tRequest request;

  a0Debug(("%d:_a0Flush (%X) BEGIN\n", a0SelfNode, pipe));

  /* if pipe is synchronous, just send the buffer */

  if (pipe->flags & A0SyncPipe) {
    a0Debug(("%d:_a0Flush (%X) sync pipe\n", a0SelfNode, pipe));
    if (pipe->flags & A0OutputPipe)
      return _a0PipeSendBuffer(pipe);
    else
      return A0ErrOk;
  }

  a0Debug(("%d:_a0Flush (%X) default pipe\n", a0SelfNode, pipe));

  /* enqueue fake data */
  if (ret = _a0PipeEnqueue(pipe, 0, 0, 0, 0, 0, &request))
	return ret;

  a0Debug(("%d:_a0Flush (%X) queued\n", a0SelfNode, pipe));

  ret = _a0WaitRequest(&request);

  a0Debug(("%d:_a0Flush (%X) END\n", a0SelfNode, pipe));

  return ret;
}


/*************************************************************************
	_a0FPut
*************************************************************************/

extern a0tError _a0FPut(a0tPipe *pipe, a0tRequest *request,
	a0tFormat *format, void *address, int count) {

  /* blocking version of _a0IFPut */

  a0tError ret;
  a0tRequest aux_req;

  a0Debug(("%d:_a0FPut (%X) BEGIN %X %X %d %X\n", a0SelfNode,
		pipe, format, address, count, request));

  if (pipe->flags & A0SyncPipe) {

    /* check/set output flag */
    if (pipe->flags & A0InputPipe) return A0ErrInvPipe;
    pipe->flags |= A0OutputPipe;

    _a0LockMutex(pipe->mutex);
    ret = _a0FPack(pipe->fake_buffer, format, address, count);
    _a0UnlockMutex(pipe->mutex);
    if (request) request->wc = 0;
  }
  else {
    if (!request) request = &aux_req;
    if (ret = _a0IFPut(pipe, request, format, address, count))
	return ret;

    ret = _a0WaitRequest(request);
  }

  a0Debug(("%d:_a0FPut (%X) END %X %X %d %X\n", a0SelfNode,
		pipe, format, address, count, request));
  return ret;
}


/*************************************************************************
	_a0FGet
*************************************************************************/

extern a0tError _a0FGet(a0tPipe *pipe, a0tRequest *request,
	a0tFormat *format, void *address, int count) {

  /* blocking version of _a0IFGet */

  a0tError ret;
  a0tRequest aux_req;

  a0Debug(("%d:_a0FGet (%X) BEGIN %X %X %d %X\n", a0SelfNode,
		pipe, format, address, count, request));

  if (pipe->flags & A0SyncPipe) {

    /* check/set input flag */
    if (pipe->flags & A0OutputPipe) return A0ErrInvPipe;
    pipe->flags |= A0InputPipe;

    _a0LockMutex(pipe->mutex);
    ret = _a0FUnpack(pipe->fake_buffer, format, address, count);
    _a0UnlockMutex(pipe->mutex);
    if (request) request->wc = 0;
    return ret;
  }
  else {
    if (!request) request = &aux_req;
    if (ret = _a0IFGet(pipe, request, format, address, count))
	return ret;

    ret = _a0WaitRequest(request);
  }
  a0Debug(("%d:_a0FGet (%X) END %X %X %d %X\n", a0SelfNode,
		pipe, format, address, count, request));

  return ret;
}


/*************************************************************************
	_a0NewPipe
*************************************************************************/

extern a0tError _a0NewPipe(a0tPipe *newpipe,
			a0tPort *port, int node, int tag,
			size_t bufsize, int protocol, int prio) {

/*	newpipe: new pipe to create
	port:	port to use to communicate (must be created before)
	node:	node to talk to (dest in put, src in get)
	tag:	tag to use in the port
	bufsize: size of the buffer to use (can be A0DefaultPipeSize)
	protocol: protocol to use (can be A0AsyncPipe)
	prio:	priority of the pipe (asynchronous)  */

  a0tError ret;

  newpipe->port = 0;
  newpipe->port = port;
  newpipe->tag = tag;
  newpipe->node = node;
  newpipe->flags = protocol | A0PipeOpen;
  newpipe->queue_head = 0;
  newpipe->queue_tail = &(newpipe->queue_head);

  ret = _a0NewBuffer(newpipe->buffer, A0SendBufferType, bufsize);
  if (ret) return ret;

  newpipe->initsiz = newpipe->buffer->size;
  newpipe->bufsize = bufsize;
#if defined(A0_MPICH) || defined(A0_MPIT3E)
  newpipe->flags |= A0FlowPipe;
#endif
  newpipe->flowct = 0;

  newpipe->fake_buffer->bufftype = A0PipeBufferType;
  newpipe->fake_buffer->storage = (void*) newpipe;

  /* The mutex is used for two different things. If the pipe has NOT
     A0SyncPipe, the mutex is used to enforce mutual exclusion
     to insert asynchronous requests in a queue. Puts and gets are
     never mixed, since they are all performed by the secondary thread.
     If the pipe has A0SyncPipe, the mutex serves to avoid two
     threads to interleave their calls to put/get.
  */

  a0Debug(("%d:_a0NewPipe (%X) initsiz=%d bufsiz=%d\n",
		a0SelfNode, newpipe, newpipe->initsiz,
		newpipe->bufsize));

  if (ret = _a0NewMutex(newpipe->mutex)) return ret;
  if (!(protocol & A0SyncPipe)) {
    if (ret = _a0NewSemaphore(newpipe->sleep, 0)) return ret;
    if (ret = _a0NewSlave(newpipe->daemon, A0InheritScheduling,
			prio, A0DefaultStack,
			(a0tFunction) _a0PipeDaemon, (void *) newpipe))
	return ret;
  }

  return A0ErrOk;
}


/*************************************************************************
	_a0DisposePipe
*************************************************************************/

extern a0tError _a0DisposePipe(a0tPipe *pipe) {

  void *dummy;
  a0tError ret;

  a0Debug(("%d:_a0DisposePipe (%X) BEGIN\n", a0SelfNode, pipe));

  /* signal the daemon to die */
  pipe->flags &= ~A0PipeOpen;
  if (ret = _a0Flush(pipe)) return ret;
  a0Debug(("%d:_a0DisposePipe (%X) flushed\n", a0SelfNode, pipe));

  if (!(pipe->flags & A0SyncPipe))
    _a0JoinThread(pipe->daemon, &dummy);

  a0Debug(("%d:_a0DisposePipe (%X) joined\n", a0SelfNode, pipe));

  pipe->port = 0;
  ret = _a0DisposeBuffer(pipe->buffer);

  a0Debug(("%d:_a0DisposePipe (%X) END\n", a0SelfNode, pipe));
  return ret;
}


/************************************************************************
	_a0PipePack - 
************************************************************************/

extern a0tError _a0PipePack(a0tBuffer *buffer,
		 a0tFormat *formatp, void *address, int count) {

  int size, space;
  a0tError ret;
  a0tPipe *pipe = (a0tPipe *) buffer->storage;
  void *offset;

  /* Only map formats arrive here. Get one item's size */
  if (ret = _a0GetMapSize(formatp, address, 1, &size))
	return ret;

  /* one element alone must fit in the buffer */
  /* TODO: new versions should be able to cut a map format. */
  if (!buffer_empty_fit(pipe, size))
	_a0Error(A0ErrFatal, "pipe pack elem>buff");

  while (count) {
    if (!(space = (buffer_available(pipe)/size))) {
      /* no more space. send it */
      if (ret = _a0PipeSendBuffer(pipe)) return ret;
      space = buffer_available(pipe)/size;
      /* space MUST be > 0 now */
    }

    /* space is the number of items that could fit in buffer */
    if (space > count) space = count;
    a0Debug(("%d:_a0PipePack (%X) packing %d elements\n",
		a0SelfNode, pipe, space));
    if (ret = _a0MapPack(pipe->buffer, formatp, address, space))
	return ret;
    pipe->flags |= A0PipeHasData;

    if (count -= space) {
      /* more items to pack. skip space items and restart */
      if (ret = _a0GetMapOffset(formatp, address, space, &offset))
	return ret;
      address = offset;
    }
  }

  return A0ErrOk;
}


/************************************************************************
	_a0PipeUnpack - 
************************************************************************/

extern a0tError _a0PipeUnpack(a0tBuffer *buffer,
		 a0tFormat *formatp, void *address, int count) {

  int size, space;
  a0tError ret;
  a0tPipe *pipe = (a0tPipe *) buffer->storage;
  void *offset;

  /* Only map formats arrive here. Get one item's size */
  if (ret = _a0GetMapSize(formatp, address, 1, &size))
	return ret;

  /* one element alone must fit in the buffer */
  if (!buffer_empty_fit(pipe, size))
	_a0Error(A0ErrFatal, "pipe pack elem>buff");

  a0Debug(("%d:_a0PipeUnpack %x\n", a0SelfNode, pipe->flags));

  /* firs call to this function must block and receive a buffer */
  if (!(pipe->flags & A0PipeReceived)) {
    pipe->flags |= A0PipeReceived;
    if (ret = _a0PipeReceiveBuffer(pipe))
	return ret;
  }

  while (count) {
    if (!(space = (buffer_available(pipe)/size))) {
      /* no more data. receive more */
      if (ret = _a0PipeReceiveBuffer(pipe))
	return ret;

      space = buffer_available(pipe)/size;
      /* space MUST be > 0 now */
    }

    /* space is the number of items that could fit in buffer */
    if (space > count) space = count;
    a0Debug(("%d:_a0PipeUnpack (%X) unpacking %d elements\n",
		a0SelfNode, pipe, space));
    if (ret = _a0MapUnpack(pipe->buffer, formatp, address, space))
	return ret;

    if (count -= space) {
      /* more items to pack. skip space items and restart */
      if (ret = _a0GetMapOffset(formatp, address, space, &offset))
	return ret;
      address = offset;
    }
  }

  return A0ErrOk;
}
