/*
 *
 * akmpi.c : mpi stubs for thread safety.
 *
 *
 * Feb 3, 1996. Ilan Ginzburg, Marcelo Pasin.
 *
 * Athapascan Kernel.
 *
 */

#include <pthread.h>
#include <mpi.h>

#define AKERNEL_SOURCE
#include "akthread.h"
#include "akmpi.h"
#include "akerror.h"

/* debugging function */
#define akDebugOff(a) printf a
#define akDebug(a)


/* Mutex protecting accesses to MPI library */
akpthread_mutex_t akDaemonMutex;
/* In the current version, there is only one mutex akDaemonMutex,
   which protects all low-level operations of aKernel. */


/*************************************************************************
 *
 * MPI functions
 *
 * We give higher layers the use of MPI non-communicating and non-
 * blocking functions (datatypes, packing...). Calls must be protected for
 * thread reentrance.
 *
 * The parameters for the functions here are the same as for the associated
 * MPI functions.
 *
 * The #defines of MPI can be used as-is.
 *
 */


/***************************************************************************
			MPI type functions
***************************************************************************/


int akMPI_Type_contiguous(int count, MPI_Datatype oldtype, 
				MPI_Datatype *newtype) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Type_contiguous(count, oldtype, newtype);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Type_vector(int count, int blocklength, int stride, 
			MPI_Datatype oldtype, MPI_Datatype *newtype) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Type_vector(count, blocklength, stride, oldtype, newtype);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Type_hvector(int count, int blocklength, MPI_Aint stride, 
			MPI_Datatype oldtype, MPI_Datatype *newtype) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Type_hvector(count, blocklength, stride, oldtype, newtype);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Type_indexed(int count, int *array_of_blocklengths, 
			int *array_of_displacements, MPI_Datatype oldtype, 
			MPI_Datatype *newtype) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Type_indexed(count, array_of_blocklengths,
			array_of_displacements, oldtype, newtype);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Type_hindexed(int count, int *array_of_blocklengths, 
			MPI_Aint *array_of_displacements, MPI_Datatype oldtype, 
			MPI_Datatype *newtype) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Type_hindexed(count, array_of_blocklengths, 
			array_of_displacements, oldtype, newtype);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Type_struct(int count, int *array_of_blocklengths, 
			MPI_Aint *array_of_displacements, 
			MPI_Datatype *array_of_types, MPI_Datatype *newtype) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Type_struct(count, array_of_blocklengths, 
			array_of_displacements, array_of_types, newtype);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Address(void* location, MPI_Aint *address) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Address(location, address);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Type_extent(MPI_Datatype datatype, MPI_Aint *extent) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Type_extent(datatype, extent);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Type_size(MPI_Datatype datatype, int *size) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Type_size(datatype, (void *) size);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Type_commit(MPI_Datatype *datatype) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Type_commit(datatype);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Get_count(MPI_Status *status, MPI_Datatype datatype, int *count) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Get_count(status, datatype, count);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Pack(void* inbuf, int incount, MPI_Datatype datatype, void *outbuf,
             int outsize, int *position,  MPI_Comm comm) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Pack(inbuf, incount, datatype, outbuf, outsize, position, comm);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Unpack(void* inbuf, int insize, int *position, void *outbuf,
               int outcount, MPI_Datatype datatype, MPI_Comm comm) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Unpack(inbuf, insize, position, outbuf, outcount, datatype, comm);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Pack_size(int incount, MPI_Datatype datatype, MPI_Comm comm,
                  int *size) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Pack_size(incount, datatype, comm, size);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


/*************************************************************************
		Communicator Functions
*************************************************************************/


int akMPI_Comm_rank(MPI_Comm comm, int *rank) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Comm_rank(comm, rank);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Comm_dup(MPI_Comm comm, MPI_Comm *newcomm) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Comm_dup(comm, newcomm);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


int akMPI_Comm_size(MPI_Comm comm, int *size) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Comm_size(comm, size);
  akpthread_mutex_unlock(&akDaemonMutex);
  return r;
}


/***************************************************************************
			MPI functions used only by akernel
***************************************************************************/


int akMPI_Init(int *argcp, char ***argvp) {
  int r;

  /* This function must execute in single-thread mode. Must be
     the first of everything. */

  r = MPI_Init(argcp, argvp);

  /* init MPI mutex */
  if (akpthread_mutex_init(&akDaemonMutex,akPTHREAD_MUTEXATTR_DEFAULT_NP))
	akError(akERRMUTEXINIT, "MPI");

  return r;
}


int akMPI_Finalize(void) {
  int r;

  akpthread_mutex_lock(&akDaemonMutex);
  r = MPI_Finalize();

  /* never unlock the mutex again */
  /* akpthread_mutex_unlock(&akDaemonMutex); */
  /* destroy MPI mutex */
  /* akpthread_mutex_destroy(&akDaemonMutex); */

  return r;
}
