
/*
 *
 * akwc.c : Kernel wait context operations
 *
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * Mar 13, 1997
 *
 */


/* Kernel source file */
#define	AKERNEL_SOURCE
#include "akthread.h"
#include "akwc.h"
#include "akcomm.h"
#include "akdaemon.h"
#include "akerror.h"

/* debugging function */
#define akDebug_on(a) printf a
#define akDebug(a)


/* Wait context table */
aktWaitContext akWCTable[akMAXWC];

/* wait context lists heads */
aktWaitContext *akWCFree = NULL;
aktWaitContext *akWCInUse2 = NULL;

/* wait context InUse2 list counter */
int akWCInUse2Count = 0;


/* dummy handle to give to the daemon store the status, 
   when a request is not expected to be waited or tested on */
static MPI_Status akDummyStatus;


/*************************************************************************
akInsertWCTable - inserts a new table of wait contexts in the free-list.
*************************************************************************/

static int akInsertWCTable(aktWaitContext *tab, int size) {

  int i;

  /* Links and state on free list */
  for (i=0; i<size; i++) {
    tab[i].flags = akWCFREE; /* Other fields we don't care about */
    tab[i].next = tab+i+1; /* Last one will be improperly set */
    akpthread_cond_init(&tab[i].cond, akPTHREAD_CONDATTR_DEFAULT_NP);
  }

  /* Correct last link... */
  tab[akMAXWC-1].next = akWCFree;

  /* Head of free list to first wait context */
  akWCFree = tab;

  return 0;
}


/*************************************************************************
akInitWC - Inits the wait contexts
*************************************************************************/

extern int akInitWC(void) {

  int i;

  akInsertWCTable(akWCTable, akMAXWC);

  return 0;
}


/*************************************************************************
akWCMoveToInUse1 - Finds a free w.c. and turns it into the state InUse1
*************************************************************************/

extern int akWCMoveToInUse1(aktWaitContext **wcpp) {

  aktWaitContext *wcp;

  /* takes the head of the list and advance it*/
  if (!akWCFree) {
    aktWaitContext *tab;

    tab = (aktWaitContext *) malloc (akMAXWC * sizeof(aktWaitContext));
    if (!tab) akError(akERRNOWC, "WCMTU1");
    akInsertWCTable(tab, akMAXWC);
  }

  wcp = akWCFree;
  akWCFree = akWCFree->next;

  wcp->flags = akWCINUSE1;

  *wcpp = wcp;

  return 0;
}


/*************************************************************************
akWCSetNoWait - Sets a W.C. to not be waited for
*************************************************************************/
#ifdef __GNUC__
inline
#endif
int akWCSetNoWait(aktWaitContext *wcp) {

  /* mark flags */
  wcp->flags |= akWCNOWAIT;
 /* give a dummy handle to the daemon store the status */
  wcp->statusp = &akDummyStatus;
}


/*************************************************************************
akWCMoveToInUse2 - Turns a w.c. into the state InUse2 (daemon control)
*************************************************************************/

extern int akWCMoveToInUse2(aktWaitContext *wcp, void *hp) {

  int emptylist;
  aktHandle *handlep = (aktHandle *) hp;

  /* set up the flags */
  wcp->flags &= ~akWCSTATE;
  wcp->flags |= akWCINUSE2;

  /* if it's not going to be waited, mark it. It is going to be 
     set free by the daemon as soon as the send completes. */
  if (!handlep) akWCSetNoWait(wcp);
  else {
    /* remember handle address to the daemon store MPI status */
    wcp->statusp = &(handlep->status);
    /* remember w.c. address to the user wait/test on it */
    handlep->wc = wcp;
  }

  /* inserts in the head of the InUse2 list */
  /* test if list is empty */
  emptylist = ! (int) akWCInUse2;
  wcp->next = akWCInUse2;
  if (akWCInUse2) akWCInUse2->prev = wcp;
  wcp->prev = NULL;
  akWCInUse2 = wcp;

  akWCInUse2Count ++;	/* count one more w.c. */

  akDebug(("akWCMoveToInUse2 :Inserted wc=%X f=%x t=%s\n",
		wcp, wcp->flags,
		((wcp->flags&akWCTYPE)==akWCSEND)?"send":
		(((wcp->flags&akWCTYPE)==akWCRECV)?"recv":
		(((wcp->flags&akWCTYPE)==akWCPROBE)?"probe":
		"other(?)"))));

  /* if the list was empty, wake up the daemon */
  if (emptylist) akAdvance();

  return 0;
}


/*************************************************************************
akWCMoveToInUse3 - Turns a w.c. into the state InUse3 (w.c. completed)
*************************************************************************/
#ifdef __GNUC__
inline
#endif
int akWCMoveToInUse3(aktWaitContext *wcp) {

/* This function takes elem out of the InUse2 list into InUse3 darkness
 * (sorry, but it's not a list :-)
 */
  if ((wcp->flags & akWCSTATE) != akWCINUSE2) return -1;

  /* check if we're stealing the daemon's next w.c. */
  if (wcp == akDaemonNext) akDaemonNext = wcp->next;

  if (wcp->next) wcp->next->prev = wcp->prev;
  if (wcp->prev) wcp->prev->next = wcp->next;
  else akWCInUse2 = wcp->next;
  wcp->flags = (wcp->flags & ~akWCSTATE) | akWCINUSE3;

  akWCInUse2Count --;	/* count one less w.c. */
}


/*************************************************************************
akWCMoveToFree - Turns a w.c. into the state Free (w.c. unused)
*************************************************************************/
#ifdef __GNUC__
inline
#endif
int akWCMoveToFree(aktWaitContext *wcp) {

  /* set up the flags (clean it HERE!) */
  wcp->flags = akWCFREE;

  /* inserts in the head of the Free list */
  wcp->next = akWCFree;
  akWCFree = wcp;

  return 0;
}


#ifdef A0_TRACE
/*************************************************************************
akWCCallbackTrace - unlock the mutex, call the callback, trace completion
*************************************************************************/
#ifdef __GNUC__
inline
#endif
int akWCCallbackTrace(void (*callback)(), void *callback_par,
		int type, MPI_Status *statusp, unsigned long serial_no) {

  /* This function is called by the user interface function or by the
     daemon to call the callback function and trace the completion of a
     request. */

  int size;

  size = 0;
  switch (type) {
  case akWCRECV:
  case akWCPROBE:
    /* get packed data size. MPI can be called, the mutex is locked. */
    if (MPI_Get_count(statusp, MPI_PACKED, &size))
	akError(akERRMPI, "compl-req");
    break;
  }

  if (akpthread_mutex_unlock(&akDaemonMutex))
	akError(akERRMUTEXUNLOCK, "compl-req");

  /* call the function that traces the completion event */
  a0_perf_completed_request(type, serial_no, size,
				statusp->MPI_SOURCE);

  if (callback) callback(callback_par);

  /* if a wait context is passed, the mutex must be relocked,
     because it's called by the daemon. */

  if (akpthread_mutex_lock(&akDaemonMutex))
	akError(akERRMUTEXLOCK, "compl-req");
}
#endif /* A0_TRACE */


#ifndef A0_TRACE
/*************************************************************************
akWCCallback - unlock mutex and call the callback
*************************************************************************/
#ifdef __GNUC__
inline
#endif
int akWCCallback(void (*callback)(), void *callback_par) {

  /* This function is called by the user interface function or by the
     daemon to call the callback function. */

  if (callback) {
    if (akpthread_mutex_unlock(&akDaemonMutex))
	akError(akERRMUTEXUNLOCK, "compl-req");
    (callback)(callback_par);
    if (akpthread_mutex_lock(&akDaemonMutex))
	akError(akERRMUTEXLOCK, "compl-req");
  }
}
#endif /* A0_TRACE */


/*************************************************************************
akWCUnblock - complete and free the w.c., signal waiting thread
*************************************************************************/
#ifdef __GNUC__
inline
#endif
int akWCUnblock(aktWaitContext *wcp) {

  /* unlocks the mutex only once, call callback (and trace) */
#ifdef A0_TRACE
  akWCCallbackTrace(wcp->callback, wcp->callback_par,
		wcp->flags&akWCTYPE, wcp->statusp, wcp->serial_no);
#else
  akWCCallback(wcp->callback, wcp->callback_par);
#endif

  /* Is this request gonna be waited for or tested? */
  if (wcp->flags & akWCNOWAIT) {
    /* not to be waited or tested, free it */
    akWCMoveToFree(wcp);
  }
  else {
    /* Signal on the w.c. condition so the waiting thread is happy */
    wcp->flags |= akWCDONE;
    /* check if it's a part of a wait/test any */
    if (wcp->flags & akWCANY) {
      if (akpthread_cond_signal(wcp->any)<0)
	akError(akERRSEMAP, "V any-daemon");
    }
    else {
      if ((wcp->flags&akWCTYPE)==akWCOTHER)
      akDebug(("akWCUnblock signal cond %x\n", &wcp->cond));
      if (akpthread_cond_signal(&wcp->cond)<0)
	akError(akERRSEMAP, "V daemon");
    }
  }
}


/*************************************************************************
akWCCheck - Check if a wait context is completed
*************************************************************************/

extern int akWCCheck(aktWaitContext *wcp) {

  int completed=0;


  if (wcp->flags & akWCDONE) {
     completed=0;
     return;
  }
   
  switch (wcp->flags & akWCTYPE) {
  case akWCPROBE:

    akDebug(("akWCCheck %X %d:%d comm=%d\n", wcp, wcp->u.probe.source,
    wcp->u.probe.tag, wcp->u.probe.comm));

    if (MPI_Iprobe(wcp->u.probe.source, wcp->u.probe.tag,
		wcp->u.probe.comm, &completed, wcp->statusp))
		akError(akERRMPI, "test daemon2");
    break;
  case akWCSEND:
  case akWCRECV:
    if (MPI_Test(&(wcp->u.request), &completed, wcp->statusp))
		akError(akERRMPI, "test daemon1");
    break;
  default:
    return;
  }

  if (completed) {
    akDebug(("completed wc=%X h=%x t=%s\n", wcp,
		(wcp->statusp==&akDummyStatus)? 0 : wcp->statusp,
		((wcp->flags&akWCTYPE)==akWCSEND)?"send":
		(((wcp->flags&akWCTYPE)==akWCRECV)?"recv":
		(((wcp->flags&akWCTYPE)==akWCPROBE)?"probe":
		"other(?)"))));
       
    akWCMoveToInUse3(wcp);
    akWCUnblock(wcp);
  }
}

