/*
 * a0-thr-rem.c : many sends and receives to different threads
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * May 20, 1996
 *
 * This program tests the creation of remote threads of ath0b.
 *
 * This program can run in any number of nodes. No arguments are
 * expected.
 *
 * It declares two different services for remote execution on each
 * node. Both execute the same function but have different
 * priorities. The main thread of the node 0 starts boths services
 * with remote threads, on all nodes, 5 times. It packs a reply
 * port and a randomly chosen string in the data buffer of the
 * remote threads. The service function receives in its input
 * buffer a reply port and a string. They add all the chars of the
 * string and send the result to the reply port.
 */

#include <a0b.h>
#include <stdio.h>


void PackOrigin(a0tBuffer *buf, a0tPort *port, int node, int tag)
{
	a0Pack(buf, a0_PORT, port, 1);
	a0Pack(buf, a0_INT, &node, 1);
	a0Pack(buf, a0_INT, &tag, 1);
}


void UnpackOrigin(a0tBuffer *buf, a0tPort *port, int *node, int *tag)
{
/*
   printf("unpack origin %d %d\n", buf->size, buf->position);
 */
	a0Unpack(buf, a0_PORT, port, 1);
	a0Unpack(buf, a0_INT, node, 1);
	a0Unpack(buf, a0_INT, tag, 1);
}


a0tError service_fun(a0tBuffer *input)
{
	char message[100];
	int size, sum, sender, tag;
	a0tPort reply;
	a0tRequest req;
	a0tThreadSpecific *data;
/*
	a0GetSpecificData(&data);
	printf("service spc=%X fcn=%X par=%X pri=%d (srv=%X buf=%X)\n",
		data, data->function, data->param, data->Prio,
		service_fun, input);
*/
	UnpackOrigin(input, &reply, &sender, &tag);
	a0Unpack(input, a0_INT, &size, 1);
	a0Unpack(input, a0_CHAR, message, size);
	message[size] = 0;

	sum = 0;
	while (size)
		sum += message[size--];

	printf("service %X node %d got [%s] from node %d replied %d\n",
	       service_fun, a0SelfNode, message, sender, sum);

	a0Send(&reply, sender, tag, &req, a0_INT, &sum, 1);
}


int main(int argc, char *argv[])
{

	int service_1, service_2;
	int i, r, len, rr, r1, r2;
	a0tPort port;
	int prio1, prio2, max;
	a0tRequest request;
	a0tBuffer buffer;
	char *messages[] =
	{
	  "deals", "heterogeneous", "architectures", "model", "promoted",
	      "distributed", "memory", "parallel", "machine", "portable",
	  "cooperative", "computations", "among", "processors", "levels",
		"designed", "exploit", "types", "provided", "runtime",
	      "library", "provides", "primitives", "spread", "forniture",
	};

	a0Init(&argc, &argv);

	switch (A0DefaultScheduling) {
#ifdef A0HasRRScheduling
	case A0RRScheduling:
		max = A0MaxRRPriority;
		break;
#endif
#ifdef A0HasFifoScheduling
	case A0FifoScheduling:
		max = A0MaxFifoPriority;
		break;
#endif
#ifdef A0HasOtherScheduling
	case A0OtherScheduling:
		max = A0MaxOtherPriority;
		break;
#endif
	default:
		fprintf(stderr, "don't know scheduling used\n");
		a0Abort(0);
	}

	prio1 = A0DefaultPriority + 1;
	if (prio1 > max)
		prio1 = max;

	prio2 = A0DefaultPriority + 2;
	if (prio2 > max)
		prio2 = max;

	a0NewService(&service_1, service_fun,
		     A0DefaultScheduling, prio1, A0DefaultStack);
	a0NewService(&service_2, service_fun,
		     A0DefaultScheduling, prio2, A0DefaultStack);

	a0InitCommit();

	srand(time(0) * a0SelfNode);

	a0NewPort(&port, 100, 2);
	a0NewBuffer(&buffer, A0ThreadBufferType, 1000);

	for (rr = 0; rr < 5; rr++) {

		for (i = 0; i < a0NodeCount; i++) {

			/* choose a string */
			r = rand() % 20;
			len = strlen(messages[r]);
			a0ClearBuffer(&buffer);
			PackOrigin(&buffer, &port, a0SelfNode, 0);
			a0Pack(&buffer, a0_INT, &len, 1); /* size of table */
			a0Pack(&buffer, a0_CHAR, messages[r], len);
			a0StartRemoteThread(i, service_1,
				  A0ServiceScheduling, A0ServicePriority,
				      A0ServiceStack, &request, &buffer);

			/* choose another string */
			r = rand() % 20;
			len = strlen(messages[r]);
			a0ClearBuffer(&buffer);
			PackOrigin(&buffer, &port, a0SelfNode, 1);
			a0Pack(&buffer, a0_INT, &len, 1); /* size of table */
			a0Pack(&buffer, a0_CHAR, messages[r], len);
			a0StartRemoteThread(i, service_2,
				  A0ServiceScheduling, A0ServicePriority,
				      A0ServiceStack, &request, &buffer);

		}
		for (i = 0; i < a0NodeCount; i++) {
			a0Receive(&port, i, 0, &request, a0_INT, &r1, 1);
			a0Receive(&port, i, 1, &request, a0_INT, &r2, 1);
			printf("node %d got %d and %d from node %d\n",
			       a0SelfNode, r1, r2, i);
		}

	}

	a0Terminate();
}
