/*
 * ppongath.c: Executes ping-pong using akernel calls
 *             measuring latency and banddwith. Allows multiple
 *             ping-pongs in parallel.
 *
 * by Alexandre Carissimi
 * Apache Project, IMAG, Grenoble, France
 * January 6,1998
 *
 * To execute (after source on a0setup.csh flavour):
 *
 * a0run ppongak -a0n <nodes> [-ping][-ping-pong][-help] [-plot <file>] 
 *          [-size <initial> <final> <stride>] [-log] [-nt <n>] [-reps <reps>] 
 *
 * Modifications:
 *   20.03.98: ping algorithm included
 */

#include	<akernel.h>
#include	<stdio.h>
#include	<sys/types.h>
#include	<sys/time.h>
#include	<unistd.h>
#include	<stdarg.h>
#include	<string.h>
#include	"parser.h"

#define		REPS		10
#define		MILISEC		1000.0
#define		SECONDS         1000000.0
#define		PINGTAG		0


typedef struct {
	int     i;
}Parameters;


int    Reps=100,	     /*times which each message is sended	*/
       MsgLen=0,	     /*message size (initial equal zero)	*/
       FinalSize=65536,      /*final message size			*/
       Stride=1024,          /*increment value for message size	        */
       log=0,		     /*increment step is on a log way           */
       Plot=0;	             /*control flag to generate results file	*/
       
char   *file,
       *TestName,	     /*test to be done 			        */
       *ScaleName;           /*scale type message size: byte,KB,or MB   */
       
double (*ftest)();           /*pointer to test function to be done	*/

double  	ScaleX,
        	ScaleY;


void	HelpMenu(void)
{
   printf("\n");
   printf("# This program allows the following tests: \n");
   printf("#      -ping     : realizes ping algorithm between nodes \n"); 
   printf("#      -ping-pong: realizes ping-pong algorithm between nodes \n");
   printf("#      -help     : prints this help\n#\n");
   printf("#usage: ppongath [-ping][-ping-pong][-help] [-plot <file>] \n");
   printf("        [-size <initial> <final> <stride>] [-log] [-nt <n>] \n");
   printf("        [-reps <reps>] \n\n");
   printf(" 'nt' gives the number of parallel ping executed \n");
   printf("#\n");
}
   

/*
 * This function is defined to exploits particular calls existing
 * on some systems. To compatibility reasons the times is always
 * converted on microseconds.
 */
 
double GetTime(void)
{
   struct timeval time;
   double Time;

   gettimeofday(&time, (struct timezone *) NULL);
   Time = ((double) time.tv_sec)*1000000.0 + time.tv_usec;
   return(Time);
}


double StandardRoundTrip(int rank, int next, int prev,int MsgLen, int Reps, int x)
{
     int        i, j;
     char       *Buffer;
     double     InitialTime, EndTime, ElapsedTime;
     aktHandle  handle;

     Buffer = (char *) malloc(sizeof(char)*(MsgLen?MsgLen:1));
     if (Buffer == NULL) {
         printf("ABORTING: Memory allocation error\n");
         exit(0);
     }
      
     if (rank==0) {
        akRecv(Buffer, 1, MPI_BYTE, prev, PINGTAG+x, MPI_COMM_WORLD, &handle);          
        InitialTime = GetTime();  
        for(j=0;j<Reps;j++) {           
           akSend(Buffer, MsgLen, MPI_BYTE, next, PINGTAG+x, MPI_COMM_WORLD, &handle);
           akRecv(Buffer, MsgLen, MPI_BYTE, prev, PINGTAG+x, MPI_COMM_WORLD, &handle);
        }
        EndTime = GetTime();
        ElapsedTime = (EndTime - InitialTime)/Reps;
     }
     
    if (rank!=0) {
       akSend(Buffer, 1, MPI_BYTE, next, PINGTAG+x, MPI_COMM_WORLD, &handle);            
       for(j=0;j<Reps;j++) {          
          akRecv(Buffer, MsgLen, MPI_BYTE, prev, PINGTAG+x, MPI_COMM_WORLD, &handle);
          akSend(Buffer, MsgLen, MPI_BYTE, next, PINGTAG+x, MPI_COMM_WORLD, &handle);
       }
    }
    
    free(Buffer);
    return(ElapsedTime);
}

double StandardOneWay(int rank, int next, int prev,int MsgLen, int Reps, int x)
{
     int        i, j;
     char       *Buffer;
     double     InitialTime, EndTime, ElapsedTime;
     aktHandle  handle;

     Buffer = (char *) malloc(sizeof(char)*(MsgLen?MsgLen:1));
     if (Buffer == NULL) {
         printf("ABORTING: Memory allocation error\n");
         exit(0);
     }
      
     if (rank==0) {
        akRecv(Buffer, 1, MPI_BYTE, prev, PINGTAG+x, MPI_COMM_WORLD, &handle);          
        InitialTime = GetTime();  
        for(j=0;j<Reps;j++)            
           akSend(Buffer, MsgLen, MPI_BYTE, next, PINGTAG+x, MPI_COMM_WORLD, &handle);
        EndTime = GetTime();
        ElapsedTime = (EndTime - InitialTime)/Reps;
     }
     
    if (rank!=0) {
       akSend(Buffer, 1, MPI_BYTE, next, PINGTAG+x, MPI_COMM_WORLD, &handle);            
       for(j=0;j<Reps;j++)          
          akRecv(Buffer, MsgLen, MPI_BYTE, prev, PINGTAG+x, MPI_COMM_WORLD, &handle);
    }
    
    free(Buffer);
    return(ElapsedTime);
}




/******************************************************************
 *                                                                *
 ******************************************************************/
  
void*	PingPongTest(void *arg) 
{  

   int		MsgSize,
		rank,
		size,
		next,
		prev,
		Tnro,
		i;
		
   double	Average,
   		RoundTrip,
		Latency,
		BandWidth;
		
   
   FILE*	fp;
   char		FileName[256]="",
   		ext[3];
   
   Tnro = ((Parameters *)arg)->i;
	
   rank = akNode;
   size = akSize;
   next = (rank+1)%size;
   prev = (rank+size-1) % size;
 
   if (rank == 0) {  
      fp=stdout;
      if (Plot) {
         strcpy(FileName, file);
         ext[0]='.';ext[1]=(char)(0x30+Tnro);ext[2]='\0';
         strcat(FileName, ext);
         fp = fopen(FileName,"a");
         if (fp==NULL) {
              printf("ERROR: File creation!! \n");
              exit(0);
         }
         fprintf(fp,"#\n#  AKERNEL:%s between %d Nodes \n#\n", TestName, akSize);
         fprintf(fp,"#\n# File: %s \n#\n", FileName);
         fprintf(fp,"# Scale to message size is: %s \n", ScaleName);
         fprintf(fp,"# Msg. Size    One-way(usecs)");
         fprintf(fp,"  RoundTrip(usecs)  BandWidth (MB/s)\n");
      }
      else
         strcpy(FileName, "none");

       printf("#\n#  AKERNEL:%s between %d Nodes\n#\n", TestName, akSize);
       printf("# File: %s \n#\n",FileName);
       printf("# Scale to message size is: %s \n", ScaleName);
       printf("# Msg. Size    One-way(usecs)");
       printf("  RoundTrip(usecs)  BandWidth (MB/s)\n");
   }

   MsgSize = MsgLen;
    
   while (MsgSize <= FinalSize) { 
      Average=0;
      for(i=0; i<REPS; i++)
          Average+=(*ftest)(rank,next,prev,MsgSize,Reps,Tnro);
      if (rank==0) {
      
          RoundTrip = Average/(double)REPS;
	  
          if (MsgSize==0)
             Latency = RoundTrip/ScaleY;
	     
          BandWidth = ((double)MsgSize*ScaleY)/RoundTrip; /* Bytes per microsec  */
	  BandWidth = BandWidth/1048576.0;                /* converts to MBytes  */
	  BandWidth = BandWidth*(double)SECONDS;          /* converts to seconds */
	  
	  printf("%d:",Tnro);
          printf(" %10.5f   %15.7f  %15.7f     %7.3f \n",
	         (double)MsgSize/ScaleX, RoundTrip/ScaleY,RoundTrip*2.0/ScaleY,
		 BandWidth);
		 
          if (Plot)
             fprintf(fp," %10.5f   %15.7f  %15.7f     %7.3f \n", 
	            (double)MsgSize/ScaleX,RoundTrip/ScaleY,RoundTrip*2.0/ScaleY,
		    BandWidth);
          }  
          
       if (log) 
          MsgSize = MsgSize?MsgSize<<1:1;
       else        
          MsgSize+=Stride;
   }
   
   if (rank == 0) {
      printf("#\n#\n#\n");
      if (fp != stdout)
         fclose(fp);
   }
      
   akpthread_exit(0);
}
	

/******************************************************************
 *                                                                *
 ******************************************************************/

	
int	main(int argc, char **argv) 
{
	akpthread_t 	 *ThreadId;
	akpthread_attr_t attr;
	char		 **FileName;
	int	   	 Nt=1,
			 i,
			 aux[3],
			 retcode;
	Parameters	 *arg;


	akInit(&argc, &argv);
	
	if (akSize < 2) {
	   printf("#\n BAD NODES NUMBERS!\n");
	   HelpMenu();
	   exit(0);
	}

       /*
        * Get arguments from the command line.
        */
       
       
        if (ArgFindName(argc, argv, "-help")!=-1) {
            HelpMenu();
	    exit(0);
	}

 	 if (ArgGetInt(argc, argv, "-nt", &Nt)==-1) {
 	     printf("#Missing number of threads  -nt x \n");
 	 exit(0);
 	 }


 	 ThreadId = (akpthread_t *)malloc(sizeof(akpthread_t)*Nt);
 	 arg = (Parameters *)malloc(sizeof(Parameters)*Nt);
 	 if ((arg==NULL) || (ThreadId==NULL)) {
 	    printf("#Memory allocation error on main ...\n");
 	    exit(0);
 	 }
 
         if (ArgGetInt(argc, argv, "-reps", &Reps)==-1) {
             printf("#Missing number of repetitions: -reps x \n");
             exit(0);
         }

         log = 0;
         if (ArgFindName(argc, argv, "-log")!=-1) {
            FinalSize = 1048576<<2;
            log = 1;
         }
 
 	 switch(ArgGetIntList(argc, argv, "-size", 3, aux)) {
 	    case -1:
 	 	    printf("#Missing correct size parameters: \
		           -size initial final stride \n");
 	 	    exit(0);
 	    case  1:
 	 	    MsgLen= aux[0];
 	 	    FinalSize = aux[1];
 	 	    Stride = aux[2];
 	 	    break;
 	    default:
 	 	    break;
 	 }
      
 	 if (FinalSize < 1024) {
	    ScaleX = 1.0;
	    ScaleName="Byte";
	 }
 	 else if (FinalSize < 1048576) {
	    ScaleX = 1024.0;
	    ScaleName="KByte";
	 }
 	 else {
	    ScaleX = 1048576.0;
	    ScaleName="MByte";
	 }
 
 
 	 if (ArgFindName(argc, argv, "-ping")!=-1) {
 	    ftest = StandardOneWay;
 	    TestName = "Ping blocking mode"; 
	    ScaleY = 1.0;
	    if (akSize > 2) {
	       printf("\n Ping test must be done with ONLY 2 nodes\n");
	       exit(0);
	    }	    
 	 }
 	 else {
 	     ftest = StandardRoundTrip;
 	     TestName = "Ping Pong Standard mode"; ScaleY = 2.0;
 	 }
	 
         Plot=ArgGetStr(argc, argv, "-plot", &file);
	
	 akpthread_attr_init(&attr);
	  	 	
	 for(i=0 ; i<Nt; i++) {
	    arg[i].i    = i;
	    retcode = akpthread_create(&ThreadId[i], &attr,
	                               PingPongTest, (void *)&arg[i]);
	    if ( retcode != 0)
	       exit(0);
	 }
	
	 for(i=0; i<Nt; i++) 
	    akpthread_join(ThreadId[i], NULL);
	
	 free(ThreadId);
	 free(arg);
	 akFinalize();
}
